/*{{{  symbol.c*/
/*{{{  decls*/
#include <stdio.h>
#include <stdlib.h>
#define STDIO_H

#include "adaptdat.h"
#include "adout.h"	/* for 'mapvec' macros */
#include "expr.h"
#include "map.h"	/* for 'default_proc_sp' */
#include "output.h"
#include "strmacs.h"
#include "symbol.h"
#include "unit.h"
#include "typedef.h"

/*--------------------------------------------------------------*/
/* 'last_sp' a global, declared in 'symbol.h'.  The only reason	*/
/* why it's not local to this file is because it's used once in	*/
/* unit.c, in 'create_unit_record'.				*/
/*	Maybe we can 'localise' this later!!			*/
/*--------------------------------------------------------------*/
Symbol	*last_sp;

static Symbol  *symtab [HASHSIZE];		/* pointers to symbol table */

/*{{{  'namebuf' decls*/
/*--------------------------*/
/* Name buffer declarations */
/*--------------------------*/
#define  BUFLEN  256	/* (8 * (MAX_ID_LEN+1)); MAX_ID_LEN is	*/
			/* largest space that might be wasted!	*/
typedef struct Name {
	char  buf [BUFLEN];
	struct Name  *next;
} Name;

static Name  *first_np = NULL;
static int    buf_rem = 0;	/* number of free bytes left in name buffer */
static char  *namebufp = NULL_str;

char *make_namebuf ();
/*}}}*/

Align  *copy_align ();
Distr  *copy_distr ();
/*}}}*/

/*{{{  <<< symbol table functions >>>*/
/*------------------------*/
/* symbol table functions */
/*------------------------*/
/*{{{  unsigned hash (char *name)*/
unsigned hash (name)
/*----------------*/
	register char  *name;
/*--------------------------------------------*/
/*    Compute hash value for string 'name'    */
/*--------------------------------------------*/
{
	register unsigned  hashval;

	for (hashval = 0;  *name;  name++)
		hashval = *name + 31 * hashval;

	return  hashval % HASHSIZE;
}
/*}}}*/

/*{{{  Symbol  *lookup_or_install_id (name, len, hashval) -- ok*/
Symbol  *lookup_or_install_id (name, len, hashval)
/*----------------------------------------------*/
	char *name;
	int  len;
	unsigned  hashval;
/*--------------------------------------------------------------*/
/*	If 'name' is found in the symbol table, with an		*/
/* appropriate scope number, return a pointer to its entry,	*/
/* otherwise install it.					*/
/*	In the latter case, note that currently 'name' points	*/
/* to a copy of the name in 'textbuf', but this is only		*/
/* temporary storage - it will be overwritten when the next	*/
/* statement is read.  Hence, this function copies 'name' into	*/
/* the 'namebuf' for long-term storage.				*/
/*--------------------------------------------------------------*/
/*{{{  Note: order of multiple entries of 'name' in symtab*/
/*--------------------------------------------------------------*/
/* Note: order of multiple entries of 'name' in symtab:		*/
/* ----------------------------------------------------		*/
/*	Outside of type defs, names are discovered in the	*/
/* symtab in order of locality -- the most local first.  This	*/
/* is because:							*/
/*								*/
/*  (i) IDs are discovered in the reverse order to their order	*/
/* of installation in the symtab:  the latest entry is		*/
/* discovered first.						*/
/*								*/
/*  (ii) scope_type == SC_LOCAL or SC_DEF_LOCAL, so when a new	*/
/* ID is installed, it's local.					*/
/*								*/
/*  (iii) There is a stack of units currently 'in scope', and	*/
/* only names local to these units are in scope -- those local	*/
/* to other units are 'detached' from the symtab.  (If this	*/
/* were not so, a unit like an interface block or type def	*/
/* intervening between the decl of a name and its reference,	*/
/* could install a duplicate name which would be discovered	*/
/* in the symtab before the entry that's local to *this* unit).	*/ 
/*								*/
/*	However, if we're in a type def, 'scope_type' ==	*/
/* SC_DEF_HOST, so a name local to the *host* can be installed	*/
/* *after* (and hence discovered *before*) a local name.	*/
/*  Hence, the above rule doesn't apply within type defs.	*/
/*--------------------------------------------------------------*/
/*}}}*/
{
	char  *namep = NULL;
	register Symbol  *sp;

	/*----------------------*/
	/* Search symbol table.	*/
	/*----------------------*/
	for (sp = symtab [hashval];  sp;  sp = sp->next) {
		if (str_match (sp->name, name)) {
			switch (scope_type) {
			case SC_DEF_LOCAL :
				/*{{{  */
				/*--------------------------------------------------------------*/
				/* scope_type == SC_DEF_LOCAL					*/
				/* --------------------------					*/
				/* Here, the first 'sp' found in the symtab is the most local,	*/
				/* and hence the one required.					*/
				/*	Note, this conclusion assumes that we're not in a	*/
				/* type def (see intro comments).  This follows because		*/
				/* 'scope_type' can't be SC_DEF_LOCAL in a type def (it's	*/
				/* either SC_LOCAL or SC_DEF_HOST).				*/
				/*--------------------------------------------------------------*/
				
				return sp;	/* 1st one found is most local */
				
				/*}}}*/

			case SC_LOCAL :
				/*{{{  */
				/*--------------------------------------------------------------*/
				/* scope_type == SC_LOCAL					*/
				/* ----------------------					*/
				/* We're looking for an entry of 'name' with *local* scope.	*/
				/* If we're not in a type def, the first entry of 'name' found	*/
				/* in the symtab is the most local, so if it's nonlocal we must	*/
				/* install a new symtab entry with local scope.  This conclusion*/
				/* doesn't hold if we're in a type def (see intro comments).	*/
				/*								*/
				/* N.B.  If a nonlocal entry of 'name' is found, save its	*/
				/* 'sp->name' in 'namep', so that if a new entry is installed	*/
				/* it can point to the same string in 'namebuf'.		*/
				/*--------------------------------------------------------------*/
					
				if (sp->scope == this_unitp->scope)
					return  sp;	/* found local */
				
				else if (this_unitp->attrs & TYPEDEF_LEVEL) {
					/*-----------------------------------------------------------*/
					/* in type def: local may follow nonlocal, so look further   */
					/*-----------------------------------------------------------*/
					namep = sp->name;
					continue;
				}
				else {
					/*-----------------------------------------------------------*/
					/* local can't follow nonlocal, so no local entry in symtab! */
					/*-----------------------------------------------------------*/
					namep = sp->name;
					goto install;
				}			
					
				/*}}}*/

			case SC_DEF_HOST :
				/*{{{  */
				/*--------------------------------------------------------------*/
				/* scope_type == SC_DEF_HOST					*/
				/* -------------------------					*/
				/* Here, we must be in a type def, and we're looking for the	*/
				/* most local entry of 'name' that *isn't* local to the type def*/
				/* itself.  This will be the first *nonlocal* entry discovered	*/
				/* in the symtab.						*/
				/*								*/
				/* N.B.  If a nonlocal entry of 'name' is found, save its	*/
				/* 'sp->name' in 'namep', so that if a new entry is installed	*/
				/* it can point to the same string in 'namebuf'.		*/
				/*--------------------------------------------------------------*/
					
				if (sp->scope == this_unitp->scope) {
					/*----------------------------------------------*/
					/* local not wanted, so look further in symtab! */
					/*----------------------------------------------*/
					namep = sp->name;
					continue;
				}
				
				return  sp;	/* first *nonlocal* is the required one! */
				/*}}}*/
			}
		}
	}
	
install:
	/*--------------------------------------*/
	/* Create a new symbol table entry.	*/
	/*--------------------------------------*/
	sp = (Symbol *) ecalloc (sizeof (Symbol));	/* zero it! */

	/*{{{  attach a permanent copy of 'name' (in 'namebuf') to 'sp'*/
	/*--------------------------------------------------------------*/
	/*	If 'namep' is non-zero, then a permanent copy of 'name'	*/
	/* already exists in 'namebuf', so make 'sp' point to it.	*/
	/* (Note that this means that several 'sp's may point to the	*/
	/* same name entry in 'namebuf').				*/
	/*	Otherwise, make a permanent copy of 'name' in 'namebuf',*/
	/* and attach it to 'sp'.					*/
	/*--------------------------------------------------------------*/
	if (namep)
		sp->name = namep;
	else {
		len++;		/* to include the terminal '\0' */
		if (len > buf_rem)
			namebufp = make_namebuf ();
		strcpy (namebufp, name);
		sp->name = namebufp;
		buf_rem  -= len;
		namebufp += len;
	}
	/*}}}*/

	/*{{{  set 'sp->scope'*/
	/*-----------------*/
	/* set 'sp->scope' */
	/*-----------------*/
	switch (scope_type) {
	case  SC_LOCAL :
	case  SC_DEF_LOCAL :
		sp->scope = this_unitp->scope;
		break;
	
	case  SC_DEF_HOST :
		sp->scope = this_unitp->prev->scope;
		break;
	}
	/*}}}*/

	/*{{{  link 'sp' into the hash table and ID list.*/
	/*----------------------------------------------*/
	/*  link symtab entry into the hash table	*/
	/*----------------------------------------------*/
	sp->next = symtab [hashval];  /* add to front of list */
	symtab [hashval] = sp;
	
	/*--------------------------------------*/
	/*  attach symtab entry to the ID list	*/
	/*--------------------------------------*/
	if (!this_unitp->first_sp)
		this_unitp->first_sp = sp;
	
	if (last_sp)
		last_sp->nextid = sp;
	
	last_sp = sp;
	
	/*}}}*/

	return sp;
}
/*}}}*/
/*{{{  install_alt_return () -- ok*/
install_alt_return ()
/*-----------------*/
/*----------------------------------------------------------------
	This is called by 'yyparse' when an alternate return
dummy argument is parsed.  It sets up a symbol table entry
for the alternate return, with name '*', linked into the
symtab as normal.  This is done to satisfy the method for
checking an actual argument list against the corresponding
dummy list (if the interface is explicit), whereby the dummy
argument list is given by the chain of consecutive symbols
(linked through 'sp->nextid') with attribute AT_DUMMY
immediately following the symbol for the procedure itself.
	Notice that:

-- this symbol can never be confused with a proper
identifier, as its name is not a valid ID name;

-- it's never searched for by 'lookup_or_install_id()':
it's only accessed as described above;

-- if there are several alternate return dummy arguments,
there will be one symtab entry for each, all with the same
name (viz. "*").  This will slightly unbalance the hash
tables -- lookup will be longer for names with the same
hash value as "*"!.

	It's a shame to spoil the symtab by this ugly, ad hoc
addition, but still -- that's Fortran!
----------------------------------------------------------------*/
{
	char  *name = "*";
	int  len = 1;
	Symbol *sp;
	unsigned  hashval = hash (name);

	/*----------------------------------------------*/
	/* Create a new symbol table entry for '*'	*/
	/*----------------------------------------------*/
	sp = (Symbol *) ecalloc (sizeof (Symbol));	/* zero it! */

	/*{{{  make a permanent copy of 'name' (in 'namebuf') & attach to 'sp'*/
	/*--------------------------------------------------------------*/
	/* make a permanent copy of 'name' (in 'namebuf'), and attach	*/
	/* to 'sp'.							*/
	/*--------------------------------------------------------------*/
	len++;		/* to include the terminal '\0' */
	if (len > buf_rem)
		namebufp = make_namebuf ();
	strcpy (namebufp, name);
	sp->name = namebufp;
	buf_rem  -= len;
	namebufp += len;
	/*}}}*/

	/*{{{  set 'sp->scope'*/
	/*-----------------*/
	/* set 'sp->scope' */
	/*-----------------*/
	switch (scope_type) {
	case  SC_LOCAL :
	case  SC_DEF_LOCAL :
		sp->scope = this_unitp->scope;
		break;
	
	case  SC_DEF_HOST :
		sp->scope = this_unitp->prev->scope;
		break;
	}
	/*}}}*/

	/*{{{  link 'sp' into the hash table and ID list.*/
	/*----------------------------------------------*/
	/*  link symtab entry into the hash table	*/
	/*----------------------------------------------*/
	sp->next = symtab [hashval];  /* add to front of list */
	symtab [hashval] = sp;
	
	/*--------------------------------------*/
	/*  attach symtab entry to the ID list	*/
	/*--------------------------------------*/
	if (!this_unitp->first_sp)
		this_unitp->first_sp = sp;
	
	if (last_sp)
		last_sp->nextid = sp;
	
	last_sp = sp;
	/*}}}*/

	/*------------------------*/
	/* Install its attributes */
	/*------------------------*/
	sp->attrs |= AT_DUMMY;
}
/*}}}*/

/*{{{  remove_id (Symbol *sp) -- ok*/
remove_id (remove_sp)
/*-----------------*/
	Symbol *remove_sp;
/*--------------------------------------------------------------*/
/*	Remove an ID with no 'attrs' from the symbol table,	*/
/* to reclaim the space it occupies.  It's only removed if it's	*/
/* local to the current unit, but otherwise there are no	*/
/* restrictions (i.e. 'remove_sp' may be anywhere in the linked	*/
/* list of this unit's symbols!).  Currently, this is only	*/
/* used to remove a symbol corresponding to a letter in an	*/
/* IMPLICIT statement.						*/
/*								*/
/*	N.B. We *don't* remove 'name' from the name buffer,	*/
/* firstly because it's complicated (it may leave a 'hole' in	*/
/* the name buffer) and secondly because we may want to keep	*/
/* the name even if not the symbol table entry (i.e. when we're	*/
/* really just parsing a name rather than an ID, e.g.		*/
/* 'constr_heading'!)						*/
/*	(Q: Is it worth removing symbols??)			*/
/*--------------------------------------------------------------*/
{
	Symbol *sp,  *prev_sp;
	unsigned  hashval;

	/*{{{  detach 'remove_sp' from the ID list - return if not found*/
	/*--------------------------------------------------------------*/
	/* (1) Detach the entry from the ID linked list.		*/
	/* Don't delete it yet, however, as it still has to be detached	*/
	/* from the hash table linked list !				*/
	/*	N.B. If 'remove_sp' isn't found, return immediately	*/
	/* (it *doesn't* point to a symbol table entry!!).		*/
	/*--------------------------------------------------------------*/
	for (sp = this_unitp->first_sp,  prev_sp = this_unitp->prev_sp;
	     sp && (sp != remove_sp);
	     prev_sp = sp,  sp=sp->nextid)
		;
	
	if (!sp)
		return;		/* 'remove_sp' not found! */
	
	/*-------------------------*/
	/* Adjust links in ID list */
	/*-------------------------*/
	if (prev_sp)
		prev_sp->nextid = sp->nextid;
	
	/*---------------------------------------------------------*/
	/* If 'sp' is first in unit, adjust 'this_unitp->first_sp' */
	/*---------------------------------------------------------*/
	if (sp == this_unitp->first_sp)
		this_unitp->first_sp = sp->nextid;
	
	/*------------------------------------------------------*/
	/* If 'sp' is the last in the ID list, adjust 'last_sp' */
	/*------------------------------------------------------*/
	if (sp == last_sp)
		last_sp = prev_sp;
	/*}}}*/
	
	/*{{{  detach 'remove_sp' from the symtab list*/
	/*--------------------------------------------------------------*/
	/* (2)	Detach the entry from the symtab linked	list (if it	*/
	/* hasn't already been detached -- e.g. by 'detach_symtab()).	*/
	/*--------------------------------------------------------------*/
	hashval = hash (remove_sp->name);
	for (sp = symtab [hashval],  prev_sp = NULL_Sym;
	     sp && (sp != remove_sp);
	     prev_sp = sp,  sp = sp->next)
		;
	
	if (sp) {
		if (prev_sp)
			prev_sp->next = sp->next;
		else
			symtab [hashval] = sp->next;
	}
	/*}}}*/
	
	/*{{{  free 'remove_sp'*/
	/*--------------------------------------*/
	/* (3) 'free' the symbol table entry.	*/
	/*--------------------------------------*/
	free  ((char *) remove_sp);
	/*}}}*/
}
/*}}}*/
/*{{{  clear_symtab () -- ok*/
clear_symtab ()
/*-----------*/
/*--------------------------------------------------------------*/
/*	Delete all symbol table entries and their associated	*/
/* records, e.g. 'typeinfo.char_len', 'info.init' & 'arp'.	*/
/*	It's assumed that all the symbol table entries are in	*/
/* the ID linked list, and that the current unit record is the	*/
/* base one (i.e. this_unitp->level = EXTERNAL_LEVEL).		*/
/*--------------------------------------------------------------*/
{
	Symbol  *sp,  *nextid;
	int  i;

	for (sp = this_unitp->first_sp;  sp;  sp = nextid) {
		nextid = sp->nextid;

		/*{{{  delete attached structures as appropriate*/
		/*------------------------------------------------------*/
		/* delete attached structures if present, viz...	*/
		/*------------------------------------------------------*/
		/*--------------------------------------*/
		/* ... sp->typeinfo.char_len  (Expr *)	*/
		/*--------------------------------------*/
		if ((sp->type & TY_CHARACTER)  &&  sp->typeinfo.char_len)
			delete_expr (sp->typeinfo.char_len,  DECR_NLINKS);
		
		
		/*------------------------------*/
		/* ... sp->info.init  (Expr *)	*/
		/*------------------------------*/
		if ((sp->attrs & AT_INIT)  &&  (sp->attrs & (AT_NORMAL_VAR | AT_PARAM)) )
			delete_expr (sp->info.init, DECR_NLINKS);
		
		
		/*-----------------------------------------------------------------*/
		/* ... sp->arp (Array *) & sp->alignp (Align *) or distp (Distr *) */
		/*-----------------------------------------------------------------*/
		if (sp->arp)
			delete_array (sp->arp, DECR_NLINKS);
		
		if (sp->alignp)
			delete_align (sp->alignp, DECR_NLINKS);
		
		if (sp->distp)
			delete_distr (sp->distp, DECR_NLINKS);
		/*}}}*/
#ifdef V_IDA
		/*{{{  delete 'Idref' records attached to 'sp->first_refp'*/
		/*--------------------------------------*/
		/* ... sp->first_refp  (Idref *)	*/
		/*--------------------------------------*/
		if (sp->first_refp)
			delete_ida_recs (sp->first_refp);
		/*}}}*/
#endif
		free ((char *) sp);
	}
	this_unitp->first_sp = this_unitp->prev_sp = last_sp = NULL_Sym;

	for (i = 0;  i < HASHSIZE; i++)
		symtab [i] = NULL_Sym;
}
/*}}}*/
/*{{{  clear_local_symtab () -- ok*/
clear_local_symtab ()
/*-----------------*/
/*--------------------------------------------------------------*/
/*	Delete all symbol table entries associated with the	*/
/* *current* unit (which *isn't* the base unit).		*/
/*	This is invoked by 'set_part_and_label()' when an	*/
/* illegal stmt (i.e. any stmt except a unit header or END	*/
/* stmt) appears in a CONTAINS part, to remove any IDs installed*/
/* in the symbol table as a result of the stmt.  It isn't used	*/
/* in any other circumstance!					*/
/*	It's assumed that all the symbol table entries are in	*/
/* the ID linked list.						*/
/*--------------------------------------------------------------*/
{
	int      hashval;
	Symbol  *sp,  *sp1,  *nextid;

	for (sp = this_unitp->first_sp;  sp;  sp = nextid) {
		nextid = sp->nextid;

		/*{{{  delete attached structures as appropriate*/
		/*------------------------------------------------------*/
		/* delete attached structures if present, viz...	*/
		/*------------------------------------------------------*/
		/*--------------------------------------*/
		/* ... sp->typeinfo.char_len  (Expr *)	*/
		/*--------------------------------------*/
		if ((sp->type & TY_CHARACTER)  &&  sp->typeinfo.char_len)
			delete_expr (sp->typeinfo.char_len,  DECR_NLINKS);
		
		
		/*------------------------------*/
		/* ... sp->info.init  (Expr *)	*/
		/*------------------------------*/
		if ((sp->attrs & AT_INIT)  &&  (sp->attrs & (AT_NORMAL_VAR | AT_PARAM)) )
			delete_expr (sp->info.init, DECR_NLINKS);
		
		
		/*-----------------------------------------------------------------*/
		/* ... sp->arp (Array *) & sp->alignp (Align *) or distp (Distr *) */
		/*-----------------------------------------------------------------*/
		if (sp->arp)
			delete_array (sp->arp, DECR_NLINKS);
		
		if (sp->alignp)
			delete_align (sp->alignp, DECR_NLINKS);
		
		if (sp->distp)
			delete_distr (sp->distp, DECR_NLINKS);
		/*}}}*/
#ifdef V_IDA
		/*{{{  delete 'Idref' records attached to 'sp->first_refp'*/
		/*--------------------------------------*/
		/* ... sp->first_refp  (Idref *)	*/
		/*--------------------------------------*/
		if (sp->first_refp)
			delete_ida_recs (sp->first_refp);
		/*}}}*/
#endif
		/*{{{  rearrange the 'next' links to bypass 'sp'*/
		/*----------------------------------------------*/
		/* rearrange the 'next' links to bypass 'sp'	*/
		/*----------------------------------------------*/
		hashval = hash (sp->name);
		if (symtab [hashval] == sp)
			symtab [hashval] = sp->next;
		else {
			for (sp1 = symtab [hashval];  sp1;  sp1 = sp1->next) {
				if (sp1->next == sp) {
					sp1->next = sp->next;
					break;
				}
			}
		}
		/*}}}*/

		free ((char *) sp);
	}
	this_unitp->first_sp = NULL_Sym;	/* to restart local 'ID list'*/
	last_sp = this_unitp->prev_sp;		/* leave 'prev_sp' unchanged */
}
/*}}}*/
/*{{{  detach_symtab ()       -- ok*/
detach_symtab ()
/*------------*/
/*--------------------------------------------------------------*/
/*	Detach from 'symtab[]' all symbol table entries that	*/
/* are local to the current unit (i.e. whose 'sp->scope' is the	*/
/* same as 'this_unitp->scope), but keep them linked in the	*/
/* 'ID chain'.   This means that they aren't found by		*/
/* 'lookup_id (name)', but they can be accessed by moving along	*/
/* the ID chain from a visible symbol -- i.e. one that *is*	*/
/* found by 'lookup_id (name)'!					*/
/*--------------------------------------------------------------*/
{
	Symbol  *sp,  *sp1;
	int  hashval,  this_scope = this_unitp->scope;

	for (sp = this_unitp->first_sp;  sp;  sp = sp->nextid) {
		if (sp->scope != this_scope)
			continue;

		/*----------------------------------------------*/
		/* 'sp' is local to this unit.  Detach it from	*/
		/* 'symtab[]'.					*/
		/*----------------------------------------------*/
		hashval = hash (sp->name);
		if (symtab [hashval] == sp)
			symtab [hashval] = sp->next;
		else {
			for (sp1 = symtab [hashval];  sp1;  sp1 = sp1->next) {
				if (sp1->next == sp) {
					sp1->next = sp->next;
					break;
				}
			}
		}
	}
	this_unitp->first_sp = NULL;	/* to restart local 'ID list' */
	this_unitp->prev_sp  = last_sp;	/* leave 'last_sp' unchanged */
}
/*}}}*/

/*}}}*/

/*{{{  <<< 'namebuf' handling functions >>>*/
/*------------------------------*/
/* 'namebuf' handling functions */
/*------------------------------*/
/*{{{  char *make_namebuf () -- ok*/
char *make_namebuf ()
/*-----------------*/
/*--------------------------------------------------------------*/
/*	Allocate a new name buffer (of total length BUFLEN)	*/
/* and return a pointer to its first character position.	*/
/*	This is used to hold identifier names, which are	*/
/* installed by 'install_id ()'.  It is assumed that BUFLEN is	*/
/* big enough to hold the longest names that 'install_id ()'	*/
/* tries to install in it.					*/
/*--------------------------------------------------------------*/
{
	Name  *np;

	np = (Name *) emalloc (sizeof (Name));

	np->next = first_np;	/* add to front of list -- may be NULL! */
	first_np = np;

	buf_rem = BUFLEN;

	return  np->buf;	/* 'namebufp' */
}
/*}}}*/
/*{{{  clear_namebuf () -- ok*/
clear_namebuf ()
/*------------*/
/*--------------------------------------*/
/*	Delete all the name buffers.	*/
/*--------------------------------------*/
{
	Name  *np,  *next_np;

	for (np = first_np;  np;  np = next_np) {
		next_np = np->next;
		free  ((char *) np);
	}
	first_np = NULL;
	namebufp = NULL;
	buf_rem = 0;
}

/*}}}*/
/*}}}*/

/*{{{  bool mustbe_spec_expr (ep, sp)*/
bool mustbe_spec_expr (ep, sp)
/*--------------------------*/
	Expr  *ep;
	Symbol  *sp;
{
/*********
mustbe_spec expr:

int, scalar,
ids defined already (i.e. dummy, const or global)
impose restriction -- no fns!!

non-integer bound expr [for 'B']
non-scalar bound expr [for 'B']
bound expr [for 'B'] is not a scalar integer expr
undefined variable X in bound expression [for B]
restriction:  fn ref not permitted in bound expr [for B]
***********/

	return TRUE;
}
/*}}}*/

/*{{{  char *ecalloc (unsigned n)*/
char  *ecalloc (n)
/*--------------*/
	unsigned  n;
/*--------------------------------------------------------------*/
/*  Version of 'calloc' that checks if there is enough memory.	*/
/*  Memory is initialised to zero.				*/
/*--------------------------------------------------------------*/
{
	char  *p;

	p = (char *) calloc (1, n);
	if (!p) {
		fputs ("*** out of memory;  aborting! ***\n", stderr);
		exit (1);
	}
	return p;
}
/*}}}*/
/*{{{  char *emalloc (unsigned n)*/
char  *emalloc (n)
/*--------------*/
	unsigned  n;
/*--------------------------------------------------------------*/
/*  Version of 'malloc' that checks if there is enough memory.	*/
/*  Memory is uninitialised.					*/
/*--------------------------------------------------------------*/
{
	char  *p;

	p = (char *) malloc (n);
	if (!p) {
		fputs ("*** out of memory;  aborting! ***\n", stderr);
		exit (1);
	}
	return p;
}
/*}}}*/

/*{{{  <<< array handling functions >>>*/
/*--------------------------*/
/* array handling functions */
/*--------------------------*/
/*********************************************************/
/*** N.B. STILL HAVE TO INSTALL arp->n[] for (:) !! ******/
/*********************************************************/

/*{{{  Array *create_array () -- ok*/
Array *create_array ()
/*------------------*/
/*------------------------------*/
/* Create an 'Array' record.	*/
/*------------------------------*/
{
	Array  *arp;

	arp = (Array *) ecalloc (sizeof (Array));
	arp->attrs = DEFERRED_SHAPE;		/* initial value */

	return  arp;
}
/*}}}*/
/*{{{  bool  canbe_dimensioned (sp, in_HPF_dir) -- ok*/
bool  canbe_dimensioned (sp, in_HPF_dir)
/*------------------------------------*/
	Symbol  *sp;
	bool  in_HPF_dir;	/* TRUE if in an HPF directive */
/*--------------------------------------------------------------*/
/*	Called by 'yyparse ()' to determine whether an ID	*/
/* ('sp') can be dimensioned.  If the 'array_spec' directly	*/
/* follows the ID, this is called prior to 'create_array ()',	*/
/* while in a typedecl or HPF 'combined-directive' with a	*/
/* DIMENSION attribute and no separate 'array-spec' for the ID,	*/
/* it's called prior to 'install_array ()'.			*/
/*--------------------------------------------------------------*/
{
	int32  sp_attrs = sp->attrs;

	if (sp->arp) {
		errmsg ("dimensions of %q already specified", sp->name);
		return  FALSE;
	}
	else if (sp_attrs & (AT_UNTYPED_ENTITY & ~(AT_PROCESSORS | AT_TEMPLATE))) {
		/*{{{  give errmsg & return FALSE*/
		/*--------------------------------------------------------------*/
		/* An object that can't have a data type can't be an array	*/
		/* (except for HPF processors and templates, which are excluded	*/
		/* from this test).  Report an error, and do one of two things:	*/
		/*								*/
		/* -- if the object type is one of those in AT_OVERRIDE_ATTRS   */
		/* (i.e. AT_CONSTRUCT), then clear this flag and return TRUE.	*/
		/* In other words, forget the object's old identity and give it	*/
		/* a new one!							*/
		/* -- otherwise return FALSE.					*/
		/*--------------------------------------------------------------*/
		switch (sp_attrs & AT_UNTYPED_ENTITY) {
			case  AT_SUBR :			begin_errmsg ("subroutine");	break;
			case  AT_MODULE :		begin_errmsg ("module");	break;
			case  AT_PROGRAM :		begin_errmsg ("main program");	break;
			case  AT_BLOCK_DATA :		begin_errmsg ("block data name");  break;
			case  AT_DERIVED_TYPE :		begin_errmsg ("derived type");	break;
			case  AT_NAMELIST_GROUP :	begin_errmsg ("namelist group");  break;
			case  AT_CONSTRUCT :		begin_errmsg ("construct name");  break;
		}
		end_errmsg (" %q cannot be an array", sp->name);
		
		if (sp->attrs & AT_CONSTRUCT) {
			clear_flags (sp->attrs, AT_CONSTRUCT);
			return  TRUE;
		}
		else
			return  FALSE;
		/*}}}*/
	}
	else if (in_HPF_dir) {
		/*{{{  if 'sp' isn't template/procs, give errmsg & return FALSE*/
		/*----------------------------------------------------------------
		'array-spec' follows an ID in an HPF directive, so the ID should be
		a template or processors arrangement.  I.e. 'sp' mustn't have a
		datatype, and the only 'attrs' flags that may be set are:
		
			(AT_TEMPLATE | AT_PROCESSORS | AT_COMMON_NAME | AT_ACCESS)
		
		(In fact the last probably isn't allowed for an HPF entity either
		 -- but the 'HPF Spec' says nothing about it!).
		
		
		N.B. TEMPLATE / PROCESSORS needn't have been declared yet, so how
		do we check that DIMENSION appeared in right place (i.e. within or
		outside an HPF directive as appropriate) when the entity-type *is*
		discovered?  (Maybe one day we'll try to deduce entity-type from
		any info given here -- but that still wouldn't fully resolve the
		problem.)
		----------------------------------------------------------------*/
		if ((sp->type & TY_TYPE)  ||  ! only_flags_set (sp_attrs,
		          (AT_TEMPLATE | AT_PROCESSORS | AT_COMMON_NAME | AT_ACCESS)) ) {
			errmsg ("cannot specify dimensions of %q in an HPF directive as it is not a template or processors arrangement",
			         sp->name);
			return  FALSE;
		}
		/*}}}*/
	}
	else if (sp_attrs & (AT_PROCESSORS | AT_TEMPLATE)) {
		errmsg ("dimensions of %s %q must be specified in an HPF directive",
		         (sp_attrs & AT_PROCESSORS) ? "processors arrangement" : "template",
		          sp->name);
		return  FALSE;
	}
	else if ((sp_attrs & AT_INIT)  &&  (sp_attrs & (AT_NORMAL_VAR | AT_PARAM))) {
		begin_errmsg ("must specify array dimensions for ");
		if (sp_attrs & AT_PARAM)
			end_errmsg ("named constant %q before its value is declared",
			             sp->name);
		else
			end_errmsg ("%q before it is initialised or referenced",
			             sp->name);
		return  FALSE;
	}
	else if (sp_attrs & AT_EQUIV) {
		errmsg ("must specify array dimensions of %q before it is equivalenced",
		         sp->name);
		return  TRUE;		/* set dims anyway! */
	}
	return  TRUE;		/* OK! */
}
/*}}}*/

/*{{{  install_bounds (arp, sp, l, u) -- ok*/
install_bounds (arp, sp, l, u)
/*--------------------------*/
	Array  *arp;
	Symbol *sp;	/* may be NULL;  used only for errmsgs */
	Expr   *l, *u;	/* either may be NULL */

/*{{{  << EXPLANATION >>*/
/*----------------------------------------------------------------
	'l' and 'u' are the declared lower and upper bounds for
a dimension ('dim') of an array specification 'arp'.
'l' may be an expr or NULL;  'u' may be an expr, NULL or '*' for an
assumed size upper bound.

This function initialises 'arp->l [dim]' and 'arp->n [dim]' as follows:

	     bounds	| arp->l [dim]	|  arp->n [dim]	|
	----------------|---------------|---------------|
explicit		|		|		|
--------    (lb : ub)	|	lb	|   (ub-lb+1)	|
	    (     ub)	|	1	|	ub	|
	    (lb : * )	|	lb	|	*	|
	    (     * )	|	1	|	*	|
			|		|		|
deferred / assumed	|		|		|
------------------	|		|		|
	    (lb :   )	|	lb	|	NULL	|
	    (   :   )	|	NULL	|	NULL	|
	-------------------------------------------------
I.e.:
	if (ub == NULL) {
		arp->l = lb      (may be NULL)
		arp->n  = NULL
	}	
	else {
		arp->l = lb ? lb  : 1
		arp->n  = (ub == '*')  ? '*' : (ub-lb+1)
	}

It also performs the following checks and actions:

 -- checks that the rank <= MAX_RANK;
 -- checks that 'l' and 'u' are legal specification exprs;
 -- classifies the array shape as const, explicit, deferred, assumed,
	assumed-size or mixed (the latter is an error!)

'sp' is passed-in for the purpose of errmsgs.  If it isn't NULL,
then the errmsg refers to 'sp->name'.  Otherwise the array spec
appears in a type decl and no specific identifier is referred to!

N.B.:
(1) 'arp->attrs' is initialised to DEFERRED_SHAPE when 'arp' is created.
(2) If (ub == NULL), i.e. (arp->n [dim] == NULL), then 'process_decls()'
will initialise them for an *assumed shape* dummy array as follows:

	arp->lb [dim]:	NULL -> 1
			expr -- unchanged

	arp->n [dim]:	NULL -> mapvec_name (elem (n,dim))

[At the moment this is even done for a pointer or allocatable array,
to avoid errors, though both should be and remain NULL for 'deferred bounds'.]
----------------------------------------------------------------*/
/*}}}*/
{
	/*{{{  decls*/
	bool   exprs_ok = ! (arp->attrs & NOT_SPEC_EXPR),
	       mixed_shape_errmsg = FALSE;
	char  *name = (sp ? sp->name : NULL);
	int    prev_dim,
	       shape_type = arp->attrs & SHAPE_MASK;
	/*}}}*/
	
	/*{{{  if (rank >= MAX_RANK), give errmsg and return*/
	/*--------------------------------------------------------------*/	
	/* If (rank >= MAX_RANK), give errmsg and ignore this and	*/
	/* subsequent dimensions.					*/
	/*--------------------------------------------------------------*/	
	prev_dim = arp->attrs & RANK_MASK;	/* # of *previous* dimension */
	
	if (prev_dim >= MAX_RANK) {
		if (prev_dim == MAX_RANK) {
			/*----------------------------------------------*/
			/* give errmsg & set 'rank' to (MAX_RANK + 1)	*/
			/*----------------------------------------------*/
			begin_errmsg ((name ? "array %q has " : ""),  name);
			end_errmsg ("too many dimensions - up to %d are allowed",
			             MAX_RANK);
			arp->attrs++;		/* to set 'rank' to (MAX_RANK + 1) */
		}
		/*------------------------------------------------------*/
		/* ...else errmsg already give, so do nothing and don't	*/
		/* change 'arp->attrs'.					*/
		/*------------------------------------------------------*/
		if (l) delete_expr (l, LEAVE_NLINKS);
		if (u) delete_expr (u, LEAVE_NLINKS);
		return;
	}
	/*}}}*/
	/*{{{  give errmsg if *previous* upper bound was '*'*/
	/*--------------------------------------------------------------*/
	/* Give an errmsg if the *previous* upper bound was an assumed	*/
	/* size declarator ('*')   (i.e. '*' not in the last dimension!)*/
	/*--------------------------------------------------------------*/
	if (shape_type & (ASSUMED_SIZE | MIXED_ASSUMED_SIZE)) {
		begin_errmsg ( (name ? "array %q: " : ""), name);
		end_errmsg  ("assumed size declarator (*) can only appear in the last dimension");
		shape_type = INVALID_POSN_ASSUMED_SIZE;
	}
	/*}}}*/

	/*{{{  if 'u' is given but not 'l', set l = 1.*/
	/*--------------------------------------------------------------*/
	/* If 'u' is given (by an expr or '*') but 'l' isn't, set 'l'	*/
	/* to 1.							*/
	/*--------------------------------------------------------------*/
	if (u  && (! l))
		l = int_lit_1;
	/*}}}*/
	
	/*{{{  check 'u', derive extent 'n', and install it in 'arp->n [prev_dim]'*/
	if (! u) {
		/*---------------------*/
		/* Upper bound omitted */
		/*---------------------*/
		if (shape_type & EXPL_SHAPE)
			mixed_shape_errmsg = TRUE;
	}
	else if (u->token == ND_STAR) {
		/*------------------------------------------------------*/
		/* Upper bound is '*'.  Install it unchanged in		*/
		/* 'arp->n [prev_dim]'.					*/
		/*------------------------------------------------------*/
		if ((shape_type & EXPL_SHAPE)  ||  (prev_dim == 0))
			shape_type = ASSUMED_SIZE;
	
		else if (shape_type & (ASSUMED_SHAPE | DEFERRED_SHAPE | MIXED_SHAPE)) {
			errmsg ("leading dimensions of assumed size (*) array %q must have explicit bounds",
				name);
			shape_type = MIXED_ASSUMED_SIZE;
		}
	}
	else {
		/*{{{  explicit upper bound.  Set u = MAX (u-l+1, 0)*/
		/*------------------------------------------------------*/
		/* Explicit upper bound.  Install MAX (u-l+1, 0) in	*/
		/* 'arp->n [prev_dim]'.					*/
		/* N.B. We know in this case that (l != NULL).		*/
		/*------------------------------------------------------*/
		Expr  *n;		/* the extent */
		bool  const_n;		/* TRUE if both 'l' and 'u' are const */
		
		protect_expr (l);	/* to prevent them being deleted when 'expr' */
		protect_expr (u);	/* is called from 'array_extent_func()'!     */
		
		n = array_extent_func (l, u);
		
		l->nlinks--;		/* don't *delete* 'l', as it'll be attached */
					/* to 'arp->lb [prev_dim]' later on! */
		unprotect_expr (u);	/* this deletes 'u' if it isn't used in 'n'! */
		
		u = n;
		const_n = (n->type & EX_INT_LIT_CONST);	/* both l & u const! */
		if (prev_dim == 0)
			shape_type = const_n ? CONST_SHAPE : VAR_SHAPE;
		
		else if ((shape_type == CONST_SHAPE) && !const_n)
			shape_type = VAR_SHAPE;
		
		else if (shape_type & (DEFERRED_SHAPE | ASSUMED_SHAPE))
			mixed_shape_errmsg = TRUE;
		
		if (exprs_ok)
			exprs_ok = mustbe_spec_expr (u, sp);
		/*}}}*/
	}
	
	if (mixed_shape_errmsg) {
		begin_errmsg ((name ? "array %q: " : ""), name);
		end_errmsg  ("cannot mix explicit and assumed or deferred array dimensions");
		shape_type = MIXED_SHAPE;
	}
	
	if (u) attach_expr (arp->n [prev_dim],  u);
	/*}}}*/

	/*{{{  check 'l' and install in 'arp->l [prev_dim]'*/
	if (l) {
		/*------------------------------------------------------*/
		/* Explicit lower bound  (or upper bound was explicit	*/
		/* or '*', in which case the lower bound was set to '1'	*/
		/* if it wasn't explcitly specified!)			*/
		/*							*/
		/* N.B. If a switch from CONST_SHAPE -> VAR_SHAPE was	*/
		/* necessary, it would have been done amongst the	*/
		/* checking for 'u'.					*/
		/*------------------------------------------------------*/
		if (shape_type == DEFERRED_SHAPE)
			shape_type = ASSUMED_SHAPE;
	
		if (exprs_ok)
			exprs_ok = mustbe_spec_expr (l, sp);
	
		attach_expr (arp->l [prev_dim], l);
	}
	/*}}}*/

	arp->attrs = (prev_dim + 1) + shape_type + (exprs_ok ? 0 : NOT_SPEC_EXPR);
}
/*}}}*/
/*{{{  install_array (sp, attrs_with_arp, arp) -- OK*/
install_array (sp, attrs_with_arp, arp)
/*-----------------------------------*/
	Symbol *sp;		  /* NULL if 'arp' is a DIMENSION attr
				                     in an 'attr-list' */
	int32   attrs_with_arp;	  /* 'attrs' declared in an 'attr-list'
				       along with (but before) 'arp' */
	Array  *arp;
/*----------------------------------------------------------------
	If (sp != NULL), then array spec 'arp' directly follows
an ID ('sp').  Check that it's consistent with 'sp's attributes
-- e.g. pointer and allocatable arrays must have deferred shape,
named consts must have constant shape,  HPF processor arrangements
& templates must have explicit shape, etc.  If any inconsistency
is found give an errmsg *unless* the attr that caused the error
is also in 'attrs_with_arp' (in which case we're in a typedecl or
HPF 'combined-attr', and the errmsg has been given already).
	Attach 'arp' to 'sp->arp' (unconditionally).

	If (sp == NULL) then 'arp' is specified by a DIMENSION
attribute in a typedecl or HPF 'combined-attr'.  Give an errmsg
if it isn't consistent with 'attrs_with_arp', the prior attributes
specified therein.
----------------------------------------------------------------*/
{
	char  *str_array;
	int    compatible_shape;

	if ((arp->attrs & RANK_MASK) == (MAX_RANK + 1))  /* i.e. too many dims*/
		arp->attrs--;
	compatible_shape = SHAPE_MASK;     /* the default, i.e. any shape OK! */
	if (sp) {
		/*{{{  << Discussion >>*/
		/*----------------------------------------------------------------
			Check that 'arp' is consistent with 'sp's attributes.
		
		N.B. 'sp->attrs' **cannot** have any of the following flags set, as
		'canbe_dimensioned (sp)' would have returned FALSE if they were:
		
			AT_NORMAL_VAR	-- since this implies that 'sp' was initialised or
					   referenced, and it must be dimensioned before this;
		
			(AT_FUNC | AT_SUBR | AT_MODULE | AT_PROGRAM | AT_BLOCK_DATA |
			 AT_DERIVED_TYPE | AT_NAMELIST_GROUP | AT_CONSTRUCT | AT_EXTERNAL |
			 AT_INTRINSIC)
					-- as they're not data objects.
		
		We ignore 'AT_INIT', as it's always set in conjunction with an 'entity'
		flag which is more specific and relevant for checking purposes.
		
		The following (groups of) flags can be associated with **any** array
		shape (except for ASSUMED_SIZE or ASSUMED_SHAPE, which must be a dummy
		argument):
		
			(AT_COMMON_NAME | AT_DUMMY | AT_RESULT_VAR | AT_INTENT |
			 AT_OPTIONAL | AT_ACCESS | AT_TARGET)
		----------------------------------------------------------------*/
		/*}}}*/
		int32  sp_prev_attrs = (sp->attrs & ~attrs_with_arp);
		                        /* ...those 'sp->attrs' *not* in  */
		                        /* an 'attr-list' along with 'arp'*/
		/*{{{  ALLOCATABLE / POINTER -> deferred shape*/
		if (sp_prev_attrs & (AT_ALLOC | AT_POINTER)) {
		
			compatible_shape = (DEFERRED_SHAPE | INVALID_SHAPE);
		
			str_array = (sp_prev_attrs & AT_ALLOC) ? "allocatable array" :
			                                         "array pointer";
		}
		/*}}}*/
		/*{{{  PROCESSORS / TEMPLATE -> explicit shape*/
		else if (sp_prev_attrs & (AT_PROCESSORS | AT_TEMPLATE)) {
		
			compatible_shape = (EXPL_SHAPE | INVALID_SHAPE);
		
			str_array = (sp_prev_attrs & AT_PROCESSORS) ? "processor array" :
			                                              "template array";
		}
		/*}}}*/
		/*{{{  PARAM / EQUIV -> constant shape*/
		else if (sp_prev_attrs & (AT_PARAM | AT_EQUIV)) {
		
			compatible_shape = (CONST_SHAPE | INVALID_SHAPE);
		
			str_array = (sp_prev_attrs & AT_PARAM) ? "named constant array" :
			                                         "equivalenced array";
		}
		/*}}}*/
		/*{{{  IN_COMMON / SAVE -> constant / deferred shape*/
		else if (sp_prev_attrs & (AT_IN_COMMON | AT_SAVE)) {
		
			compatible_shape = (CONST_SHAPE | DEFERRED_SHAPE | INVALID_SHAPE);
		
			str_array = (sp_prev_attrs & AT_IN_COMMON) ? "common block array" :
			                                             "SAVEd array";
		}
		/*}}}*/
		/*{{{  IN_NAMELIST -> constant shape*/
		else if (sp_prev_attrs & AT_IN_NAMELIST) {
		
			if (! (arp->attrs & (CONST_SHAPE | INVALID_SHAPE)) )
				errmsg ("only arrays with constant bounds can be in a namelist group: %q",
				         sp->name);
		}
		/*--------------------------------------------------------------*/
		/* ... leave (compatible_shape == SHAPE_MASK) so as not to	*/
		/* trigger another errmsg in the test later.			*/	
		/*--------------------------------------------------------------*/
		/*}}}*/
		/*{{{  EXTERNAL / INTRINSIC / FUNC -> *no* array-spec allowed!*/
		else if (sp_prev_attrs & (AT_INTRINSIC | AT_EXTERNAL | AT_FUNC)) {
		
			bool  intrin_msg = FALSE,  extern_msg = FALSE,  non_intrin_msg = FALSE;
		
			if (sp == this_unitp->sp) {
				/*------------------------------------------------------*/
				/* 'sp' is the host's function name (not result var!)   */
				/* (Incidentally, in this case 'prev_sp_attrs' has	*/
				/* AT_FUNC set, not (AT_INTRINSIC | AT_EXTERNAL).)	*/
				/*------------------------------------------------------*/
				Symbol  *result_sp;
		
				for (result_sp = sp->nextid;  ! (result_sp->attrs & AT_RESULT_VAR);
				     result_sp = result_sp->nextid)
					;
		
				errmsg ("array dimensions must be specified for the result variable %q, not the function name %q",
				         result_sp->name,  sp->name);
			}
			else if (sp_prev_attrs & AT_INTRINSIC)	/* 'sp' prev decl INTRINSIC */
				intrin_msg = TRUE;
		
			else if (sp_prev_attrs & AT_EXTERNAL)	/* 'sp' prev decl EXTERNAL */
				extern_msg = TRUE;
		
			else if ( ! (attrs_with_arp & (AT_INTRINSIC | AT_EXTERNAL)) ) {
				/*{{{  AT_FUNC is set, & not because of INTRINSIC / EXTERNAL attr*/
				/*----------------------------------------------------------------
					'sp_prev_attrs' has AT_FUNC set.
					The above test is necessary because, if 'attrs_with_arp'
				contains (AT_INTRINSIC | AT_EXTERNAL) and it was successfully set
				in 'sp', then it would also cause AT_FUNC to be set therein.  This
				would appear in 'sp_prev_attrs' as it isn't in 'attrs_with_arp'.
				We exclude this case since an errmsg about the DIMENSION attr was
				given when the 'attr-list' was processed.
					(Incidentally, this test also excludes the case where 'attr'
				(AT_INTRINSIC or AT_EXTERNAL)) *wasn't* installed in 'sp->attrs'
				because AT_FUNC was set by a prior reference or explicit
				interface, and 'attr' conflicts with the external or intrinsic
				identification made then.  In that case an errmsg would have been
				given for 'sp' about the EXTERNAL / INTRINSIC clash, so it doesn't
				matter if no errmsg is given about the DIMENSION error!)
				----------------------------------------------------------------*/
				
				if ((sp->attrs & AT_INFO) && (sp->info.sp->attrs & AT_INTRINSIC))
					/*----------------------------------------------*/
					/* 'sp' established as an intrinsic function.	*/
					/*----------------------------------------------*/
					intrin_msg = TRUE;
				
				else
					/*------------------------------------------------------*/
					/* 'sp' established as a non-intrinsic function.	*/
					/*							*/
					/* N.B.  The errmsg shouldn't assume that it's external	*/
					/* (e.g. by stating that there must be an explicit	*/
					/* interface), as it could be an internal or module	*/
					/* function!						*/
					/*------------------------------------------------------*/
					non_intrin_msg = TRUE;
				/*}}}*/
			}
		
			if (intrin_msg)
				errmsg ("array dimensions must not be specified for an intrinsic function: %q",
				         sp->name);
		
			else if (extern_msg || non_intrin_msg)
				errmsg ("array dimensions for %s function %q can only be specified in an explicit interface%s",
				         (extern_msg ? "external" : NULL_str),   sp->name,
				         (non_intrin_msg ? " or function definition" : NULL_str));
		}
		/*--------------------------------------------------------------*/
		/* ... leave (compatible_shape == SHAPE_MASK) so as not to	*/
		/* trigger another errmsg in the test later.			*/	
		/*--------------------------------------------------------------*/
		
		/*}}}*/
		/*{{{  assumed shape / size -> dummy argument*/
		else if (  (arp->attrs & (ASSUMED_SHAPE | ASSUMED_SIZE))  &&
		         ! (sp_prev_attrs & AT_DUMMY) )
		
			errmsg ("array %q cannot have assumed %s as it is not a dummy argument",
			         sp->name,  ((arp->attrs & ASSUMED_SHAPE) ? "shape" : "size"));
		
		/*--------------------------------------------------------------*/
		/* ... leave (compatible_shape == SHAPE_MASK) so as not to	*/
		/* trigger another errmsg in the test later.			*/	
		/*--------------------------------------------------------------*/
		/*}}}*/

		attach_array (sp->arp, arp);

		earlier_error = TRUE;     /* in case errmsg re #(mapped dims)*/
		if (sp->alignp)
			check_alignee (sp);
		else if (sp->distp)
			check_distributee (sp);
		earlier_error = FALSE;
	}
	else {	/*----------------------------------------------*/
		/* we're in the 'attr-list' of a 'type-decl' or	*/
		/* HPF 'combined-dir'.				*/
		/*----------------------------------------------*/
		/*{{{  ALLOCATABLE / POINTER -> deferred shape*/
		if (attrs_with_arp & (AT_ALLOC | AT_POINTER)) {
		
			compatible_shape = (DEFERRED_SHAPE | INVALID_SHAPE);
		
			str_array = (attrs_with_arp & AT_ALLOC) ? "ALLOCATABLE arrays" :
			                                          "array-valued POINTERs";
		}
		/*}}}*/
		/*{{{  PARAMETER -> constant shape*/
		else if (attrs_with_arp & AT_PARAM) {
		
			compatible_shape = (CONST_SHAPE | INVALID_SHAPE);
		
			str_array = "PARAMETER arrays";
		}
		/*}}}*/
		/*{{{  PROCESSORS / TEMPLATE -> explicit shape*/
		else if (attrs_with_arp & (AT_PROCESSORS | AT_TEMPLATE)) {
		
			compatible_shape = (EXPL_SHAPE | INVALID_SHAPE);
		
			str_array = (attrs_with_arp & AT_PROCESSORS) ? "PROCESSORS arrays" :
			                                               "TEMPLATE arrays";
		}
		/*}}}*/
		/*{{{  SAVEd arrays -> const / deferred shape*/
		else if (attrs_with_arp & AT_SAVE) {
		
			compatible_shape = (CONST_SHAPE | DEFERRED_SHAPE | INVALID_SHAPE);
		
			str_array = "SAVEd arrays";
		}
		/*}}}*/
		/*{{{  EXTERNAL / INTRINSIC -> *no* array-spec allowed*/
		else if (attrs_with_arp & AT_INTRINSIC)
		
			errmsg ("array dimensions must not be specified for INTRINSIC functions");
		
		else if (attrs_with_arp & AT_EXTERNAL)
		
			errmsg ("array dimensions for EXTERNAL functions can only be specified in an explicit interface");
		
		/*--------------------------------------------------------------*/
		/* ... in both cases leave (compatible_shape == SHAPE_MASK) so	*/
		/* as not to trigger another errmsg in the test later.		*/	
		/*--------------------------------------------------------------*/
		/*}}}*/
	}
	if ( ! (arp->attrs & compatible_shape))
		shape_errmsg (sp, str_array, compatible_shape);
}
/*}}}*/
/*{{{  delete_array (arp,  decr_nlinks) -- ok*/
delete_array (arp,  decr_nlinks)
/*----------------------------*/
	Array  *arp;
	bool    decr_nlinks;
/*--------------------------------------------------------------*/
/*	Called by 'clear_symtab ()' and 'yyparse ()' to delete	*/
/* an array record.						*/
/*								*/
/* N.B. It's assumed that (arp != NULL)!			*/
/*--------------------------------------------------------------*/
{
	Expr  **lpp,  **npp;
	int   dim,  rank = (arp->attrs & RANK_MASK);

	if (decr_nlinks)
		arp->nlinks--;

	if (arp->nlinks)
		/*----------------------------------------------*/	
		/* 'arp' is still in use, so don't delete it.	*/
		/*----------------------------------------------*/
		return;

	for (lpp = arp->l, npp = arp->n, dim = 0;  dim < rank;  dim++) {
		delete_expr (*lpp++, DECR_NLINKS);
		delete_expr (*npp++, DECR_NLINKS);
	}
	free ((char *) arp);
}
/*}}}*/

/*}}}*/

/*{{{  <<< 'char-len' handling functions >>>*/
/*-----------------------------*/
/* char-len handling functions */
/*-----------------------------*/
/*{{{  install_char_len (sp, char_len)*/
install_char_len (sp, char_len)
/*---------------------------*/
	Symbol  *sp;
	Expr    *char_len;

/*************
IN HERE, WE SHOULD ALSO CHECK ALL THE RULES FOR char-len, E.G.:
	-- THAT '*' IS ONLY USED WHERE ALLOWED;
	-- THAT char_len IS CONST IF NECESSARY
	-- THAT char-len IS A spec-expr (?? OR DO THIS IN hpf.y ??)
*************/
{
	attach_expr (sp->typeinfo.char_len, char_len);
}
/*}}}*/

/*}}}*/

/*{{{  <<< HPF data mapping functions >>>*/
/*----------------------------*/
/* HPF data mapping functions */
/*----------------------------*/
/*{{{  *** STILL TO DO -- data mapping *****/
/*------------------------------------
STILL TO DO -- data mapping
---------------------------
-- In interface block, 'process_decls()' not called, so final checks
& processing not done!

-- Insert flag IN_HPF_DIR to 'arp->attrs' if appropriate, in
'install_array ()', etc.  Then use this to determine if subsequent decls
are in accordance with it.  Also, in 'install_align ()', remove
uncomment this part of the test which precedes the call to
'check_align_target ()'.

-- In 'install_array()', ensure that 'arp->l []' or 'n []' are *never* NULL !!

-- When have all info, do suitable checks and give suitable errmsgs, e.g:
	-- if aligned outside bounds (alignp->al[] < 0, or l+n > arp->n)
	-- if src & tgt don't conform.
------------------------------------*/
/*}}}*/

/*{{{  <<< HPF ALIGN functions >>>*/
/*------------------------------*/
/* HPF ALIGN handling functions */
/*------------------------------*/
/*{{{  Align *create_align () -- OK*/
Align *create_align ()
/*------------------*/
/*------------------------------*/
/* Create an 'Align' record.	*/
/*------------------------------*/
{
	return  (Align *) ecalloc (sizeof (Align));
}
/*}}}*/
/*{{{  bool  canbe_aligned (sp) -- ok*/
bool  canbe_aligned (sp)
/*--------------------*/
	Symbol  *sp;
/*--------------------------------------------------------------*/
/*	Called by 'yyparse ()' to determine whether an ID ('sp')*/
/* can be aligned.  If the 'align-spec' directly follows the	*/
/* ID in an ALIGN directive, this is called prior to		*/
/* 'create_align ()', while in an HPF 'combined-directive' with	*/
/* an ALIGN attribute, it's called prior to 'attach_align ()'.	*/
/*--------------------------------------------------------------*/
{
	if (sp->alignp) {
		if (sp->alignp->tgt_sp)
			errmsg ("alignment of %q already specified", sp->name);
		else
			errmsg ("cannot ALIGN %q as it has INHERIT specified",
			         sp->name);
		return  FALSE;
	}
	else if (sp->distp) {
		errmsg ("cannot ALIGN %q as it is explicitly DISTRIBUTEd", sp->name);
		return  FALSE;
	}
	else if (sp->attrs & AT_UNTYPED_ENTITY) {
		/*{{{  give errmsg & return FALSE*/
		/*--------------------------------------------------------------*/
		/* A non-data entity can't be aligned.  Report an error, and do	*/
		/* one of two things:						*/
		/*								*/
		/* -- if the object type is one of those in AT_OVERRIDE_ATTRS   */
		/* (i.e. AT_CONSTRUCT), then clear this flag and return TRUE.	*/
		/* In other words, forget the object's old identity and give it	*/
		/* a new one!							*/
		/* -- otherwise return FALSE.					*/
		/*--------------------------------------------------------------*/
		switch (sp->attrs & AT_UNTYPED_ENTITY) {
		case  AT_SUBR :			begin_errmsg ("subroutine");		break;
		case  AT_MODULE :		begin_errmsg ("module");		break;
		case  AT_PROGRAM :		begin_errmsg ("main program");		break;
		case  AT_BLOCK_DATA :		begin_errmsg ("block data name");	break;
		case  AT_DERIVED_TYPE :		begin_errmsg ("derived type");		break;
		case  AT_NAMELIST_GROUP :	begin_errmsg ("namelist group");	break;
		case  AT_CONSTRUCT :		begin_errmsg ("construct name");	break;
		case  AT_PROCESSORS :		begin_errmsg ("processors arrangement");break;
		case  AT_TEMPLATE :		begin_errmsg ("template");		break;
		}
		end_errmsg (" %q cannot be aligned", sp->name);
		
		if (sp->attrs & AT_CONSTRUCT) {
			clear_flags (sp->attrs, AT_CONSTRUCT);
			return  TRUE;
		}
		else
			return  FALSE;
		/*}}}*/
	}
	else if (sp->attrs & (AT_FUNC | AT_INTRINSIC | AT_EXTERNAL)) {
		/*{{{  give errmsg & return FALSE*/
		/*--------------------------------------------------------------*/
		/* Only data *objects* can be aligned -- i.e. functions can't	*/
		/* be.  Report an error and return FALSE.			*/
		/*--------------------------------------------------------------*/
		
		if (sp->attrs & AT_INTRINSIC)
			/*------------------------------------------------------*/
			/* ...must be an intrinsic function, as AT_SUBR was	*/
			/* caught earlier.					*/
			/*------------------------------------------------------*/
			errmsg ("intrinsic function %q cannot be aligned", sp->name);
		
		else if (sp->attrs & AT_FUNC)
			errmsg ("can only specify alignment of function result %q in an explicit interface or function definition",
			         sp->name);
		
		else	/*-----------------------*/
			/* an EXTERNAL procedure */
			/*-----------------------*/
			errmsg ("external procedure %q cannot be aligned", sp->name);
		
		return  FALSE;
		/*}}}*/
	}

	return  TRUE;		/* OK! */
}
/*}}}*/
/*{{{  bool  canbe_inherited (sp) -- ok*/
bool  canbe_inherited (sp)
/*----------------------*/
	Symbol  *sp;
/*--------------------------------------------------------------*/
/*	Called by 'yyparse ()' to determine whether an ID ('sp')*/
/* can be given the INHERIT attribute.				*/
/*--------------------------------------------------------------*/
{
	if ( ! (sp->attrs & AT_DUMMY)) {
		errmsg ("INHERIT specified for non dummy argument %s", sp->name);
		return  FALSE;
	}
	else if (sp->alignp) {
		if (sp->alignp->tgt_sp)
			errmsg ("cannot INHERIT %q as it has ALIGN specified",
			         sp->name);
		else
			errmsg ("INHERIT already specified for %q", sp->name);
		return  FALSE;
	}
	else if (sp->attrs & AT_PROCEDURE) {
		errmsg ("cannot specify INHERIT for %s %q",
		          ((sp->attrs & AT_FUNC) ? "function" :
		          ((sp->attrs & AT_SUBR) ? "subroutine" :
		                                   "external procedure")),
		          sp->name);
		return  FALSE;
	}

	return  TRUE;		/* OK! */
}
/*}}}*/

/*{{{  install_align_source (src_ep, src_buf, alignp, sp) -- ok*/
install_align_source (src_ep, src_buf, alignp, sp)
/*----------------------------------------------*/
	Expr   *src_ep,			/* an 'align source' spec -- NULL,
					   ND_STAR or ND_SIMPLE_SCALAR */
	       *src_buf [MAX_RANK];	/* the buffer to copy it into */
	Align  *alignp;
	Symbol *sp;			/* may be NULL; only for errmsgs */
{
	char  *name = (sp ? sp->name : NULL);
	int    prev_rank = alignp->attrs & AL_SRC_RANK_MASK;
					/* i.e. rank *before* this dimension */

	alignp->attrs |= AL_EXPL_SRC_LIST;

	/*------------------------------------------------------*/	
	/* If (rank > MAX_RANK), give errmsg and ignore this	*/
	/* and subsequent dimensions.				*/
	/*------------------------------------------------------*/	
	if (prev_rank == MAX_RANK) {
		if ( ! (alignp->attrs & AL_TOO_MANY_SRC_DIMS)) {
			/*--------------------------------------*/
			/* give errmsg, and record that it's	*/
			/* been given by setting flag		*/
			/* 'AL_TOO_MANY_SRC_DIMS'.		*/
			/*--------------------------------------*/
			begin_errmsg ((name ? "%q" : "align source list"), name);
			end_errmsg (" has too many dimensions - up to %d are allowed",
			             MAX_RANK);
			alignp->attrs |= AL_TOO_MANY_SRC_DIMS;
		}
		/*--------------------------------------*/
		/* ...else errmsg already given.	*/
		/*--------------------------------------*/
		if (src_ep)
			delete_expr (src_ep, LEAVE_NLINKS);
		return;
	}

	alignp->attrs++;	/* to store the current 'source' rank */

	src_buf [prev_rank] = src_ep;		/* may be NULL */

	return;		
}
/*}}}*/
/*{{{  check_align_source_list (src_buf, alignp) -- ok*/
check_align_source_list (src_buf, alignp)
/*-------------------------------------*/
	Expr   *src_buf [MAX_RANK];
	Align  *alignp;
/*--------------------------------------------------------------*/
/*	Check that align dummies aren't repeated in the source	*/
/* list.  If they are, give an errmsg and replace repeats by	*/
/* '*' (i.e. make them collapsed dimensions).			*/
/*								*/
/* N.B.  We do this before the 'WITH' clause is parsed, rather	*/
/* than afterwards in 'install_align ()', so that any errmsgs	*/
/* relating solely to the source list are produced *before* any	*/
/* errmsgs about exprs in the align target list.		*/
/*--------------------------------------------------------------*/
{
	int     src_rank = alignp->attrs & AL_SRC_RANK_MASK;
	bool    repeats;
	int     i,  j;
	Expr   *ep,  *src_ep;
	Symbol *dummy_sp;

	for (i = 0;   i < src_rank;   i++) {
		src_ep = src_buf [i];
		if ( ! src_ep  ||  (src_ep->token != ND_SIMPLE_SCALAR))
			continue;

		dummy_sp = src_ep->val.sp;	/* an align dummy variable */

		/*----------------------------------------------*/
		/* 'dummy_sp' is an align dummy variable.  If	*/
		/* it's repeated in the source list give an	*/
		/* errmsg and replace the repeats by '*'.	*/
		/*----------------------------------------------*/
		repeats = FALSE;
		for (j = i+1;   j < src_rank;   j++) {
			ep = src_buf [j];
			if (ep  &&  (ep->token == ND_SIMPLE_SCALAR)  &&
			            (ep->val.sp == dummy_sp)) {
				repeats = TRUE;
				delete_expr (ep, LEAVE_NLINKS);
				src_buf [j] = star_primary;
			}
		}
		if (repeats)
			errmsg ("dummy variable %q is repeated in the ALIGN source list",
			         dummy_sp->name);
	}
}
/*}}}*/
/*{{{  install_align_subscr (subscr_ep, tgt_buf, alignp) -- ok*/
install_align_subscr (subscr_ep, tgt_buf, alignp)
/*---------------------------------------------*/
	Expr   *subscr_ep,		/* an 'align subscript' spec (!= NULL).
					   It may be an ordinary expr, ND_STAR
					   or a ND_SUBSCR_DOUBLET /_TRIPLET */
	       *tgt_buf [MAX_RANK];	/* the buffer to copy it into */
	Align  *alignp;
{
	int    prev_rank = (alignp->attrs & AL_TGT_RANK_MASK) >> AL_TGT_RANK_SHIFT;
					/* i.e. rank *before* this dimension */

	alignp->attrs |= AL_EXPL_TGT_LIST;

	/*------------------------------------------------------*/	
	/* If (rank > MAX_RANK), give errmsg and ignore this	*/
	/* and subsequent dimensions.				*/
	/*------------------------------------------------------*/	
	if (prev_rank == MAX_RANK) {
		if ( ! (alignp->attrs & AL_TOO_MANY_TGT_DIMS)) {
			/*--------------------------------------*/
			/* give errmsg, and record that it's	*/
			/* been given by setting flag		*/
			/* 'AL_TOO_MANY_TGT_DIMS'.		*/
			/*--------------------------------------*/
			errmsg ("%q has too many dimensions - up to %d are allowed",
			         alignp->tgt_sp->name,  MAX_RANK);
			alignp->attrs |= AL_TOO_MANY_TGT_DIMS;
		}
		/*--------------------------------------*/
		/* ...else errmsg already given.	*/
		/*--------------------------------------*/
		if (subscr_ep)
			delete_expr (subscr_ep, LEAVE_NLINKS);
		return;
	}

	alignp->attrs += (1 << AL_TGT_RANK_SHIFT);	/* to store the current
							   'target' rank */
	tgt_buf [prev_rank] = subscr_ep;

	return;		
}
/*}}}*/

/*{{{  install_align (sp, alignp, src_buf, tgt_buf) -- ok*/
install_align (sp, alignp, src_buf, tgt_buf)
/*----------------------------------------*/
	Symbol *sp;		  /* NULL for ALIGN attr in an 'attr-list' */
	Align  *alignp;
	Expr   *src_buf [MAX_RANK],  *tgt_buf [MAX_RANK];

/*{{{  << DESCRIPTION -- rewrite it! >>*/
/*----------------------------------------------------------------
	Called by 'yyparse ()' after an ALIGN directive or attr
has been parsed.  ('decl_sp' is defined in the former case but
is NULL in the latter).





*** REWRITE THIS ***
	If (sp != NULL), then array spec 'arp' directly follows
an ID ('sp').  Check that it's consistent with 'sp's attributes
-- e.g. pointer and allocatable arrays must have deferred shape,
named consts must have constant shape,  HPF processor arrangements
& templates must have explicit shape, etc.  If any inconsistency
is found give an errmsg *unless* the attr that caused the error
is also in 'attrs_with_arp' (in which case we're in a typedecl or
HPF 'combined-attr', and the errmsg will have been given already).
	Attach 'arp' to 'sp->arp' (unconditionally).

	If (sp == NULL) then 'arp' is specified by a DIMENSION
attribute in a typedecl or HPF 'combined-attr'.  Give an errmsg
if it isn't consistent with 'attrs_with_arp', the prior attributes
specified therein.
----------------------------------------------------------------*/
/*}}}*/
{
	/*{{{  decls*/
	bool     error;
	
	/**** int      i;	**** ONLY FOR TESTING -- DELETE LATER ***/
	
	int      actual_src_rank,  actual_tgt_rank,  embed_rank,  replic_rank,
	         src_rank,  tgt_dim,  tgt_rank,  triplet_dim;
	short   *dim_p,  *ip,  *max_dim_p,  short_buf [MAX_RANK];
	Expr    *ep,  **epp,  **lpp,  **max_src_epp,  **max_tgt_epp,
	       **prev_err_epp, **spp, *src_ep, **src_epp, *tgt_ep, **tgt_epp;
	Symbol  *dummy_sp,
	        *tgt_sp = alignp->tgt_sp;	/* NULL for INHERIT; not otherwise */
	char    *conform_msg = "aligned dimensions of alignee and align target do not conform";
	
	/*}}}*/

	if ( ! alignp->tgt_sp) {
		/*{{{  INHERIT:  install if 'sp' defined, and return*/
		/*--------------*/
		/* INHERIT	*/
		/*--------------*/
		if (sp) {
			attach_align (sp->alignp, alignp);
			if (sp->arp)
				check_alignee (sp);
		}
		return;
		/*}}}*/
	}
	/*{{{  give errmsg & return if 'sp' is self-aligned!*/
	if (sp  &&  (sp == tgt_sp)) {
		errmsg ("%q is aligned with itself", sp->name);
		return;
	}
	/*}}}*/
	/*{{{  give errmsg if descriptive ALIGN specified for non-dummy 'sp'*/
	if (sp  &&  (alignp->attrs & AL_ASSERT)  &&  ! (sp->attrs & AT_DUMMY))
	
		errmsg ("descriptive (i.e. \'*\') ALIGN specified for non dummy argument %q",
		         sp->name);
	/*}}}*/
	/*{{{  find 'src_ / tgt_rank', and clear any 'AL_TOO_MANY_DIMS' flags*/
	/*----------------------------------------------------------------*/
	/* Find 'src_rank' & 'tgt_rank', and clear any 'AL_TOO_MANY_DIMS' */
	/* flags that may be set.					  */
	/*----------------------------------------------------------------*/
	src_rank = alignp->attrs & AL_SRC_RANK_MASK;
	tgt_rank = (alignp->attrs & AL_TGT_RANK_MASK) >> AL_TGT_RANK_SHIFT;
	
	clear_flags (alignp->attrs, (AL_TOO_MANY_SRC_DIMS | AL_TOO_MANY_TGT_DIMS));
	
	max_src_epp = src_buf + src_rank;	/* ptr to cell *after* valid ones */
	max_tgt_epp = tgt_buf + tgt_rank;	/*    "    "      "      "     "  */
	/*}}}*/
	/*{{{  find 'actual_src_ / tgt_rank' if known*/
	/*----------------------------------------------------------------
		If the alignee 'sp' is given and is already dimensioned
	(i.e. (sp  &&  sp->arp)), then set 'actual_src_rank' to #(alignee dims).
	Likewise, if the dimensions of the align target 'tgt_sp' are known
	(i.e. (tgt_sp->arp != NULL) -- we know 'tgt_sp' is given), then set
	'actual_tgt_rank' to #(target dims).
		They're used for the following purposes:
	
	--  To ensure that we don't look-up a lower bound 'sp / tgt_sp->arp->l [dim]'
	of a non-existent dimension, in case the alignment erroneously
	specifies too many dimensions.  [However, we don't yet report or
	correct this error.  That's done by 'check_alignee ()' or
	check_align_target ()' after 'alignp' has been set up.]
	
	--  If the *target* list is omitted from ALIGN directive / attr
	(i.e. it has the form:
	
		!HPF$ ALIGN [s] [(...)] WITH t  [:: ...] )
	
	it's taken to be a list of ':' entries of length 'actual_tgt_rank',
	and 'tgt_buf []' and 'alignp->attrs' are set-up accordingly.
		[Incidentally, we can't do a similar thing for the
	source list, since if it's omitted then this must be an ALIGN *attr*
	(not directive), and 'sp' is undefined!]
	----------------------------------------------------------------*/
	if (sp  &&  sp->arp)
		actual_src_rank = (sp->arp->attrs & RANK_MASK);
	
	if (tgt_sp->arp) {
		actual_tgt_rank = (tgt_sp->arp->attrs & RANK_MASK);
		if ( ! tgt_rank) {
			/*----------------------------------------------*/
			/* Target list is omitted from ALIGN directive,	*/
			/* which is equivalent to a list of ':' entries	*/
			/* of length 'actual_tgt_rank'.  Set-up		*/
			/* 'tgt_buf []' and 'alignp->attrs' accordingly.*/
			/*----------------------------------------------*/
			alignp->attrs += ((tgt_rank = actual_tgt_rank) << AL_TGT_RANK_SHIFT);
			max_tgt_epp = tgt_buf + tgt_rank;
	
			for (tgt_epp = tgt_buf;   tgt_epp < max_tgt_epp;   tgt_epp++)
				*tgt_epp = subscr_triplet (NULL_Expr, NULL_Expr, NULL_Expr);
								/* i.e. ':' */
		}
	}
	/*}}}*/
	/*{{{  if only one of 'src/tgt list' is known, deduce the other*/
	/*----------------------------------------------------------------
		If one of the alignee or target lists is omitted from the
	ALIGN directive and cannot be deduced because its dimensions
	aren't declared yet, but the other *is* is given or was deduced,
	then *assume* that the missing list is a list of N ':'s, where
	N is the number of ':'s in the given list (as it must be unless
	there's an error).  Set-up 'src/tgt_buf []' and 'alignp->attrs'
	accordingly.
	
	N.B. 'src/tgt_rank' could be assumed to be 0 (if the other list
	has no ':' entries).
	----------------------------------------------------------------*/
	if (src_rank  &&  ! tgt_rank) {
		/*------------------------------------------------------*/
		/*	The target list is omitted from the ALIGN	*/
		/* directive but the source list is given or was	*/
		/* deduced, so assume that the former is a list of N	*/
		/* ':'s, where N is the number of ':'s in the latter.	*/
		/* Set 'tgt_buf []' and 'alignp->attrs' accordingly.	*/
		/*------------------------------------------------------*/
		for (src_epp = src_buf;   src_epp < max_src_epp;   src_epp++)
			if ( ! *src_epp)
				/*------------------------------*/
				/* a ':' in the source list	*/
				/*------------------------------*/
				tgt_rank++;
	
		alignp->attrs += (tgt_rank << AL_TGT_RANK_SHIFT);
		max_tgt_epp = tgt_buf + tgt_rank;
	
		for (tgt_epp = tgt_buf;   tgt_epp < max_tgt_epp;   tgt_epp++)
			*tgt_epp = subscr_triplet (NULL_Expr, NULL_Expr, NULL_Expr);
							/* i.e. a ':' entry */
		alignp->attrs |= AL_DEDUCED_TGT_LIST;
	}
	else if (tgt_rank  &&  ! src_rank) {
		/*----------------------------------------------*/
		/*	As above with 'source' <-> 'target'.	*/
		/*----------------------------------------------*/
		for (tgt_epp = tgt_buf;   tgt_epp < max_tgt_epp;   tgt_epp++)
			if ( ((*tgt_epp)->token == ND_SUBSCR_DOUBLET)  ||
			     ((*tgt_epp)->token == ND_SUBSCR_TRIPLET) )
				/*--------------------------------------*/
				/* a '([l]:[u][:s])' in the target list	*/
				/*--------------------------------------*/
				src_rank++;
	
		alignp->attrs += src_rank;
		max_src_epp = src_buf + src_rank;
	
		for (src_epp = src_buf;   src_epp < max_src_epp;   src_epp++)
			*src_epp = NULL_Expr;		/* i.e. ':' */
	
		alignp->attrs |= AL_DEDUCED_SRC_LIST;
	}
	/*}}}*/
	/*{{{  allocate arrays 'alignp->al []', 'as []' & 'd []'*/
	/*------------------------------------------------------*/
	/* Allocate arrays 'alignp->al []', 'as []' and 'd []'.	*/
	/*------------------------------------------------------*/
	if (src_rank) {
		alignp->al = (Expr **) emalloc (src_rank * sizeof (Expr *));
		alignp->as = (Expr **) emalloc (src_rank * sizeof (Expr *));
		alignp->d  = (short *) emalloc (src_rank * sizeof (short));
	}
	max_dim_p = alignp->d + src_rank;
	/*}}}*/

	/*{{{  check align dummies & set their 'actual_arg' to point to tgt expr*/
	/*--------------------------------------------------------------*/
	/*	Check that each align dummy variable 'dummy_sp' is used	*/
	/* OK, and set 'dummy_sp->actual_arg' to point to the		*/
	/* corresponding target expr.					*/
	/*--------------------------------------------------------------*/
	/*----------------------------------------------------------------
		For each 'dummy_sp', set 'dummy_sp->actual_arg' to point
	to the (first) expr in the align target list in which it appears,
	and store its align target dimension in 'alignp->d [].  If it
	doesn't occur in the align target list, this dimension of the
	alignee is collapsed so replace it by '*'.  If it's repeated in
	the target list give an errmsg and treat it as a normal variable
	in subsequent dims.
	
	N.B.  'check_align_source_list()' has already checked that align
	dummies aren't repeated in the *source* list, and replaced any
	such repeats by '*'s.
	----------------------------------------------------------------*/
	for (src_epp = src_buf,  dim_p = alignp->d;   src_epp < max_src_epp;
	     src_epp++,  dim_p++) {
		src_ep = *src_epp;
		if ( ! src_ep  ||  (src_ep->token != ND_SIMPLE_SCALAR))
			continue;
	
		dummy_sp = src_ep->val.sp;	/* an align dummy variable */
	
		/*{{{  set 'dummy_sp->actual_arg' to point to corresp tgt expr*/
		/*--------------------------------------------------------------*/
		/* If there are no occurrences of 'dummy_sp' in the align	*/
		/* target, this dimension of the align source is collapsed so	*/
		/* replace it by '*'.  Otherwise, find the first expr in the	*/
		/* align target list in which 'dummy_sp' occurs, set		*/
		/* 'dummy_sp->actual_arg' to point to it, and store its dim	*/
		/* in 'alignp->d [i].  Give an errmsg if 'dummy_sp' occurs in	*/
		/* >1 target dimensions.					*/
		/*--------------------------------------------------------------*/
		for (tgt_epp = tgt_buf;   tgt_epp < max_tgt_epp;   tgt_epp++) {
			tgt_ep = *tgt_epp;
		
			if ( tgt_ep->token == ND_STAR  ||
			     tgt_ep->token == ND_SUBSCR_DOUBLET  ||
			     tgt_ep->token == ND_SUBSCR_TRIPLET ||
			   ! occurs_in_expr (dummy_sp, tgt_ep))
				continue;
		
			/*----------------------------------------------*/
			/* 'dummy_sp' occurs in this target dimension	*/
			/*----------------------------------------------*/
			if ( ! dummy_sp->actual_arg) {
				dummy_sp->actual_arg = tgt_ep;	/* pointer to tgt expr */
				*dim_p = tgt_epp - tgt_buf;	/* i.e. tgt dimension */
			}
			else {
				errmsg ("align dummy variable %q appears in more than one dimension of the align target %q",
				         dummy_sp->name,  tgt_sp->name);
		
				break;			/* quit loop -- needn't check
							   for more repeats! */
			}
		}
		
		if ( ! dummy_sp->actual_arg) {
			/*------------------------------------------------------*/
			/* 'dummy_sp' doesn't appear in the align target -- i.e.*/
			/* it's a collapsed dim!  Replace it by '*' in alignee.	*/
			/*------------------------------------------------------*/
			delete_expr (src_ep, LEAVE_NLINKS);
			*src_epp = star_primary;
		}
		/*}}}*/
	}
	
	/*----------------------------------------------------------------
		Now for each align dummy, check that its target expr
	doesn't involve other align dummies (i.e. that there aren't other
	align dummies 'sp' with the same 'sp->actual_arg').  If it does,
	give an errmsg and replace all but the first of these align dummies
	in the align source list by '*'.
	
	N.B. This only checks for *1st* appearances of align dummies in
	the target list, so an error is only reported if >1 align dummy
	has its *1st* appearance in the same target expr.  Appearance
	of an align dummy in more than one target expr has already been
	reported as an error above, so it's OK to ignore 2nd and subsequent
	appearances now.
	----------------------------------------------------------------*/
	for (src_epp = src_buf;   src_epp < max_src_epp;   src_epp++) {
		src_ep = *src_epp;
		if ( ! src_ep  ||  (src_ep->token != ND_SIMPLE_SCALAR))
			continue;
	
		tgt_ep = src_ep->val.sp->actual_arg;	/* the corresponding tgt expr */
	
		/*{{{  check that other align dummies don't have the same 'tgt_ep'*/
		/*--------------------------------------------------------------*/
		/* Check that align dummies in alignee dimensions j > i don't	*/
		/* have target expr 'tgt_ep' (which is the target expr for	*/
		/* alignee dim i).						*/
		/*--------------------------------------------------------------*/
		prev_err_epp = NULL_Expr_p;
		for (epp = src_epp + 1;   epp < max_src_epp;   epp++) {
			ep = *epp;
			if ( ! ep  ||  (ep->token != ND_SIMPLE_SCALAR)  ||
			        (ep->val.sp->actual_arg != tgt_ep))
				continue;
		
			/*------------------------------------------------------*/
			/* Align dummy from this alignee dim also appears in	*/
			/* 'tgt_ep' (along with that from alignee dim 'i').	*/
			/* Give errmsg.						*/
			/*------------------------------------------------------*/
			if ( ! prev_err_epp)
				begin_errmsg ("dimension %d of align target %q involves more than one align dummy variable, namely %q",
				               alignp->d [src_epp - src_buf] + 1,
				               tgt_sp->name,
				               src_ep->val.sp->name);
			else {
				/*----------------------------------------------*/
				/* replace align dummy 'prev_err_ep->val.sp' by	*/
				/* '*' in the alignee				*/
				/*----------------------------------------------*/
				cont_errmsg (", %q",  (*prev_err_epp)->val.sp->name);
				delete_expr (*prev_err_epp, LEAVE_NLINKS);
				*prev_err_epp = star_primary;
			}
			prev_err_epp  = epp;
		}
		if (prev_err_epp) {
			/*----------------------------------------------*/
			/* replace align dummy 'prev_err_ep->val.sp' by	*/
			/* '*' in the alignee				*/
			/*----------------------------------------------*/
			end_errmsg (" and %q",  (*prev_err_epp)->val.sp->name);
			delete_expr (*prev_err_epp, LEAVE_NLINKS);
			*prev_err_epp = star_primary;
		}
		/*}}}*/
	}
	/*}}}*/

	for (src_epp = src_buf,  lpp = alignp->al,  spp = alignp->as,
	           dim_p = alignp->d,  triplet_dim = -1;
	     src_epp < max_src_epp;    src_epp++,  lpp++,  spp++,  dim_p++) {
		src_ep = *src_epp;
		if (! src_ep) {
			/*{{{  ':' in alignee;  match with triplet in target*/
			/*--------------------------------------------------------------*/
			/* A ':' in the alignee dimension.  Look for the next subscript	*/
			/* triplet in the align target.					*/
			/*--------------------------------------------------------------*/
			for (triplet_dim++;   triplet_dim < tgt_rank;   triplet_dim++) {
				tgt_ep = tgt_buf [triplet_dim];
				if ((tgt_ep->token == ND_SUBSCR_DOUBLET)  ||
				    (tgt_ep->token == ND_SUBSCR_TRIPLET))
					break;
			}
			
			if (triplet_dim >= tgt_rank) {
				if (triplet_dim == tgt_rank)	/* give errmsg for 1st unmatched ':' */
					errmsg (conform_msg);
				/*------------------------------------------------------*/
				/* corrective action -- treat it as a collapsed dim!	*/
				/*------------------------------------------------------*/
				attach_expr (*lpp, star_primary);
				attach_expr (*spp, star_primary);
				*dim_p = -1;
				continue;
			}
			
			/*-------------------------*/
			/* install 'alignp->as []' */
			/*-------------------------*/
			if (tgt_ep->token == ND_SUBSCR_TRIPLET) {
				attach_expr (*spp, tgt_ep->right);		/* explicit stride */
				tgt_ep = tgt_ep->left;		/* the subscript doublet ([l]:[u]) */
			}
			else
				attach_expr (*spp, int_lit_1);			/* default stride */
			
			/*-------------------------*/
			/* install 'alignp->al []' */
			/*-------------------------*/
			if (tgt_ep->left) {			/* explicit lower bound	*/
				if (tgt_sp->arp  &&  (triplet_dim < actual_tgt_rank))
					attach_expr (*lpp, expr (tgt_ep->left,  ND_MINUS_OP,
					                         tgt_sp->arp->l [triplet_dim]));
				else {
					attach_expr (*lpp, tgt_ep->left);
					alignp->attrs |= (AL_NEED_TGT_LB << triplet_dim);
				}
			}
			else				/* default lower bound	*/
				attach_expr (*lpp, int_lit_0);
			
			/*------------------------*/
			/* install 'alignp->d []' */
			/*------------------------*/
			*dim_p = triplet_dim;
			
			/*}}}*/
		}
		else if (src_ep->token == ND_STAR) {
			/*{{{  '*' in alignee;  a collapsed dimension*/
			/*--------------------------------------------------------------*/
			/* A '*' in the alignee dimension -- i.e. a collapsed dimension.*/
			/*--------------------------------------------------------------*/
			attach_expr (*lpp, star_primary);
			attach_expr (*spp, star_primary);
			*dim_p = -1;
			/*}}}*/
		}
		else {
			/*{{{  align dummy var;  match with expr in target*/
			/*{{{  << EXPLANATION >>*/
			/*----------------------------------------------------------------
				(src_ep->token == ND_SIMPLE_SCALAR), i.e. this alignee
			dimension ('src_dim') contains an align dummy variable.
				Say the dummy variable is 'i' and the corresponding target
			expr (in dimension 'tgt_dim' of the align target) has the form:
			
				tgt_e = e1 + (i * e2)
			
			If 'La' is the lower bound of dimension 'src_dim' of the alignee
			'sp', and 'Lt' is the lower bound of dimension 'tgt_dim' of the
			align target, then store the following exprs in 'alignp':
			
				alignp->al [src_dim] = e1 + (La * e2) - Lt
								-- the offset from lb (align tgt)
								   of the target element aligned
								   with lb (alignee dim);
				alignp->as [src_dim] = e2	-- the stride
			N.B:
			----
			(1) A pointer to the target expr ('tgt_e') is stored in:
				src_ep->val.sp->actual_arg
			
			and its dimension in the align target is stored in:
				alignp->d [src_dim].
			
			(2) If there's no alignee (! sp), or it's shape hasn't been
			declared yet (! sp->arp), then 'La' is unknown.  In that case
			evaluate 'alignp->al [src_dim]' assuming that (La == 0) for the time
			being, and set the flag (AL_NEED_SRC_LB << src_dim) in 'alignp->attrs'
			to signal this.
				When the actual value for 'La' becomes known, function
			'check_alignee()' will evaluate and install the correct value
			of 'alignp->al [src_dim]', namely:
			
				alignp->al [src_dim] += (La * alignp->as [src_dim])
			----------------------------------------------------------------*/
			/*}}}*/
			
			int      src_dim = (src_epp - src_buf),  tgt_dim = *dim_p;
			
			dummy_sp = src_ep->val.sp;	/* the align dummy variable */
			
			extract_l_and_s (
				dummy_sp->actual_arg,		/* target expr */
				dummy_sp,			/* align dummy var */
			
				((sp  &&  sp->arp  &&  (src_dim < actual_src_rank)) ?
				      sp->arp->l [src_dim] : int_lit_0),	/* source lower bound */
			
				((tgt_sp->arp  &&  (tgt_dim < actual_tgt_rank)) ?
				      tgt_sp->arp->l [tgt_dim] : int_lit_0),	/* target lower bound */
			
				lpp,				/* ptr to result 1 (alignp->al [src_dim]) */
				spp,				/*  "   "   "    2 (alignp->as [src_dim]) */
				tgt_sp,				/* for errmsgs */
				tgt_dim				/* for errmsgs */
			);
			
			
			if ((*lpp)->token == ND_STAR) {
				/*------------------------------------------------------*/
				/* Align target expr is erroneous, so the source dim	*/
				/* is collapsed as a simple 'fix'.  In this case we	*/
				/* also replace the target expr by '*', i.e. replicate	*/
				/* the corresponding target dimension.			*/
				/*------------------------------------------------------*/
				*dim_p = -1;		/* to match (l,s) = ('*','*')	*/
				delete_expr (dummy_sp->actual_arg, LEAVE_NLINKS);
				tgt_buf [tgt_dim] = star_primary;
			}
			else {
				if ( ! (sp  &&  sp->arp))
					alignp->attrs |= (AL_NEED_SRC_LB << src_dim);
			
				if ( ! tgt_sp->arp)
					alignp->attrs |= (AL_NEED_TGT_LB << tgt_dim);
			}
			
			dummy_sp->actual_arg = NULL_Expr;
			/*}}}*/
		}
	}

	/*{{{  store #s of embedded & replicated dims in 'short_buf[]'*/
	/*--------------------------------------------------------------*/
	/* Find embedded and replicated dims, and store their dimension	*/
	/* numbers in 'short_buf []'.					*/
	/*--------------------------------------------------------------*/
	error = FALSE;
	
	for (tgt_dim = embed_rank = replic_rank = 0;
	     tgt_dim < tgt_rank;
	     tgt_dim++) {
		/*--------------------------------------*/
		/* Is 'tgt_dim' in the 'align' set?	*/
		/*--------------------------------------*/
		for (dim_p = alignp->d;   (dim_p < max_dim_p) && (*dim_p != tgt_dim);
		     dim_p++)
			;
		if (dim_p < max_dim_p)
			continue;	/* 'tgt_dim' included in 'align' set */
	
		/*------------------------------------------------------*/
		/* 'tgt_dim' isn't in the 'align' set, so it must be an	*/
		/* embedded, replicated or 'erroneous' dim.		*/
		/*------------------------------------------------------*/
		switch (tgt_buf [tgt_dim]->token) {
		case  ND_SUBSCR_DOUBLET :
		case  ND_SUBSCR_TRIPLET :
			if (! error) {
				errmsg (conform_msg);
				error = TRUE;
			}
			/*----------------------------------------------*/
			/* ...now fall through to next case -- i.e.	*/
			/* make it a replicated dim!			*/
			/*----------------------------------------------*/
		case  ND_STAR :
			short_buf [MAX_RANK - ++replic_rank] = tgt_dim;
			break;
	
		default:	/* i.e. a scalar expr */
			short_buf [embed_rank++] = tgt_dim;
			break;
		}
	}
	/*}}}*/

	/*{{{  store 'embed_rank' & 'replic_rank' in 'alignp->attrs'*/
	/*--------------------------------------------------------------*/
	/* Store 'embed_rank' & 'replic_rank' in 'alignp->attrs'	*/
	/*--------------------------------------------------------------*/
	alignp->attrs += (embed_rank << AL_EMBED_RANK_SHIFT) +
	                 (replic_rank << AL_REPLIC_RANK_SHIFT);
	/*}}}*/

	if (embed_rank) {
		/*{{{  allocate and set 'alignp->ei []' and 'ed []'*/
		/*--------------------------------------------------------------*/
		/* Allocate and set arrays 'alignp->ei []' and 'ed []' for	*/
		/* embedded dimensions:						*/
		/*								*/
		/*	alignp->ei []   = subscript-expr [tgt-dim] - Lt		*/
		/*	alignp->ed [] = tgt_dim					*/
		/*								*/
		/* where 'Lt' is the lower bound of dimension 'tgt_dim' of the	*/
		/* align target.						*/
		/*--------------------------------------------------------------*/
		alignp->ei = (Expr **) emalloc (embed_rank * sizeof (Expr *));
		alignp->ed = (short *) emalloc (embed_rank * sizeof (short));
		
		for (ip = short_buf,  lpp = alignp->ei,  dim_p = alignp->ed;
		     ip < short_buf + embed_rank;
		     ip++,  lpp++,  dim_p++) {
		
			*dim_p = tgt_dim = *ip;
		
			if (tgt_sp->arp  &&  (tgt_dim < actual_tgt_rank))
				attach_expr (*lpp,  expr (tgt_buf [tgt_dim],  ND_MINUS_OP,
				                          tgt_sp->arp->l [tgt_dim]));
			else {
				attach_expr (*lpp,  tgt_buf [tgt_dim]);
				alignp->attrs |= (AL_NEED_TGT_LB << tgt_dim);
			}
		}
		/*}}}*/
	}
	if (replic_rank) {
		/*{{{  allocate and set 'alignp->rd []'*/
		/*--------------------------------------------------------------*/
		/* Allocate and set array 'alignp->rd []' for replicated	*/
		/* dimensions:							*/
		/*	alignp->rd [] = tgt_dim					*/
		/*--------------------------------------------------------------*/
		short  *end_short_buf = (short_buf + MAX_RANK - 1);
							/* last cell in 'short_buf []' */
		
		alignp->rd = (short *) emalloc (replic_rank * sizeof (short));
		
		for (ip = end_short_buf,  dim_p = alignp->rd;
		     ip > (end_short_buf - replic_rank);
		     ip--,  dim_p++)
		
			*dim_p = *ip;
		/*}}}*/
	}
	/*{{{  delete any unused exprs in 'src_buf []' and 'tgt_buf []'*/
	/*--------------------------------------------------------------*/
	/* Delete any unused exprs (or parts thereof) in 'src_buf []'	*/
	/* and 'tgt_buf []'.						*/
	/*--------------------------------------------------------------*/
	for (src_epp = src_buf;   src_epp < max_src_epp;   src_epp++)
		if (*src_epp)
			delete_expr (*src_epp, LEAVE_NLINKS);
	
	for (tgt_epp = tgt_buf;   tgt_epp < max_tgt_epp;   tgt_epp++)
		if (*tgt_epp)
			delete_expr (*tgt_epp, LEAVE_NLINKS);
	/*}}}*/
	if (sp) {
		attach_align (sp->alignp, alignp);
		if (sp->arp)
			check_alignee (sp);
	}
	/*{{{  call 'check_align_target ()' if we have enough info to do checks*/
	/*--------------------------------------------------------------*/
	/* Call 'check_align_target ()' if we already have enough info	*/
	/* to do the checks.						*/
	/*								*/
	/* N.B.								*/
	/* ----								*/
	/* -- If the 1st part of the test below is TRUE, then 'tgt_sp'	*/
	/* is a dimensioned data object, template or processor array,	*/
	/* so its entity type is known and, if OK, we can check its	*/
	/* dims.  If the 2nd test is passed then 'tgt_sp' *isn't* a	*/
	/* data object or template, so we can give an errmsg. 		*/
	/* -- If 'sp' is defined, reference 'sp->alignp' rather than	*/
	/* 'alignp' in case 'check_alignee ()' has modified it.		*/
	/*--------------------------------------------------------------*/
	
	if (  (tgt_sp->arp  &&  ( (tgt_sp->attrs & (AT_ENTITY | AT_PROCEDURE))  ||
	                          (tgt_sp->type & TY_TYPE) /****** ||
	                       ! (tgt_sp->arp & IN_HPF_DIR) ******/ ))
	   || (tgt_sp->attrs & (AT_NONDATA_ENTITY & ~AT_TEMPLATE)) )
	
		check_align_target ((sp ? sp->alignp : alignp), sp);
	/*}}}*/
}
/*}}}*/

/*{{{  delete_align (alignp,  decr_nlinks) -- ok*/
delete_align (alignp,  decr_nlinks)
/*-------------------------------*/
	Align  *alignp;
	bool    decr_nlinks;
/*--------------------------------------------------------------*/
/*	Called by 'clear_symtab ()' and 'yyparse ()' to delete	*/
/* an align record.						*/
/* N.B. We assume that (alignp != NULL)!			*/
/*--------------------------------------------------------------*/
{
	Expr  **epp;
	int     dim,  src_rank,  embed_rank,  replic_rank;
	int32   attrs;
	
	if (decr_nlinks)
		alignp->nlinks--;

	if (alignp->nlinks)
		/*----------------------------------------------*/	
		/* 'alignp' is still in use, so don't delete it.*/
		/*----------------------------------------------*/
		return;

	/*{{{  find 'src_rank', 'embed_rank' & 'replic_rank'*/
	attrs = alignp->attrs;
	
	src_rank    = (attrs & AL_SRC_RANK_MASK);
	embed_rank  = (attrs & AL_EMBED_RANK_MASK) >> AL_EMBED_RANK_SHIFT;
	replic_rank = (attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT;
	/*}}}*/
	if (src_rank  &&  alignp->al) {
		/*----------------------------------------------*/
		/* Delete the exprs attached to 'alignp->al []'	*/
		/* and 'as []', as well as arrays 'al []',	*/
		/* 'as []' and 'd []' themselves.		*/
		/*						*/
		/* N.B. The above test checks that 'alignp->al'	*/
		/* is allocated, in case this is called as a	*/
		/* result of a syntax error during the parsing	*/
		/* of an ALIGN attr, when (src_rank > 0) but	*/
		/* before 'install_align ()' has been called.	*/
		/*----------------------------------------------*/
		for (epp = alignp->al,  dim = src_rank;   dim;   --dim)
			delete_expr (*epp++, DECR_NLINKS);

		for (epp = alignp->as,  dim = src_rank;   dim;   --dim)
			delete_expr (*epp++, DECR_NLINKS);

		free  ((char *) alignp->al);
		free  ((char *) alignp->as);
		free  ((char *) alignp->d);
	}
	if (embed_rank) {
		/*----------------------------------------------*/
		/* Delete the exprs attached to 'alignp->ei []'	*/
		/* and the arrays 'ei []' and 'ed []'.		*/
		/*----------------------------------------------*/
		for (epp = alignp->ei,  dim = embed_rank;   dim;   --dim)
			delete_expr (*epp++, DECR_NLINKS);

		free  ((char *) alignp->ei);
		free  ((char *) alignp->ed);
	}
	if (replic_rank)
		/*--------------------------------------*/
		/* Delete array 'alignp->rd []'.	*/
		/*--------------------------------------*/
		free  ((char *) alignp->rd);

	free  ((char *) alignp);
}
/*}}}*/
/*{{{  Align  *copy_align (alignp) -- ok*/
Align  *copy_align (alignp)
/*-----------------------*/
	Align  *alignp;
/*--------------------------------------------------------------*/
/*	Called by 'check_alignee ()' if it has to modify	*/
/* an 'alignp' record that has (nlinlks > 1).			*/
/*--------------------------------------------------------------*/
{
	Align  *new_alignp;
	Expr  **f_epp,  **t_epp;
	int     dim,  src_rank,  embed_rank,  replic_rank;
	short  *f_ip,  *t_ip;
	
	new_alignp = create_align ();

	new_alignp->tgt_sp = alignp->tgt_sp;
	new_alignp->attrs  = alignp->attrs;
	new_alignp->nlinks = 0;		/* not linked to anything yet */

	src_rank    = (alignp->attrs & AL_SRC_RANK_MASK);
	embed_rank  = (alignp->attrs & AL_EMBED_RANK_MASK) >> AL_EMBED_RANK_SHIFT;
	replic_rank = (alignp->attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT;

	if (src_rank) {
		/*{{{  allocate and copy 'alignp->al []', 'as []' and 'd []'*/
		new_alignp->al = (Expr **) emalloc (src_rank * sizeof (Expr *));
		new_alignp->as = (Expr **) emalloc (src_rank * sizeof (Expr *));
		new_alignp->d  = (short *) emalloc (src_rank * sizeof (short));
		
		for (f_epp = alignp->al,  t_epp = new_alignp->al,  dim = src_rank;
		     dim;   --dim,  ++f_epp,  ++t_epp)
			attach_expr (*t_epp, *f_epp);
		
		for (f_epp = alignp->as,  t_epp = new_alignp->as,  dim = src_rank;
		     dim;   --dim,  ++f_epp,  ++t_epp)
			attach_expr (*t_epp, *f_epp);
		
		for (f_ip = alignp->d,  t_ip = new_alignp->d,  dim = src_rank;
		     dim;   --dim,  ++f_ip,  ++t_ip)
			*t_ip = *f_ip;
		/*}}}*/
	}
	if (embed_rank) {
		/*{{{  allocate and copy 'alignp->ei []' and 'ed []'*/
		new_alignp->ei = (Expr **) emalloc (embed_rank * sizeof (Expr *));
		new_alignp->ed = (short *) emalloc (embed_rank * sizeof (short));
		
		for (f_epp = alignp->ei,  t_epp = new_alignp->ei,  dim = embed_rank;
		     dim;   --dim,  ++f_epp,  ++t_epp)
			attach_expr (*t_epp, *f_epp);
		
		for (f_ip = alignp->ed,  t_ip = new_alignp->ed,  dim = embed_rank;
		     dim;   --dim,  ++f_ip,  ++t_ip)
			*t_ip = *f_ip;
		/*}}}*/
	}
	if (replic_rank) {
		/*{{{  allocate and copy 'alignp->rd []'*/
		new_alignp->rd = (short *) emalloc (replic_rank * sizeof (short));
		
		for (f_ip = alignp->rd,  t_ip = new_alignp->rd,  dim = replic_rank;
		     dim;   --dim,  ++f_ip,  ++t_ip)
			*t_ip = *f_ip;
		/*}}}*/
	}

	return  new_alignp;
}
/*}}}*/

/*{{{  check_alignee (sp) -- OK*/
check_alignee (sp)
/*--------------*/
	Symbol  *sp;		/* has 'sp->alignp' defined! */
/*----------------------------------------------------------------
This is called by:
-- 'install_align (sp,...)', after installing alignment info ('alignp')
	in an ID 'sp' that has array dimensions ('sp->arp');
-- 'yyparse()', after attaching alignment info ('alignp') declared
	by an ALIGN attr to an ID 'sp' that has array dimensions.
-- 'install_array (sp,...)', after installing array dimensions ('arp')
	in an ID 'sp' that has alignment info ('sp->alignp');
-- 'process_decls ()', when it sees a scalar alignee 'sp'.

It does the following:
-- checks that the align source has the correct rank (whether derived
	from an explicit source list or deduced from the target), and if
	not gives an errmsg and adjusts the alignment info so that it does;
-- in any dimension 'dim' for which a '(AL_NEED_SRC_LB << dim)' flag
	is set, updates 'alignp->al [dim]' to according to the actual
	lower bound value.

N.B. If any modifications are made to 'sp->alignp' they are specific
to this alignee, so a new copy is made if  (sp->alignp->nlinks > 1).
----------------------------------------------------------------*/
{
	int     align_src_rank,  dim,  real_src_rank;
	Align  *alignp = sp->alignp;

	real_src_rank = sp->arp ? (sp->arp->attrs & RANK_MASK) : 0;

	if ( ! alignp->tgt_sp) {
		/*{{{  INHERIT:  setup 'alignp' and return*/
		if (real_src_rank) {
			int      dim;
			short   *dimp;
			Expr   **alp,  **asp;
		
			/*------------------------------------------------------*/
			/* Make a new 'alignp', as this one ('inherit_alignp')	*/
			/* is reserved for INHERITed scalars  (and is empty	*/
			/* except for 'nlinks' and (attrs == AL_CHECKED_TGT)!).	*/
			/*	For INHERITed arrays, 'sp->alignp' must be	*/
			/* unique, as each 'alignp-> l[]' and 's[]' involves	*/
			/* 'sp->name'  (e.g. "I$AT_NAME (...)").		*/
			/*------------------------------------------------------*/
			alignp->nlinks--;		/* as it will be detached from 'sp' */
			alignp = sp->alignp = create_align ();
			alignp->nlinks = 1;		/* as it's only attached to 'sp' */
		
			/*------------------------------------------------------*/
			/* allocate arrays 'alignp->al []', 'as []' and 'd []'	*/
			/*------------------------------------------------------*/
			alp  = alignp->al = (Expr **) emalloc (real_src_rank * sizeof (Expr *));
			asp  = alignp->as = (Expr **) emalloc (real_src_rank * sizeof (Expr *));
			dimp = alignp->d  = (short *) emalloc (real_src_rank * sizeof (short));
		
			/*----------------------*/
			/* initialise them	*/
			/*----------------------*/
			for (dim = 0;   dim < real_src_rank;   ++dim) {
				attach_expr (*alp++,
				              map_primary (sp->name, 0, map_subscr_al (real_src_rank, dim)));
				attach_expr (*asp++,
				              map_primary (sp->name, 0, map_subscr_as (real_src_rank, dim)));
				*dimp++ = -2;		/* i.e. unknown, because INHERITed! */
			}
		
			alignp->attrs = (real_src_rank | AL_CHECKED_TGT);
		}
		return;
		/*}}}*/
	}

	/*------------------------------------------------------*/
	/* From here on we're dealing with an explicit "ALIGN"	*/
	/* (not "INHERIT")...					*/
	/*------------------------------------------------------*/
	align_src_rank = (alignp->attrs & AL_SRC_RANK_MASK);

	if (real_src_rank != align_src_rank) {
		/*{{{  give errmsg & fix 'alignp' to have correct #(aligned dims)*/
		/*--------------------------------------------------------------*/
		/* Wrong rank is specified or deduced for the alignee in ALIGN.	*/
		/* Give an errmsg and 'fix' 'sp->alignp' to have the correct	*/
		/* source rank.							*/
		/*								*/
		/* N.B.  Since we're about to modify 'sp->alignp', make a new	*/
		/* copy of it if (sp->alignp->nlinks > 1).			*/
		/*--------------------------------------------------------------*/
		Expr  **f_epp,  **new_al,  **new_as,  **t_epp;
		int     extra_replic_dims;
		short  *f_ip,  short_buf [MAX_RANK],  *new_d,  *t_ip;
		
		
		if (alignp->nlinks > 1) {
			alignp->nlinks--;		/* as it will be detached from 'sp' */
			alignp = sp->alignp = copy_align (alignp);
			alignp->nlinks = 1;		/* as it's only attached to 'sp' */
		}
		
		/*{{{  print an errmsg*/
		/*{{{  << EXPLANATION >>*/
		/*----------------------------------------------------------------
		<< EXPLANATION >>
		-----------------
		Giving an appropriate errmsg is quite complicated!
			The possible situations are as follows (where  S = source,
		T = target,  ES = EXPL_SRC_LIST,  ET = EXPL_TGT_LIST,  DS = DEDUCED_SRC_LIST,
		DT = DEDUCED_TGT_LIST,  #S & #T = rank of S and T):
		
		case	ALIGN directive / attr		    flags;  (#dims)
		~~~~	~~~~~~~~~~~~~~~~~~~~~~		    ~~~~~~~~~~~~~~~
		after 'install_align()'
		----------------------
		1:	ALIGN (s-list) WITH T (t-list) ::   [ES | ET]	(#S,#T > 0)
		
		2a: }	ALIGN (s-list) WITH T ::	  { [ES]  if T-dims known before ALIGN
		    }					  {		(#S,#T > 0);
		2b: }					  { [ES | DT]  otherwise
		    }					  {		(#S > 0, #T >=0)
		
		3:	ALIGN  WITH T (t-list) ::	    [DS | ET]	(#S >=0, #T > 0)
		
		4a: }	ALIGN  WITH T ::		  { [DS]  if T-dims known before ALIGN
		    }					  {		(#S == #T > 0);
		4b: }					  { [-]   otherwise
		    }					  {		(#S == #T == 0) for now
		after 'check_align_target ()'
		-----------------------------
		5:	ALIGN WITH T ::			  { if (#T > 0)
							  {	converts [-]  to  [DS]
							  {	sets (#S == #T)
		
		Errmsgs if (align_src_rank != real_src_rank):
		~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		
		if (real_src_rank > 0)
		----------------------
		if (ES set)		-- cases 1 & 2   (ALIGN [S](s-list) WITH ...)
			"SRC has too many/few dimensions in ALIGN directive"
		
		else if (ET set)	-- case 3   (ALIGN WITH T (t-list) :: ...S)
			"ALIGN mismatch: %d-dimensional array SRC is aligned with
			 %d dimensions of TGT"
		
		else if (DS set)	-- cases 4a & 5  (ALIGN WITH T :: ...S  ! #T known)
			"aligned arrays SRC and TGT do not conform"
		
		else		-- case 4b i.e. no flags set  (ALIGN WITH T ::...S ! #T unknown)
			-- No errmsg (since (#S == #T) could be anything, but wasn't
			established before now).
			-- Set  (#S = #T = real_src_rank), and set [DT].
			-- 'check_align_target ()' still has to be called, and will give
			an errmsg if (#T) isn't as expected.  [How do we know this?
			Because #T is currently unknown, or the [DS] flag would have been set!]
		
		else	-- i.e. (real_src_rank == 0)
		------------------------------------
		if (ES set)		-- cases 1 & 2   (ALIGN [S](s-list) WITH ...)
			"scalar object SRC has subscripts in an ALIGN directive"
		
		else if (ET or DS set)	-- cases 3, 4a or 5  (ALIGN WITH T [(t-list)]...)
			"in ALIGNment of scalar SRC with array TGT, each dimension of SRC
			must specify replication ('*') or embedding (a subscript value)"
		
		-- else case 4b -- no flags set ([-]), which is OK!
		
		----------------------------------------------------------------*/
		/*}}}*/
		
		if (real_src_rank > 0) {
			if (alignp->attrs & AL_EXPL_SRC_LIST)		/* cases 1 & 2 */
		
				errmsg ("%q has too %s dimensions in ALIGN directive",
				         sp->name, ((align_src_rank > real_src_rank) ? "many" : "few"));
		
			else if (alignp->attrs & AL_EXPL_TGT_LIST)	/* case 3 */
		
				errmsg ("ALIGN mismatch: %d-dimensional array %q is aligned with %d dimension%c of %q",
				         real_src_rank,  sp->name,  align_src_rank,
				         ((align_src_rank != 1) ? 's' : '\0'),  alignp->tgt_sp->name);
		
			else if (alignp->attrs & AL_DEDUCED_SRC_LIST)	/* cases 4a & 5 */
		
				errmsg ("aligned arrays %q and %q do not conform",
				         sp->name,  alignp->tgt_sp->name);
		
			else {		/* case 4b, i.e. no flags set ([-]) */
			
				/*{{{  no errmsg;  update info in 'alignp' & return*/
				/*----------------------------------------------------------------
				The ALIGN directive / attr was:
				
					ALIGN WITH T ::  [...] S
				
				where (#S == #T) was unknown in 'install_align ()', so was set to 0.
					Now the rank of S, #S, is known, so:
				-- set {(l, s, dim)} == {(0,1,d),  0 <= d < #S};
				-- set (#S = #T = real_src_rank),
				-- set [DT].
				
				N.B.  'check_align_target ()' still has to be called, and
				will give an errmsg if (#T) isn't as expected.  [How do we know that?
				Because #T is currently unknown, or the [DS] flag would have been set!]
				----------------------------------------------------------------*/
				Expr  **lpp,  **spp;
				short  *dim_p;
				
				/*--------------------------------------------------------------*/
				/* allocate arrays 'alignp->al []', 's []' and 'dim []', and	*/
				/* initialise each dimension 'd' to (0,1,d).			*/
				/*--------------------------------------------------------------*/
				lpp   = alignp->al = (Expr **) emalloc (real_src_rank * sizeof (Expr *));
				spp   = alignp->as = (Expr **) emalloc (real_src_rank * sizeof (Expr *));
				dim_p = alignp->d  = (short *) emalloc (real_src_rank * sizeof (short));
				
				for (dim = 0;   dim < real_src_rank;   ++lpp,  ++spp,  ++dim_p,  ++dim) {
					attach_expr (*lpp, int_lit_0);
					attach_expr (*spp, int_lit_1);
					*dim_p = dim;
				}
				
				/*--------------------------------------------------------------*/
				/* set (align_src_rank == tgt_rank) & the 'DT' flag in		*/
				/* 'alignp->attrs'						*/
				/*--------------------------------------------------------------*/
				alignp->attrs &= AL_ASSERT;	/* ...to leave only 'AL_ASSERT' (if set) */
				alignp->attrs |= real_src_rank | (real_src_rank << AL_TGT_RANK_SHIFT) |
				                 AL_DEDUCED_TGT_LIST;
				
				return;		/* since all done now! */
				/*}}}*/
			}
		}
		else {		/* i.e. (real_src_rank == 0) */
			if (alignp->attrs & AL_EXPL_SRC_LIST)		/* cases 1 & 2 */
		
				errmsg ("scalar object %q has subscripts in an ALIGN directive",
				         sp->name);
		
			else		/* cases 3, 4a & 5, i.e. ET or DS set */
				errmsg ("in ALIGNment of scalar %q with array %q, each dimension of %q must specify replication (\'*\') or embedding (a subscript value)",
				         sp->name,  alignp->tgt_sp->name, alignp->tgt_sp->name);
		}
		/*}}}*/
		
		alignp->attrs += (real_src_rank - align_src_rank);
		
		if (real_src_rank > align_src_rank) {
			/*{{{  make new 'alignp->al[]', 'as[]' & 'd[]' with extra collapsed dims*/
			/*--------------------------------------------------------------*/
			/* (real_src_rank > align_src_rank).				*/
			/*	Create new arrays 'alignp->al []', 'as []' and 'd []'.	*/
			/* Copy the old entries to them, and make the required extra	*/
			/* dimensions collapsed ones.					*/
			/*--------------------------------------------------------------*/
			/*--------------------------------------------------------------*/
			/* allocate arrays 'new_al []', 'new_as []' and 'new_d []'	*/
			/*--------------------------------------------------------------*/
			new_al = (Expr **) emalloc (real_src_rank * sizeof (Expr *));
			new_as = (Expr **) emalloc (real_src_rank * sizeof (Expr *));
			new_d  = (short *) emalloc (real_src_rank * sizeof (short));
			
			/*-----------------------*/
			/* initialise 'new_al []' */
			/*-----------------------*/
			for (f_epp = alignp->al,  t_epp = new_al,  dim = 0;   dim < align_src_rank;
			                                                      ++dim)
				*t_epp++ = *f_epp++;
			
			for (  ;   dim < real_src_rank;   ++dim,  ++t_epp)
				attach_expr (*t_epp, star_primary);
			
			/*-----------------------*/
			/* initialise 'new_as []' */
			/*-----------------------*/
			for (f_epp = alignp->as,  t_epp = new_as,  dim = 0;   dim < align_src_rank;
			                                                      ++dim)
				*t_epp++ = *f_epp++;
			
			for (  ;   dim < real_src_rank;   ++dim,  ++t_epp)
				attach_expr (*t_epp, star_primary);
			
			/*-------------------------*/
			/* initialise 'new_d []' */
			/*-------------------------*/
			for (f_ip = alignp->d,  t_ip = new_d,  dim = 0;  dim < align_src_rank;   ++dim)
				*t_ip++ = *f_ip++;
			
			for (  ;   dim < real_src_rank;   ++dim,  ++t_ip)
				*t_ip = -1;
			
			/*--------------------------------------------------------------*/
			/* delete the old arrays and attach the new ones to 'alignp'	*/
			/*--------------------------------------------------------------*/
			if (align_src_rank) {
				free  ((char *) alignp->al);
				free  ((char *) alignp->as);
				free  ((char *) alignp->d);
			}
			alignp->al = new_al;
			alignp->as = new_as;
			alignp->d  = new_d;
			/*}}}*/
		}
		else {
			/*{{{  convert superfluous aligned dims to replicated ones*/
			/*----------------------------------------------------------------
			(real_src_rank < align_src_rank).
				Delete the superfluous entries in 'alignp-> al []' and 'as[]',
			and convert any superfluous *aligned* (as opposed to collapsed)
			dimensions to replicated ones, i.e. add their target dimensions
			('alignp->d []') to the end of 'alignp->rd []'.  This will
			require the creation of a new 'alignp->rd []' array.
			
			N.B.  We *delete* the extra 'alignp->al[]' and 'as[]' entries
			because they're exprs.  If we didn't, it won't be done by
			'delete_align ()', so they would remain as garbage.
			----------------------------------------------------------------*/
			/*--------------------------------------------------------------*/
			/* delete unwanted entries in 'alignp->al []', and record	*/
			/* superfluous *aligned* dims in 'short_buf []'.	 	*/
			/*--------------------------------------------------------------*/
			extra_replic_dims = 0;
			for (f_epp = alignp->al + real_src_rank,  dim = real_src_rank;
			     dim < align_src_rank;   ++dim) {
			
				if ((*f_epp)->token != ND_STAR) {
					int  tgt_dim = alignp->d [dim];
					
					short_buf [extra_replic_dims++] = tgt_dim;
					clear_flags (alignp->attrs, (AL_NEED_SRC_LB << dim));
					clear_flags (alignp->attrs, (AL_NEED_TGT_LB << tgt_dim));
									/* ... in case they're set! */
				}
				delete_expr (*f_epp++, DECR_NLINKS);
			}
			
			/*----------------------------------------------*/
			/* delete unwanted entries in 'alignp->as []'	*/
			/*----------------------------------------------*/
			for (f_epp = alignp->as + real_src_rank,  dim = real_src_rank;
			     dim < align_src_rank;   ++dim)
				delete_expr (*f_epp++, DECR_NLINKS);
			
			/*------------------------------------------------------*/
			/* if the alignee is actually a scalar, delete arrays	*/
			/* 'alignp->al []', 's []' and 'dim []' themselves.	*/
			/*------------------------------------------------------*/
			if ( ! real_src_rank) {
				free  ((char *) alignp->al);
				free  ((char *) alignp->as);
				free  ((char *) alignp->d);
			}
			
			if (extra_replic_dims) {
				/*{{{  create new 'alignp->rd []' array containing extra replic dims*/
				/*--------------------------------------------------------------*/
				/* Create a new 'alignp->rd []' array containing the extra	*/
				/* replicated dims.						*/
				/*--------------------------------------------------------------*/
				int     new_replic_rank,  old_replic_rank;
				short  *new_rd;
				
				old_replic_rank = (alignp->attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT;
				new_replic_rank = old_replic_rank + extra_replic_dims;
				
				new_rd = (short *) emalloc (new_replic_rank * sizeof (short));
				     
				for (f_ip = alignp->rd,  t_ip = new_rd,  dim = 0;
				     dim < old_replic_rank;   ++dim)
				
					*t_ip++ = *f_ip++;
				
				for (f_ip = short_buf;   dim < new_replic_rank;   ++dim)
					*t_ip++ = *f_ip++;
				
				/*--------------------------------------------------------------*/
				/* delete old 'rd' record and attach new one to 'alignp', and	*/
				/* enter new value of 'replic_rank' in 'alignp->attrs'.		*/
				/*--------------------------------------------------------------*/
				if (old_replic_rank)
					free  ((char *) alignp->rd);
				
				alignp->rd = new_rd;
				
				alignp->attrs += (extra_replic_dims << AL_REPLIC_RANK_SHIFT);
				/*}}}*/
			}
			/*}}}*/
		}
		/*}}}*/
	}

	if (alignp->attrs & AL_NEED_SRC_LB_MASK) {
		/*{{{  adjust 'alignp->al []' according to value of lower bound.*/
		/*--------------------------------------------------------------*/
		/* Each dimension 'dim' for which the flag:			*/
		/*	(AL_NEED_SRC_LB << dim)					*/
		/* is set had its alignment specified by an align dummy		*/
		/* variable, in which case the alignee's lower bound 'lb' must	*/
		/* be known to determine 'alignp->al [dim]'.  Correct the	*/
		/* latter as follows:						*/
		/*								*/
		/*	alignp->al [dim] += (lb * alignp->as [dim])		*/
		/*								*/
		/* (See 'install_align ()' for a full explanation.)		*/
		/*								*/
		/* N.B.  Since we're about to modify 'sp->alignp', make a new	*/
		/* copy of it if (sp->alignp->nlinks > 1).			*/
		/*--------------------------------------------------------------*/
		Expr  **lpp,  **lbpp,  **spp;
		
		if (alignp->nlinks > 1) {
			alignp->nlinks--;		/* as it will be detached from 'sp' */
			alignp = sp->alignp = copy_align (alignp);
			alignp->nlinks = 1;		/* as it's only attached to 'sp' */
		}
		
		for (lpp = alignp->al,  spp = alignp->as,  lbpp = sp->arp->l,  dim = 0;
		     dim < real_src_rank;    ++lpp,  ++spp,  ++lbpp,  ++dim) {
		
			if (alignp->attrs & (AL_NEED_SRC_LB << dim)) {
				(*lpp)->nlinks--;	/* to detach it from '*lpp' */
				attach_expr (*lpp,  expr (*lpp,
				                          ND_PLUS_OP,
				                          expr (*lbpp, ND_TIMES_OP, *spp)));
			}
		}
		clear_flags (alignp->attrs, AL_NEED_SRC_LB_MASK);
		/*}}}*/
	}
}
/*}}}*/
/*{{{  check_align_target (alignp, sp) -- OK*/
check_align_target (alignp, sp)
/*---------------------------*/
	Align   *alignp;
	Symbol  *sp;		/* alignee -- may be NULL; for errmsgs only */
/*----------------------------------------------------------------
This is called by:
-- 'install_align ()', when installing alignment info ('alignp')
	where the align target's array dimensions are already known;
-- 'process_decls ()', when it sees an 'alignp' whose align
	target wasn't 'checked out' as above (in which case the
	flag 'AL_CHECKED_TGT' isn't set in 'alignp->attrs').
	
It does the following:
-- checks that the align target has the correct rank (whether derived
	from an explicit subscript list or deduced from the source),
	and if not gives an errmsg and adjusts the alignment info so
	that it does;
-- for each dimension 'tgt_dim' for which a '(AL_NEED_TGT_LB << tgt_dim)'
	flag is set, updates 'alignp->al [i]' or 'ei [i]'  (where 'i'
	is s.t. (alignp->d / ed [i] == tgt_dim))  to correct for
	the align target's lower bound in dimension 'tgt_dim'.

N.B.  If any modifications are made to 'alignp' they apply to *all*
alignees that 'alignp' is attached to, so we *don't* copy it if
(alignp->nlinks > 1).  (This is different from 'check_alignee ()').
----------------------------------------------------------------*/
{
	bool    given_errmsg;
	int     actual_tgt_rank,  dim,  tgt_rank;
	Symbol  *tgt_sp = alignp->tgt_sp;

	given_errmsg = FALSE;	/* ...used to ensure <= 1 errmsg given */

	if (tgt_sp->attrs & (AT_UNTYPED_ENTITY & ~AT_TEMPLATE)) {
		/*{{{  give errmsg*/
		/*--------------------------------------------------------------*/
		/* A non-data entity can't be an align target.  Report an error,*/
		/* and do one of two things:					*/
		/*								*/
		/* -- if the object type is one of those in AT_OVERRIDE_ATTRS   */
		/* (i.e. AT_CONSTRUCT), then clear this flag.  In other words,	*/
		/* forget the object's old identity and give it a new one!	*/
		/* -- otherwise quit.						*/
		/*--------------------------------------------------------------*/
		switch (tgt_sp->attrs & AT_UNTYPED_ENTITY) {
		case  AT_SUBR :			begin_errmsg ("subroutine");		break;
		case  AT_MODULE :		begin_errmsg ("module");		break;
		case  AT_PROGRAM :		begin_errmsg ("main program");		break;
		case  AT_BLOCK_DATA :		begin_errmsg ("block data name");	break;
		case  AT_DERIVED_TYPE :		begin_errmsg ("derived type");		break;
		case  AT_NAMELIST_GROUP :	begin_errmsg ("namelist group");	break;
		case  AT_CONSTRUCT :		begin_errmsg ("construct name");	break;
		case  AT_PROCESSORS :		begin_errmsg ("processors arrangement");break;
		}
		end_errmsg (" %q cannot be an ALIGN target", tgt_sp->name);
		
		if (tgt_sp->attrs & AT_CONSTRUCT)
			clear_flags (tgt_sp->attrs, AT_CONSTRUCT);
		
		given_errmsg = TRUE;
		/*}}}*/
	}
	else if (tgt_sp->attrs & (AT_FUNC | AT_INTRINSIC | AT_EXTERNAL)) {
		/*{{{  give errmsg*/
		/*--------------------------------------------------------------*/
		/* Only data *objects* can be align targets -- i.e. functions	*/
		/* can't.  Report an error and return.				*/
		/*--------------------------------------------------------------*/
		errmsg ("%s %q cannot be an ALIGN target",
		         ((tgt_sp->attrs & (AT_INTRINSIC | AT_FUNC)) ? "function" : "procedure"),
		           tgt_sp->name);
		
		given_errmsg = TRUE;
		/*}}}*/
	}

	tgt_rank = (alignp->attrs & AL_TGT_RANK_MASK) >> AL_TGT_RANK_SHIFT;
	actual_tgt_rank = tgt_sp->arp ? (tgt_sp->arp->attrs & RANK_MASK) : 0;

	if (actual_tgt_rank != tgt_rank) {
		/*{{{  give errmsg & 'fix' 'alignp' to have correct #(target dims)*/
		/*----------------------------------------------------------------
		Wrong rank is specified or deduced for the align target in ALIGN.
		Give an errmsg and 'fix' 'alignp' to have the correct target rank.
		
		N.B.  In this case, although we're about to modify 'alignp', we
		don't make a copy of it if (alignp->nlinks > 1) as these changes
		apply to *all* alignees that 'alignp' is attached to.  (This is
		different from 'check_alignee ()').
		----------------------------------------------------------------*/
		Expr  **f_epp,  **lpp,  **spp,  **t_epp;
		short  *dim_p,  *f_ip,  *new_rd,  *t_ip;
		
		
		/*{{{  print errmsg*/
		/*{{{  << EXPLANATION >>*/
		/*----------------------------------------------------------------
		<< EXPLANATION >>
		-----------------
		Giving an appropriate errmsg is quite complicated!
			The possible situations are as follows (where  S = source,
		T = target,  ES = EXPL_SRC_LIST,  ET = EXPL_TGT_LIST,  DS = DEDUCED_SRC_LIST,
		DT = DEDUCED_TGT_LIST,  #S & #T = rank of S and T):
		
		case	ALIGN directive / attr		    flags;  (#dims)
		~~~~	~~~~~~~~~~~~~~~~~~~~~~		    ~~~~~~~~~~~~~~~
		after 'install_align()'
		----------------------
		1:	ALIGN (s-list) WITH T (t-list) ::   [ES | ET]	(#S,#T > 0)
		
		2:	ALIGN  WITH T (t-list) ::	    [DS | ET]	(#S >=0, #T > 0)
		
		3a: }	ALIGN (s-list) WITH T ::	  { [ES]  if #T known before ALIGN
		    }					  {		(#S,#T > 0);
		3b: }					  { [ES | DT]  otherwise
		    }					  {		(#S > 0, #T >=0)
		
		4a: }	ALIGN  WITH T ::		  { [DS]  if #T known before ALIGN
		    }					  {		(#S == #T > 0);
		4b: }					  { [-]   otherwise
		    }					  {		(#S == #T == 0) for now
		after 'check_alignee ()'
		------------------------
		5:	ALIGN WITH T ::			  { if (#S > 0)
							  {	converts [-]  to  [DT]
							  {	sets (#S == #T)
		
		Errmsgs if (tgt_rank != actual_tgt_rank):
		~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		N.B.:
		--  Cases (3a & 4a) needn't be considered, as they imply that #T is correct.
		
		--  In all cases except (1 & 2) (viz cases 3b, 4b & 5), this must
		be called from 'process_decls()' rather than 'install_align()', so an
		alignee 'sp' is known and can be named in errmsgs  (though it may
		just be the 1st of several with the same 'alignp').
			[How do we known that?  Because for it to be called from
		'install_align()', #T must be known before the ALIGN directive / attr.
		In that case, however, we wouldn't *deduce* #T from the alignee (i.e set
		[DT], cases 3b & 5), and wouldn't have *no* flags set (i.e. [-], case 4b).]
		
		if (actual_tgt_rank > 0)
		------------------------
		if (ET set)		-- cases 1 & 2   (ALIGN [...] WITH T (t-list))
			"TGT has too many/few dimensions in ALIGN directive"
		
		else if (ES set)	-- case 3   (ALIGN [S](s-list) WITH T)
			"ALIGN mismatch: %d-dimensions of SRC are aligned with
			 %d dimensional array TGT"
		
		else if (DT set)	-- case 5   (ALIGN WITH T :: ...S  ! #S known)
			"aligned arrays SRC and TGT do not conform"
		
		else		-- case 4b, no flags set   (ALIGN WITH T :: ...S  ! #S unknown)
			-- No errmsg (since (#S == #T) could be anything, but wasn't
			established before now).
			-- Set  (#S = #T = actual_tgt_rank), and set [DS].
			-- 'check_alignee ()' will be called from 'process_decls()'
			immediately after this returns, and will give an errmsg
			since in fact (#S == 0) rather that (== #T > 0).
				[Why?  [-] means that #T was unknown in 'install_align()',
			so this must be called from 'process_decls()'.  The fact that it's
			still [-] means that S hasn't been declared as an array --
			otherwise 'check_alignee()' would have changed it to [DT].
			Hence S is scalar, so 'check_alignee()' is called
			immediately after this and will complain about the rank mismatch.
				N.B. An advantage of this way of organising the errmsgs
			is that a separate errmsg will be given for *every* scalar alignee
			with this 'alignp'.]
		
		else	-- i.e. (actual_tgt_rank == 0)
		--------------------------------------
		if (ET set)		-- cases 1 & 2   (ALIGN [...] WITH T (t-list))
			"scalar object TGT has subscripts in an ALIGN directive"
		
		else if (ES or DT set)	-- cases 3b & 5   (ALIGN [...] WITH T)
		
			"in ALIGNment of array SRC with scalar TGT, all dimensions of SRC
			must be collapsed (i.e. have '*' specified)"
		
		-- else case 4b -- no flags set ([-]), which is OK!
		
		----------------------------------------------------------------*/
		/*}}}*/
		
		if ((actual_tgt_rank > 0)  &&  ! (alignp->attrs &
		            (AL_EXPL_SRC_LIST | AL_EXPL_TGT_LIST | AL_DEDUCED_TGT_LIST)) ) {
						/* case 4b, i.e. no flags set ([-]) */
		
			/*{{{  no errmsg;  update info in 'alignp' & return*/
			/*----------------------------------------------------------------
			The ALIGN directive / attr was:
			
				ALIGN WITH T ::  [...] S
			
			where (#S == #T) was unknown in 'install_align ()', so was set to 0.
				Now the rank of T, #T, is known, so:
			-- set {(l, s, dim)} == {(0,1,d),  0 <= d < #T};
			-- set (#S = #T = actual_tgt_rank),
			-- set [DS].
			
			N.B.  'check_alignee ()' will be called from 'process_decls()'
			immediately after this returns, and will give an errmsg since in fact
			(#S == 0) rather that (== #T > 0).
				[Why?  Because [-] means that #T was unknown in
			'install_align()', so this must be called from 'process_decls()'.
			The fact that it's still [-] means that S hasn't been declared as
			an array -- otherwise 'check_alignee()' would have changed
			it to [DT].  Hence S is scalar, so 'check_alignee()' is called
			immediately after this, and will complain about the rank mismatch.
			----------------------------------------------------------------*/
			/*--------------------------------------------------------------*/
			/* allocate arrays 'alignp->al []', 's []' and 'dim []', and	*/
			/* initialise each dimension 'd' to (0,1,d).			*/
			/*--------------------------------------------------------------*/
			lpp   = alignp->al = (Expr **) emalloc (actual_tgt_rank * sizeof (Expr *));
			spp   = alignp->as = (Expr **) emalloc (actual_tgt_rank * sizeof (Expr *));
			dim_p = alignp->d  = (short *) emalloc (actual_tgt_rank * sizeof (short));
			
			for (dim = 0;   dim < actual_tgt_rank;   ++lpp,  ++spp,  ++dim_p,  ++dim) {
				attach_expr (*lpp, int_lit_0);
				attach_expr (*spp, int_lit_1);
				*dim_p = dim;
			}
			
			/*--------------------------------------------------------------*/
			/* set (src_rank == tgt_rank) & 'DS' flag in 'alignp->attrs'	*/
			/*--------------------------------------------------------------*/
			alignp->attrs &= AL_ASSERT;	/* ...to leave only 'AL_ASSERT' (if set) */
			alignp->attrs |= actual_tgt_rank | (actual_tgt_rank << AL_TGT_RANK_SHIFT) |
			                 AL_DEDUCED_SRC_LIST | AL_CHECKED_TGT;
			
			return;		/* since all done now! */
			/*}}}*/
		}
		else if (given_errmsg)
			;		/* don't bother with another errmsg! */
		
		else if (actual_tgt_rank > 0) {
			if (alignp->attrs & AL_EXPL_TGT_LIST)		/* cases 1 & 2*/
			
				errmsg ("%q has too %s dimensions in ALIGN directive",
				         tgt_sp->name, ((tgt_rank > actual_tgt_rank) ? "many" : "few"));
			
			else if (alignp->attrs & AL_EXPL_SRC_LIST)	/* case 3b */
			
				errmsg ("ALIGN mismatch: %d dimension%c of %q %s aligned with %d-dimensional array %q",
				         tgt_rank,  ((tgt_rank == 1) ? '\0' : 's'),
			                 sp->name,  ((tgt_rank == 1) ? "is" : "are"),
				         actual_tgt_rank,  tgt_sp->name);
			
			else if (alignp->attrs & AL_DEDUCED_TGT_LIST)	/* case 5 */
			
				errmsg ("aligned arrays %q and %q do not conform",
				         sp->name,  alignp->tgt_sp->name);
				
		}
		else {		/* i.e. (actual_tgt_rank == 0) */
			if (alignp->attrs & AL_EXPL_TGT_LIST)		/* cases 1 & 2 */
		
				errmsg ("scalar object %q has subscripts in an ALIGN directive",
				         tgt_sp->name);
		
			else		/* cases 3b & 5, i.e. ES or DT */
		
				errmsg ("in ALIGNment of array %q with scalar %q, all dimensions of %q must be collapsed (i.e. have \'*\' specified)",
				         sp->name,  alignp->tgt_sp->name, sp->name);
		}
		/*}}}*/
		
		alignp->attrs += ((actual_tgt_rank - tgt_rank) << AL_TGT_RANK_SHIFT);
		
		if (actual_tgt_rank > tgt_rank) {
			/*{{{  make new 'alignp->rd []' with extra replicated dims -- OK*/
			/*--------------------------------------------------------------*/
			/* (actual_tgt_rank > tgt_rank).				*/
			/*	Create a new array 'alignp->rd []'.  Copy the old	*/
			/* entries to it and add the required extra dimensions (i.e.	*/
			/* make them replicated ones).					*/
			/*--------------------------------------------------------------*/
			int     old_replic_rank,  new_replic_rank;
			short  *new_rd;
			
			old_replic_rank = (alignp->attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT;
			new_replic_rank = old_replic_rank + (actual_tgt_rank - tgt_rank);
			
			new_rd = (short *) emalloc (new_replic_rank * sizeof (short));
			
			for (f_ip = alignp->rd,  t_ip = new_rd,  dim = old_replic_rank;
			     dim;   --dim)
				*t_ip++ = *f_ip++;
			
			for (dim = tgt_rank;   dim < actual_tgt_rank;   ++dim)
				*t_ip++ = dim;		/* i.e. source is replicated over 'dim' */
			
			/*--------------------------------------------------------------*/
			/* delete the old record and attach the new one to 'alignp'	*/
			/*--------------------------------------------------------------*/
			if (old_replic_rank)
				free  ((char *) alignp->rd);
			
			alignp->rd = new_rd;
			
			alignp->attrs += ((actual_tgt_rank - tgt_rank) << AL_REPLIC_RANK_SHIFT);
			/*}}}*/
		}
		else {
			/*{{{  collapse src dims aligned with non-existent tgt dims, & remove surplus replicated & embedded dims -- OK*/
			/*----------------------------------------------------------------
			(actual_tgt_rank < tgt_rank).
				Convert source dimensions in 'alignp->al []', 'as []' and
			'd []' that are aligned with non-existent target dimensions into
			collapsed dimensions (i.e. change their entries to (*, *, -1)).
				Remove entries for non-existant target dimensions from
			the arrays 'alignp->ei []', 'ed []' and 'rd []'.
			----------------------------------------------------------------*/
			int  embed_rank,  replic_rank,  src_rank,  tgt_dim;
			
			src_rank    = (alignp->attrs & AL_SRC_RANK_MASK);
			embed_rank  = ((alignp->attrs & AL_EMBED_RANK_MASK) >> AL_EMBED_RANK_SHIFT);
			replic_rank = ((alignp->attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT);
			
			/*--------------------------------------------------------------*/
			/* Collapse source dims that are aligned with non-existent	*/
			/* target dims.							*/
			/*--------------------------------------------------------------*/
			for (lpp = alignp->al,  spp = alignp->as,  dim_p = alignp->d,  dim = 0;
			     dim < src_rank;     ++lpp,  ++spp,  ++dim_p,  ++dim) {
			
				if ((tgt_dim = *dim_p) >= actual_tgt_rank) {
					/*----------------------------------------------*/
					/* 'src_dim' is aligned with a non-existent	*/
					/* target dim, so convert it to a collapsed dim.*/
					/*----------------------------------------------*/
					delete_expr (*lpp, DECR_NLINKS);
					delete_expr (*spp, DECR_NLINKS);
					
					attach_expr (*lpp, star_primary);
					attach_expr (*spp, star_primary);
					*dim_p = -1;
			
					clear_flags (alignp->attrs, (AL_NEED_SRC_LB << dim));
					clear_flags (alignp->attrs, (AL_NEED_TGT_LB << tgt_dim));
									/* ... in case they're set! */
				}
			}
			if (embed_rank) {
				/*{{{  remove non-existent tgt dims from 'alignp->ei []' & 'ed []'*/
				/*--------------------------------------------------------------*/
				/* remove non-existant tgt dims from 'alignp->ei []' and	*/
				/* 'ed' [], and if no embedded dims remain delete the arrays	*/
				/* 'alignp->ei []' and 'ed []' themselves.			*/
				/*--------------------------------------------------------------*/
				for (f_epp = t_epp = alignp->ei,  f_ip = t_ip = alignp->ed,  dim = embed_rank;
				     dim;    ++f_epp,  ++f_ip,  --dim) {
				
					if ((tgt_dim = *f_ip) < actual_tgt_rank) {
						/*----------------------------*/
						/* tgt dim exists, so copy it */
						/*----------------------------*/
						*t_epp++ = *f_epp;
						*t_ip++  = *f_ip;
					}
					else {
						/*------------------------------------*/
						/* non-existent tgt dim, so remove it */
						/*------------------------------------*/
						delete_expr (*f_epp, DECR_NLINKS);
						alignp->attrs -= (1 << AL_EMBED_RANK_SHIFT);
						clear_flags (alignp->attrs, (AL_NEED_TGT_LB << tgt_dim));
										/* ... in case it's set! */
					}
				}
				
				if ( ! (alignp->attrs & AL_EMBED_RANK_MASK)) {
					free  ((char *) alignp->ei);
					free  ((char *) alignp->ed);
				}
				/*}}}*/
			}
			
			if (replic_rank) {
				/*{{{  remove non-existent tgt dims from 'alignp->rd []'*/
				/*--------------------------------------------------------------*/
				/* remove non-existant tgt dims from 'alignp->rd []', and if	*/
				/* no replicated dims remain delete 'alignp->rd []' itself.	*/
				/*--------------------------------------------------------------*/
				for (f_ip = t_ip = alignp->rd,  dim = replic_rank;
				     dim;    ++f_ip,  --dim){
				
					if (*f_ip < actual_tgt_rank)	/* tgt dim exists, so copy it */
						*t_ip++ = *f_ip;
					else				/* non-existent tgt dim, so forget it*/
						alignp->attrs -= (1 << AL_REPLIC_RANK_SHIFT);
				}
				
				if ( ! (alignp->attrs & AL_REPLIC_RANK_MASK))
					free  ((char *) alignp->rd);
				/*}}}*/
			}
			/*}}}*/
		}
		/*}}}*/
	}

	if (alignp->attrs & AL_NEED_TGT_LB_MASK) {
		/*{{{  subtract target lower bounds from 'alignp->al[]' & 'ei[]'*/
		/*--------------------------------------------------------------*/
		/* For each dimension 'tgt_dim' for which the flag		*/
		/* '(AL_NEED_TGT_LB << tgt_dim)' is set, update 'alignp->al [i]'*/
		/* or 'ei [i]'  (where 'i' is s.t. (alignp->d / ed [i] ==	*/
		/* tgt_dim)) to correct for the align target's lower bound in	*/
		/* dimension 'tgt_dim'.						*/
		/*								*/
		/* N.B.  In this case, although we're about to modify 'alignp',	*/
		/* we don't make a copy of it if (alignp->nlinks > 1) as these	*/
		/* changes apply to *all* alignees that 'alignp' is attached to.*/
		/* (This is different from 'check_alignee ()').			*/
		/*--------------------------------------------------------------*/
		int32   attrs = alignp->attrs;
		int     embed_rank,  src_rank,  tgt_dim;
		short  *dim_p;
		Expr  **lpp;
		
		src_rank   = (attrs & AL_SRC_RANK_MASK);
		embed_rank = ((attrs & AL_EMBED_RANK_MASK) >> AL_EMBED_RANK_SHIFT);
		
		for (lpp = alignp->al,  dim_p = alignp->d,  dim = src_rank;
		     dim;    ++lpp,  ++dim_p,  --dim) {
		
			tgt_dim = *dim_p;
		
			if (attrs & (AL_NEED_TGT_LB << tgt_dim)) {
				(*lpp)->nlinks--;	/* to detach it from '*lpp' */
				attach_expr (*lpp,  expr (*lpp,  ND_MINUS_OP,
				                           tgt_sp->arp->l [tgt_dim]));
			}
		}
		
		for (lpp = alignp->ei,  dim_p = alignp->ed,  dim = embed_rank;
		     dim;    ++lpp,  ++dim_p,  --dim) {
		
			tgt_dim = *dim_p;
		
			if (attrs & (AL_NEED_TGT_LB << tgt_dim)) {
				(*lpp)->nlinks--;	/* to detach it from '*lpp' */
				attach_expr (*lpp,  expr (*lpp,  ND_MINUS_OP,
				                           tgt_sp->arp->l [tgt_dim]));
			}
		}
		
		clear_flags (alignp->attrs, AL_NEED_TGT_LB_MASK);
		/*}}}*/
	}

	alignp->attrs |= AL_CHECKED_TGT;
}
/*}}}*/

/*{{{  compose_align (src, first_src, align_depth) -- ok*/
compose_align (src, first_src, n_src)
/*---------------------------------*/
	Symbol  *src,  *first_src;      /* ...whose 'alignp->tgt_sp's are
					   themselves aligned */
	int  n_src;			/* position of 'src' in align chain
					   from 'first_src' (starting at 1) */
/*----------------------------------------------------------------
	Called by 'process_decls ()' to 'collapse' a multiple
alignment to a single one.  I.e. 'src' is aligned as follows:

             align1        align2
	src  ----->  tgt1  ----->  tgt2  [ ---> tgt_i ]...

On return, this is collapsed to a direct alignment to the ultimate
align target, 'tgt_n':

             align1'
	src  ------>  tgt_n

(As a by-product, any other intermediate alignees in the alignment
chain also have their alignments collapsed to direct alignments to
the ultimate align target 'tgt_n'.)

N.B.
--  'tgt2' may be NULL, which means that 'tgt1' has 'INHERIT' specified.
In this case return straight away.
--  On return, 'src->alignp' isn't changed, though its contents may be.
--  This requires that all intermediate alignees are 'fully resolved',
i.e. have their dimensions declared and checked-out in the alignments.
----------------------------------------------------------------*/
{
	/*{{{  decls*/
	Align   *align1,  *align2;
	Expr   **al1p,  *al2,  **as1p,  *as2,  **f_epp,  **new_ei,  **t_epp;
	int      dim,  d2,  embed_rank_1,  embed_rank_2,  i,  new_embed_rank,
	         new_replic_rank,  replic_rank_1,  replic_rank_2,  src_rank,  tgt_dim;
	int32    compose_assert;
	short   *d1p,  *f_ip,  *new_ed,  *new_rd,  *t_ip;
	Symbol  *sp,  *tgt1,  *tgt2;
	/*}}}*/

	align1 = src->alignp;		/* not NULL! */
	tgt1   = align1->tgt_sp;	/*  "    "   */
	align2 = tgt1->alignp;		/*  "    "   */
	if ( ! (tgt2 = align2->tgt_sp))
		return;			/* i.e. 'tgt1' has INHERIT specified */

	/*{{{  if align chain has a cycle involving 'tgt2', give errmsg & return*/
	for (sp	= first_src,  i = n_src;   i;   i--,  sp = sp->alignp->tgt_sp) {
	
		if (sp == tgt2) {
			/*------------------------------------------------------*/
			/* The alignment chain from 'first_src' contains a	*/
			/* cycle involving 'tgt2'.				*/
			/* 	Give an errmsg, and break the cycle at the ID	*/
			/* *preceding* 'tgt2' in the alignment loop, viz.'tgt1'.*/
			/* This is done by setting 'tgt1->alignp' to NULL and	*/
			/* deleting its record.  Then return.			*/
			/*							*/
			/* N.B. Since 'tgt1' is now no longer itself an align	*/
			/* source, it now forms the ultimate align target for	*/
			/* 'src', i.e. the complete alignment chain for 'src'	*/
			/* is (src -> tgt1).  Hence no mods need to be made to	*/
			/* 'align1', and this can return straight away.		*/
			/*------------------------------------------------------*/
			errmsg ("%q is indirectly aligned with itself", tgt2->name);
				
			delete_align (tgt1->alignp, DECR_NLINKS);
			tgt1->alignp = NULL_Align;
				
			return;
		}
	}
	/*}}}*/

	if (tgt2->alignp) {
		/*{{{  collapse (tgt1 -> tgt2 -> ...) to (tgt1 -> tgt_n)*/
		/*--------------------------------------------------------------*/
		/* Collapse (tgt1 -> tgt2 -> ... -> tgt_n) to (tgt1 -> tgt_n),	*/
		/* where 'tgt_n' isn't aligned.					*/
		/*								*/
		/* N.B.  This also collapses all intermediate alignments as	*/
		/* well.  E.g. if we have:					*/
		/*	tgt1 -> A -> B -> C -> D -> E				*/
		/* then on return, 'tgt1' -- 'D' are all directly aligned with	*/
		/* 'E'.								*/
		/*--------------------------------------------------------------*/
		compose_align (tgt1, first_src, ++n_src);
		/*}}}*/
	}
	/*{{{  transform 'align1->al []', 'as []' & 'd []'*/
	/*------------------------------------------------------*/
	/* Transform 'align1->al []', 'as []' and 'd []'	*/
	/*------------------------------------------------------*/
	src_rank = (align1->attrs & AL_SRC_RANK_MASK);
	
	for (al1p = align1->al,  as1p = align1->as,  d1p = align1->d,
	          dim = src_rank;    dim;   ++al1p,  ++as1p,  ++d1p,  --dim) {
	
		if ((tgt_dim = *d1p) == -1)
			/*--------------------------------------*/
			/* Collapsed dim;  remains collapsed!	*/
			/*--------------------------------------*/
			continue;
	
		else if ((d2 = align2->d [tgt_dim]) == -1) {
			/*-------------------------------------------------*/
			/* Aligned with collapsed dim;  becomes collapsed! */
			/*-------------------------------------------------*/
			delete_expr (*al1p, DECR_NLINKS);
			delete_expr (*as1p, DECR_NLINKS);
	
			attach_expr (*al1p, star_primary);
			attach_expr (*as1p, star_primary);
			*d1p = -1;
		}
		else {
			/*----------------------------------------------*/
			/* Aligned with another aligned dimension whose	*/
			/* alignment parameters are (al2, as2, d2).	*/
			/* Combine their alignments as follows:		*/
			/*	al ->  al2 + (al * as2)			*/
			/*	as ->  (as * as2)			*/
			/*	 d ->  d2				*/
			/*----------------------------------------------*/
			al2 = align2->al [tgt_dim];
			as2 = align2->as [tgt_dim];
	
			(*al1p)->nlinks--;	/* to detach it from 'align1->al[]' */
			(*as1p)->nlinks--;	/*  "    "    "   "  'align1->as[]' */
	
			attach_expr (*al1p,  expr (al2,
			                           ND_PLUS_OP,
			                           expr (*al1p, ND_TIMES_OP, as2)
			                          ));
			attach_expr (*as1p,  expr (*as1p, ND_TIMES_OP, as2));
			*d1p = d2;
		}
	}
	/*}}}*/
	/*{{{  transform 'align1->ei []' & ed []'*/
	/*----------------------------------------------*/
	/* Transform 'align1->ei []' and 'ed []'	*/
	/*----------------------------------------------*/
	embed_rank_1 = (align1->attrs & AL_EMBED_RANK_MASK) >> AL_EMBED_RANK_SHIFT;
	embed_rank_2 = (align2->attrs & AL_EMBED_RANK_MASK) >> AL_EMBED_RANK_SHIFT;
	
	if (new_embed_rank = (embed_rank_1 + embed_rank_2)) {
		/*{{{  transform 'align1->ei []' and 'ed []'*/
		for (f_ip = align1->ed,  dim = embed_rank_1;    dim;   ++f_ip,  --dim) {
		
			if (align2->d [*f_ip] == -1)
				/*----------------------------------------------*/
				/* embedded in a collapsed dim;  omit this dim	*/
				/* from 'new_ei []' and 'new_ed [].		*/
				/*----------------------------------------------*/
				new_embed_rank--;
		}
		
		/*--------------------------------------------------------------*/
		/* Allocate arrays 'new_ei[]' and 'new_ed[]', unless		*/
		/* 'new_embed_rank' is 0 (i.e. all embedded dims are collapsed!)*/
		/*--------------------------------------------------------------*/
		if (new_embed_rank) {
			new_ei = (Expr **) emalloc (new_embed_rank * sizeof (Expr *));
			new_ed = (short *) emalloc (new_embed_rank * sizeof (short));
		}
		
		/*--------------------------------------------------------------*/
		/* Process the 'align1->ei []' & 'ed []' entries -- omitting	*/
		/* those whose dims are collapsed and transforming those whose	*/
		/* dims are aligned -- and write the results to 'new_ei []' and	*/
		/* 'new_ed []'.							*/
		/*--------------------------------------------------------------*/
		for (f_epp = align1->ei,  t_epp = new_ei,  f_ip = align1->ed,
		                          t_ip = new_ed,  dim = embed_rank_1;
		     dim;    ++f_epp,  ++f_ip,  --dim) {
			
			tgt_dim = *f_ip;
			if ((d2 = align2->d [tgt_dim]) == -1)
				/*-----------------------------------------------*/
				/* embedded in collapsed dim;  delete this entry */
				/*-----------------------------------------------*/
				delete_expr (*f_epp, DECR_NLINKS);
			else {
				/*--------------------------------------------------*/
				/* embedded in aligned dim;  ei -> al2 + (ei * as2) */
				/*--------------------------------------------------*/
				al2 = align2->al [tgt_dim];
				as2 = align2->as [tgt_dim];
			
				(*f_epp)->nlinks--;	/* to detach it from 'align1->ei[]' */
		
				attach_expr (*t_epp++,  expr (al2, ND_PLUS_OP,
				                              expr (*f_epp, ND_TIMES_OP, as2)));
				*t_ip++ = d2;
			}
		}
		/*--------------------------------------------------------------*/
		/* Now copy 'align2->ei []' & 'ed []' unchanged to the end of	*/
		/* 'new_ei []' & 'new_ed []'.					*/
		/*--------------------------------------------------------------*/
		for (f_epp = align2->ei,  f_ip = align2->ed,  dim = embed_rank_2;
		     dim;   --dim) {
		
			*t_epp++ = *f_epp++;
			*t_ip++  = *f_ip++;
		}
		
		/*--------------------------------------------------------------*/
		/* Replace the original 'align1->ei []' and 'ed []' arrays by	*/
		/* 'new_ei []' and 'new_ed []'.					*/
		/*--------------------------------------------------------------*/
		if (embed_rank_1) {
			free  ((char *) align1->ei);
			free  ((char *) align1->ed);
		}
		align1->ei = new_ei;
		align1->ed = new_ed;
		/*}}}*/
	}
	
	
	
	/*}}}*/
	/*{{{  transform 'align1->rd []'*/
	/*------------------------------*/
	/* Transform 'align1->rd []'	*/
	/*------------------------------*/
	replic_rank_1 = (align1->attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT;
	replic_rank_2 = (align2->attrs & AL_REPLIC_RANK_MASK) >> AL_REPLIC_RANK_SHIFT;
	
	if (new_replic_rank = (replic_rank_1 + replic_rank_2)) {
		/*{{{  transform 'align1->rd []'*/
		for (f_ip = align1->rd,  dim = replic_rank_1;    dim;    ++f_ip,  --dim) {
		
			if (align2->d [*f_ip] == -1)
				/*----------------------------------------------*/
				/* replicated over collapsed dim;  omit this	*/
				/* entry from 'new_rd [].			*/
				/*----------------------------------------------*/
				new_replic_rank--;
		}
		
		/*--------------------------------------------------------------*/
		/* Allocate array 'new_rd []', unless 'new_replic_rank' is 0	*/
		/*--------------------------------------------------------------*/
		if (new_replic_rank)
			new_rd = (short *) emalloc (new_replic_rank * sizeof (short));
		
		/*--------------------------------------------------------------*/
		/* Process the 'align1->rd []' dimension entries -- omitting	*/
		/* collapsed ones and copying aligned ones to 'new_rd []'.	*/
		/*--------------------------------------------------------------*/
		for (f_ip = align1->rd,  t_ip = new_rd,  dim = replic_rank_1;
		     dim;    ++f_ip,  --dim) {
		
			if ((d2 = align2->d [*f_ip]) != -1)
				*t_ip++ = d2;
		}
		
		/*--------------------------------------------------------------*/
		/* Now copy 'align2->rd []' unchanged to the end of 'new_rd []'.*/
		/*--------------------------------------------------------------*/
		for (f_ip = align2->rd,  dim = replic_rank_2;    dim;   --dim)
			*t_ip++  = *f_ip++;
		
		/*--------------------------------------------------------------*/
		/* Replace the original 'align1->rd []' array by 'new_rd []'	*/
		/*--------------------------------------------------------------*/
		if (replic_rank_1)
			free  ((char *) align1->rd);
		
		align1->rd = new_rd;
		/*}}}*/
	}
	/*}}}*/

	align1->tgt_sp = align2->tgt_sp;	/* the new target object */

	compose_assert = (align1->attrs & align2->attrs & AL_ASSERT);
	align1->attrs  =  src_rank | (align2->attrs & AL_TGT_RANK_MASK) |
	                 (new_embed_rank << AL_EMBED_RANK_SHIFT) |
	                 (new_replic_rank << AL_REPLIC_RANK_SHIFT) |
	                  AL_CHECKED_TGT | compose_assert;
}
/*}}}*/

/*}}}*/

/*{{{  <<< HPF DISTRIBUTE functions >>>*/
/*-----------------------------------*/
/* HPF DISTRIBUTE handling functions */
/*-----------------------------------*/
/*{{{  Distr *create_distr () -- OK*/
Distr *create_distr ()
/*------------------*/
/*------------------------------*/
/* Create a 'Distr' record.	*/
/*------------------------------*/
{
	return  (Distr *) ecalloc (sizeof (Distr));
}
/*}}}*/
/*{{{  bool  canbe_distributed (sp) -- ok*/
bool  canbe_distributed (sp)
/*------------------------*/
	Symbol  *sp;
/*--------------------------------------------------------------*/
/*	Called by 'yyparse ()' to determine whether an ID ('sp')*/
/* can be distributed.  If the 'distr-spec' directly follows	*/
/* the ID in a DISTRIBUTE directive, this is called prior to	*/
/* 'create_distr ()', while in an HPF 'combined-directive' with	*/
/* a DISTRIBUTE attr, it's called prior to 'install_distr ()'.	*/
/*--------------------------------------------------------------*/
{
	if (sp->distp) {
		errmsg ("distribution of %q already specified", sp->name);
		return  FALSE;
	}
	else if (sp->alignp  &&  sp->alignp->tgt_sp) {
		errmsg ("cannot DISTRIBUTE %q as it is explicitly ALIGNed", sp->name);
		return  FALSE;
	}
	else if (sp->attrs & (AT_UNTYPED_ENTITY & ~AT_TEMPLATE)) {
		/*{{{  give errmsg & return FALSE*/
		/*--------------------------------------------------------------*/
		/* Except for templates, non-data entities can't be distributed.*/
		/* Report an error, and do one of two things:			*/
		/*								*/
		/* -- if the object type is one of those in AT_OVERRIDE_ATTRS   */
		/* (i.e. AT_CONSTRUCT), then clear this flag and return TRUE.	*/
		/* In other words, forget the entity's old identity and give it	*/
		/* a new one!							*/
		/* -- otherwise return FALSE.					*/
		/*--------------------------------------------------------------*/
		switch (sp->attrs & (AT_UNTYPED_ENTITY & ~AT_TEMPLATE)) {
		case  AT_SUBR :			begin_errmsg ("subroutine");		break;
		case  AT_MODULE :		begin_errmsg ("module");		break;
		case  AT_PROGRAM :		begin_errmsg ("main program");		break;
		case  AT_BLOCK_DATA :		begin_errmsg ("block data name");	break;
		case  AT_DERIVED_TYPE :		begin_errmsg ("derived type");		break;
		case  AT_NAMELIST_GROUP :	begin_errmsg ("namelist group");	break;
		case  AT_CONSTRUCT :		begin_errmsg ("construct name");	break;
		case  AT_PROCESSORS :		begin_errmsg ("processors arrangement");break;
		}
		end_errmsg (" %q cannot be distributed", sp->name);
		
		if (sp->attrs & AT_CONSTRUCT) {
			clear_flags (sp->attrs, AT_CONSTRUCT);
			return  TRUE;
		}
		else
			return  FALSE;
		/*}}}*/
	}
	else if (sp->attrs & (AT_FUNC | AT_INTRINSIC | AT_EXTERNAL)) {
		/*{{{  give errmsg & return FALSE*/
		/*--------------------------------------------------------------*/
		/* Only data *objects* or templates can be distributed -- i.e.	*/
		/* functions can't be.  Report an error and return FALSE.	*/
		/*--------------------------------------------------------------*/
		
		if (sp->attrs & AT_INTRINSIC)
			/*------------------------------------------------------*/
			/* ...must be an intrinsic function, as AT_SUBR was	*/
			/* caught earlier.					*/
			/*------------------------------------------------------*/
			errmsg ("intrinsic function %q cannot be distributed", sp->name);
		
		else if (sp->attrs & AT_FUNC)
			errmsg ("can only specify distribution of function result %q in an explicit interface or function definition",
			         sp->name);
		
		else	/*-----------------------*/
			/* an EXTERNAL procedure */
			/*-----------------------*/
			errmsg ("external procedure %q cannot be distributed", sp->name);
		
		return  FALSE;
		/*}}}*/
	}

	return  TRUE;		/* OK! */
}
/*}}}*/
/*{{{  install_distr_fmt (fmt, fmt_buf, b, b_buf, distp, sp) -- ok*/
install_distr_fmt (fmt, fmt_buf, b, b_buf, distp, sp)
/*-------------------------------------------------*/
	short   fmt,  fmt_buf [MAX_RANK];   /* 'fmt' is a 'DFMT_xxx' flag */
	Expr   *b,  *b_buf [MAX_RANK];      /* 'b' (blocksize) may be NULL */
	Distr  *distp;
	Symbol *sp;			/* may be NULL;  only for errmsgs */
{
	char  *name = (sp ? sp->name : NULL);
	int    prev_rank = (distp->attrs & DI_SRC_RANK_MASK);
					/* i.e. rank *before* this dimension */

	/*------------------------------------------------------*/	
	/* If (rank > MAX_RANK), give errmsg and ignore this	*/
	/* and subsequent dimensions.				*/
	/*------------------------------------------------------*/	
	if (prev_rank == MAX_RANK) {
		if ( ! (distp->attrs & DI_TOO_MANY_SRC_DIMS)) {
			/*--------------------------------------*/
			/* give errmsg, and record that it's	*/
			/* been given by setting flag		*/
			/* 'DI_TOO_MANY_SRC_DIMS'.		*/
			/*--------------------------------------*/
			begin_errmsg ((name ? "%q" : "distribution format list"), name);
			end_errmsg (" has too many dimensions - up to %d are allowed",
			             MAX_RANK);
			distp->attrs |= DI_TOO_MANY_SRC_DIMS;
		}
		/*--------------------------------------*/
		/* ...else errmsg already given.	*/
		/*--------------------------------------*/
		if (b)
			delete_expr (b, LEAVE_NLINKS);
		return;
	}

	distp->attrs |= DI_EXPL_FMT_LIST;

	distp->attrs++;		/* to store the current 'source' rank */

	if (fmt & (DFMT_BLOCK | DFMT_CYCLIC))
		distp->attrs += (1 << DI_PROC_RANK_SHIFT);
				/* to store the current 'proc_rank' -- the
				   # (distributed --not collapsed-- dims) */

	fmt_buf [prev_rank] = fmt;
	b_buf   [prev_rank] = b;	/* may be NULL */
	
	return;		
}
/*}}}*/

/*{{{  install_distr (sp, distp, fmt_buf, b_buf)  -- ok, but check*/
install_distr (sp, distp, fmt_buf, b_buf)
/*-------------------------------------*/
	Symbol *sp;		  /* NULL for DISTRIBUTE attr in 'attr-list' */
	Distr  *distp;
	short   fmt_buf [MAX_RANK];
	Expr   *b_buf [MAX_RANK];
{
	int      src_rank;
	short   *f_ip,  *ip,  *max_fmt_ip,  *t_ip;
	Expr   **f_epp,  **t_epp;
	Symbol  *proc_sp = distp->proc_sp;	/* may be NULL! */

	/*{{{  give errmsg if 'DISTRIBUTE *' used for non-dummy 'sp'*/
	if (sp && (distp->attrs & (DI_ASSERT | DI_INHERIT)) && ! (sp->attrs & AT_DUMMY))
	
		errmsg ("descriptive or transcriptive (i.e. \'*\') DISTRIBUTE specified for non dummy argument %q",
		         sp->name);
	/*}}}*/

	if ( ! (distp->attrs & DI_INHERIT_FMT)) {

		/*{{{  find 'src_rank', and clear 'DI_TOO_MANY_SRC_DIMS' if set*/
		/*--------------------------------------------------------------*/
		/* Find 'src_rank' and clear 'DI_TOO_MANY_SRC_DIMS' if set.	*/
		/*--------------------------------------------------------------*/
		src_rank = distp->attrs & DI_SRC_RANK_MASK;
		
		clear_flags (distp->attrs, DI_TOO_MANY_SRC_DIMS);
		
		/*}}}*/
		/*{{{  if 'src_rank' unknown, deduce it from procs rank if known*/
		/*----------------------------------------------------------------
		If (! src_rank)  (i.e. no format list was specified, so the
		DISTRIBUTE attr was:
		
			DISTRIBUTE [*] [ONTO [*] [P]]
		
		) then if a target processors arrangement 'P' is specified and its
		rank #P is known, set (src_rank = #P) and set the distribution
		format to 'DFMT_BLOCK' in each dimension.  [The latter is an
		arbitrary choice!]
			Signal that this has been done by setting the
		'DI_DEDUCED_FMT_LIST' flag in 'distp->attrs'.
		----------------------------------------------------------------*/
		if (! src_rank  &&  proc_sp  &&  proc_sp->arp) {
			int     dim;
		
			src_rank = proc_sp->arp->attrs & RANK_MASK;
		
			for (ip = fmt_buf,  dim = src_rank;   dim;   ip++,  dim--)
				*ip = DFMT_BLOCK;	/* an arbitrary choice! */
		
			distp->attrs |= (src_rank | (src_rank << DI_PROC_RANK_SHIFT) |
			                 DI_DEDUCED_FMT_LIST);
		}
		/*}}}*/
		/*{{{  allocate arrays 'distp->fmt [src_rank]' & 'b [src_rank]'*/
		/*--------------------------------------------------------------*/
		/* Allocate arrays 'distp->fmt [src_rank]' and 'b [src_rank]'.	*/
		/*								*/
		/* N.B. We zero the latter as its entries default to NULL.	*/
		/*--------------------------------------------------------------*/
		if (src_rank) {
			distp->fmt = (short *) emalloc (src_rank * sizeof (short));
			distp->b   = (Expr **) ecalloc (src_rank * sizeof (Expr *));
		}
		
		max_fmt_ip = distp->fmt + src_rank;
		/*}}}*/
		
		for (f_ip = fmt_buf,  t_ip = distp->fmt,  f_epp = b_buf,
		                                          t_epp = distp->b;
		     t_ip < max_fmt_ip;    ++f_ip, ++t_ip, ++f_epp, ++t_epp) {
		
			*t_ip = *f_ip;
			if (*f_epp)
				attach_expr (*t_epp, *f_epp);
		
			else if (*f_ip & DFMT_BLOCK)
				/*-------------------------------------------*/
				/* BLOCK distribution with default blocksize */
				/*-------------------------------------------*/
				distp->attrs |= (DI_NEED_BLOCKSIZE << (f_ip - fmt_buf));
		}
	}
	if (sp) {
		attach_distr (sp->distp, distp);
		if (sp->arp)
			check_distributee (sp);
	}
	if (proc_sp) {
		/*{{{  check 'proc_sp's type and rank if we have enough info*/
		/*----------------------------------------------------------------
			A distribution target 'proc_sp' was specified.  If we
		have enough information then check the following:
		
		-- that 'proc_sp' is a processors arrangement (setting flag
		'DI_CHECKED_PROC_TYPE' in 'distp->attrs' when this check is done);
		
		-- if it's an array, that the distribution format list or the
		distributee rank matches its rank.  (This check is performed by
		'check_distr_target ()', which sets flag 'DI_CHECKED_PROC_DIMS'
		in 'distp->attrs'.)
		
		N.B.:   If 'sp' is defined, then reset 'distp' to 'sp->distp'
		in case it was modified by 'check_distributee ()'.
		----------------------------------------------------------------*/
		if (sp)
			distp = sp->distp;
		
		if ((proc_sp->attrs & (AT_ENTITY | AT_PROCEDURE)) || (proc_sp->type & TY_TYPE)) {
		
			if ( ! (proc_sp->attrs & AT_PROCESSORS))
				errmsg ("DISTRIBUTE target %q is not a processors arrangement",
				         proc_sp->name);
		
			distp->attrs |= DI_CHECKED_PROC_TYPE;
		}		
		if (proc_sp->arp)
			check_distr_target (distp, sp);
		/*}}}*/
	}
}
/*}}}*/

/*{{{  delete_distr (distp,  decr_nlinks) -- ok*/
delete_distr (distp,  decr_nlinks)
/*------------------------------*/
	Distr  *distp;
	bool    decr_nlinks;
/*--------------------------------------------------------------*/
/*	Called by 'clear_symtab ()' and 'yyparse ()' to delete	*/
/* a 'Distr' record.						*/
/*								*/
/* N.B. We assume that (distp != NULL)!				*/
/*--------------------------------------------------------------*/
{
	Expr  **epp;
	int     dim,  src_rank;
	
	if (decr_nlinks)
		distp->nlinks--;

	if (distp->nlinks)
		/*----------------------------------------------*/	
		/* 'distp' is still in use, so don't delete it.*/
		/*----------------------------------------------*/
		return;

	src_rank = distp->attrs & DI_SRC_RANK_MASK;
	if (src_rank  &&  distp->fmt) {
		/*----------------------------------------------*/
		/* Delete any exprs attached to 'distp->b []',	*/
		/* and delete the arrays 'fmt []' and 'b []'	*/
		/* themselves.					*/
		/*						*/
		/* N.B. The above test checks that 'distp->fmt'	*/
		/* is allocated, in case this is called as a	*/
		/* result of a syntax error during the parsing	*/
		/* of a DISTRIBUTE attr, when (src_rank > 0) but*/
		/* before 'install_distr ()' has been called.	*/
		/*----------------------------------------------*/
		for (epp = distp->b, dim = src_rank;  dim;  --dim, ++epp)
			if (*epp)
				delete_expr (*epp, DECR_NLINKS);

		free  ((char *) distp->fmt);
		free  ((char *) distp->b);
	}

	free  ((char *) distp);
}
/*}}}*/
/*{{{  Distr  *copy_distr (distp)  -- ok*/
Distr  *copy_distr (distp)
/*-----------------------*/
	Distr  *distp;
/*--------------------------------------------------------------*/
/*	Called by 'check_distributee ()' or			*/
/* 'calc_default_blocksizes ()' if they have to modify a 'distp'*/
/* record that has (nlinks > 1).				*/
/*--------------------------------------------------------------*/
{
	Distr  *new_distp;
	Expr  **f_epp,  **t_epp;
	int     dim,  src_rank;
	short  *f_ip,  *t_ip;
	
	new_distp = create_distr ();

	new_distp->proc_sp = distp->proc_sp;
	new_distp->attrs   = distp->attrs;
	new_distp->nlinks  = 0;		/* not linked to anything yet */

	if (src_rank = (distp->attrs & DI_SRC_RANK_MASK)) {
		/*{{{  allocate and copy 'distp->fmt []' and 'b []'*/
		/*--------------------------------------------------------------*/
		/* Allocate and copy arrays 'distp->fmt []' & 'b []'.		*/
		/*								*/
		/* N.B. We zero the latter as its entries are NULL by default.	*/
		/*--------------------------------------------------------------*/
		new_distp->fmt = (short *) emalloc (src_rank * sizeof (short));
		new_distp->b   = (Expr **) ecalloc (src_rank * sizeof (Expr *));
		
		for (f_ip = distp->fmt,  t_ip = new_distp->fmt,  f_epp = distp->b,
		                         t_epp = new_distp->b,  dim = src_rank;
		     dim;    --dim,  ++f_ip,  ++t_ip,  ++f_epp,  ++t_epp) {
		
			*t_ip = *f_ip;
			if (*f_epp)
				attach_expr (*t_epp, *f_epp);
		}
		/*}}}*/
	}

	return  new_distp;
}
/*}}}*/

/*{{{  check_distributee (sp)  -- OK!*/
check_distributee (sp)
/*------------------*/
	Symbol  *sp;		/* has 'sp->distp' defined! */
/*----------------------------------------------------------------
This is called by:
-- 'install_distr (sp,...)', after installing distribution info ('distp')
	in an ID 'sp' that has array dimensions ('sp->arp');
-- 'yyparse()', after attaching distribution info ('distp') declared
	by an ALIGN attr to an ID 'sp' that has array dimensions.
-- 'install_array (sp,...)', after installing array dimensions ('arp')
	in an ID 'sp' that has distribution info ('sp->distp');
-- 'process_decls ()', when it sees a scalar distributee 'sp'.

It does the following:
-- checks that the format list has the correct rank (whether
	given explicitly or deduced from the rank of the target
	processor array), and if not gives an errmsg and adjusts
	the distribution info so that it does;
-- supplies any missing default blocksizes in 'distp->b []',
	provided that 'check_distr_target ()' has already been called.

N.B. If any modifications are made to 'sp->distp' they are specific
to this distributee, so a new copy is made if  (sp->distp->nlinks > 1).
----------------------------------------------------------------*/
{
	int     actual_src_rank,  dim,  src_rank;
	Distr  *distp = sp->distp;
	
	src_rank        = (distp->attrs & DI_SRC_RANK_MASK);
	actual_src_rank = sp->arp ? (sp->arp->attrs & RANK_MASK) : 0;

	if (actual_src_rank != src_rank) {
		/*{{{  give errmsg & fix 'distp' to correct #(dims) in 'fmt[]'*/
		/*--------------------------------------------------------------*/
		/* Wrong rank is specified or deduced for the distributee in	*/
		/* DISTRIBUTE.  Give an errmsg and 'fix' 'sp->distp' to have	*/
		/* the correct number of dims in 'fmt []' & 'b []'.		*/
		/*								*/
		/* N.B.								*/
		/* ----								*/
		/* -- Since we're about to modify 'sp->distp', make a new copy	*/
		/* of it if (sp->distp->nlinks > 1).				*/
		/* -- This may change the value of 'proc_rank' (the #(BLOCK /	*/
		/* CYCLIC entries in 'fmt []')).  If so then we reset 'proc_sp'	*/
		/* to point to the default processor array of the appropriate	*/
		/* rank (viz. "$pN", where N = proc_rank).			*/
		/*--------------------------------------------------------------*/
		Expr  **f_epp,  **new_b,  **t_epp;
		short  *f_ip,  *new_fmt,  *t_ip;
		
		if (distp->nlinks > 1) {
			distp->nlinks--;		/* as it will be detached from 'sp' */
			distp = sp->distp = copy_distr (distp);
			distp->nlinks = 1;		/* as it's only attached to 'sp' */
		}
		
		/*{{{  print an errmsg*/
		/*{{{  << EXPLANATION >>*/
		/*----------------------------------------------------------------
		<< EXPLANATION >>
		-----------------
		Giving an appropriate errmsg is quite complicated!
			The possible situations are as follows,  where S = source,
		P = target procs,  EF = EXPL_FMT_LIST,  DF = DEDUCED_FMT_LIST,
		DP = DEDUCED_PROC_RANK,  #S & #P = rank of S and P:
		
		case	DISTRIBUTE directive / attr	    flags;	(#dims)
		~~~~	~~~~~~~~~~~~~~~~~~~~~~~~~~~	    ~~~~~~~~~~~~~~~~~~~
		after 'install_distr()'
		----------------------
		1:	DISTRIBUTE [S][*] (fmt-list) ...  { [EF]	(#S > 0);
		
		2a: }	DISTRIBUTE [*] ONTO [*] P ::	  { [DF]  if #P known before DISTRIBUTE
		    }					  {          and not INHERIT_FMT:
		    }					  {		(#S == #P > 0);
		2b: }					  { [-]   otherwise
		    }					  {		(#S == #P == 0) for now
		
		3:	DISTRIBUTE [*] ONTO * ::	    [-]		(#S == #P == 0) for now
		
		after 'check_distr_target ()'
		-----------------------------
		4:	DISTRIBUTE [*] ONTO [*] P ::	  { if (not INHERIT_FMT && #P > 0)
							  {	converts [-]  to  [DF]
							  {	sets (#S == #P > 0)
		Errmsgs if (src_rank != actual_src_rank):
		~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		if (actual_src_rank > 0)
		------------------------
		if (EF set)		-- case 1   (DISTRIBUTE [S] (fmt-list) ...)
			"SRC has too many/few dimensions in DISTRIBUTE directive"
		
		else if (DF set)	-- cases 2a & 4  (DISTRIBUTE ONTO [*] P   ! #P known)
		
			"SRC does not have the same rank as PROC, onto which it is distributed"
		
		else			-- cases 2b & 3,  i.e. no flags set
					               (DISTR [*] ONTO [*] [P]   ! #P unknown)
			-- No errmsg, since #S could be anything but wasn't established
				before now.
			-- Set #S = actual_src_rank.
			-- if (P defined and not INHERIT_FMT)  set (#P == #S) and set [DP].
			N.B.:  'check_distr_target ()' still has to be called, and
			will give an errmsg if (not INHERIT_FMT  &&  #P != #S).
			[We know this because if it had already been called then
			 the [DF] flag would have been set!]
		
		else    -- i.e. (actual_src_rank == 0)
		--------------------------------------
		if (EF set)		-- case 1   (DISTRIBUTE [S] (fmt-list) ...)
			"scalar object SRC has a distribution format list in a DISTRIBUTE directive"
		
		else if (DF set)	-- cases 2a & 4  (DISTRIBUTE ONTO [*] P)
			"scalar SRC cannot be distributed onto processor array PROC  (it
		can only be distributed onto a scalar processor)"
		
		-- else cases 2b or 3 -- no flags set ([-]), which is OK!
		----------------------------------------------------------------*/
		/*}}}*/
		
		if (actual_src_rank > 0) {
			if (distp->attrs & DI_EXPL_FMT_LIST)		/* case 1 */
		
				errmsg ("%q has too %s dimensions in DISTRIBUTE directive",
				         sp->name, ((src_rank > actual_src_rank) ? "many" : "few"));
		
			else if (distp->attrs & DI_DEDUCED_FMT_LIST)	/* cases 2a & 4 */
		
				errmsg ("%q does not have the same rank as %q, onto which it is distributed",
				         sp->name,  distp->proc_sp->name);
		
			else {		/* cases 2b or 3, i.e. no flags set ([-]) */
		
				/*{{{  no errmsg;  update info in 'distp' & return*/
				/*----------------------------------------------------------------
				The DISTRIBUTE directive / attr was either:
				
					DISTRIBUTE ONTO [*] [P] :: ... S
				
				where #S was unknown in 'install_distr ()' and so was set to 0, or:
				
					DISTRIBUTE * ONTO [*] [P] :: ... S
				
				(the 'INHERIT_FMT' case) in which case 'install_distr ()' did
				nothing, leaving it for this function to handle.
				
					Now the rank of S, #S, is known, so:
				if (INHERIT_FMT) {
					-- set (#S = actual_src_rank);
					-- set {fmt [i] = DFMT_INHERIT,    0 <= i < #S}
					       {  b [i] = "mapvec (..)",       "     }
				}
				else {
					-- set (#S = actual_src_rank);
					-- set {fmt [i] = DFMT_BLOCK,  0 <= i < #S}    [...this is arbitrary !];
					-- if P is defined, set (#P == #S) and set flag [DP] ('DEDUCED_PROC_RANK')
				}
				N.B.  'check_distr_target ()' still has to be called, and
				will give an errmsg if (#P != #S).  [We know this because if it
				had already been called then the [DF] flag would have been set!]
				----------------------------------------------------------------*/
				/*--------------------------------------------------------------*/
				/* Allocate arrays 'distp->fmt [src_rank]' and 'b [src_rank]'.	*/
				/*--------------------------------------------------------------*/
				f_ip  = distp->fmt = (short *) emalloc (actual_src_rank * sizeof (short));
				f_epp = distp->b   = (Expr **) ecalloc (actual_src_rank * sizeof (Expr *));
				
				/*----------------------------------------------------------------*/
				/* Initialise 'distp->fmt / b []', and set flags in 'distp->attrs'*/
				/*----------------------------------------------------------------*/
				if (distp->attrs & DI_INHERIT_FMT) {
					for (dim = 0;   dim < actual_src_rank;   ++dim) {
						*f_ip++  = DFMT_INHERIT;
						*f_epp++ = map_primary (sp->name, 0, map_subscr_b (actual_src_rank, dim));
					}
					distp->attrs |= actual_src_rank;
				}
				else {
					for (dim = 0;   dim < actual_src_rank;   ++dim) {
						*f_ip++ = DFMT_BLOCK;
						distp->attrs |= (DI_NEED_BLOCKSIZE << dim);
					}
					distp->attrs |= (actual_src_rank | (actual_src_rank << DI_PROC_RANK_SHIFT));
					if (distp->proc_sp)
						distp->attrs |= DI_DEDUCED_PROC_RANK;
				}
				return;		/* since all done now! */
				/*}}}*/
			}
		}
		else {		/* i.e. (actual_src_rank == 0) */
			if (distp->attrs & DI_EXPL_FMT_LIST)		/* case 1 */
		
				errmsg ("scalar object %q has a distribution format list in a DISTRIBUTE directive",
				         sp->name);
		
			else		/* cases 2a & 4, i.e. DF set */
		
				errmsg ("scalar %q cannot be distributed onto processor array %q  (it can only be distributed onto a scalar processor)",
				         sp->name,  distp->proc_sp->name);
		}
		/*}}}*/
		
		distp->attrs += (actual_src_rank - src_rank);	/* .. to correct 'src_rank' */
		
		if (actual_src_rank > src_rank) {
			/*{{{  make new 'distp->fmt[]' and 'b[]' with extra collapsed dims--OK*/
			/*--------------------------------------------------------------*/
			/* (actual_src_rank > src_rank).				*/
			/*	Create new arrays 'distp->fmt []' and 'b []'.  Copy the	*/
			/* old entries to them, and make the required extra dimensions	*/
			/* collapsed ones.						*/
			/*--------------------------------------------------------------*/
			/*--------------------------------------------------------------*/
			/* Allocate arrays 'new_fmt []' and 'new_b []'.			*/
			/*								*/
			/* N.B. We zero the latter as its entries are NULL by default.	*/
			/*--------------------------------------------------------------*/
			new_fmt = (short *) emalloc (actual_src_rank * sizeof (short));
			new_b   = (Expr **) ecalloc (actual_src_rank * sizeof (Expr *));
			
			/*-------------------------*/
			/* Initialise 'new_fmt []' */
			/*-------------------------*/
			for (f_ip = distp->fmt,  t_ip = new_fmt,  dim = 0;   dim < src_rank;   ++dim)
				*t_ip++ = *f_ip++;
			
			for (  ;   dim < actual_src_rank;   ++dim)
				*t_ip++ = DFMT_COLLAPSED;
			
			/*--------------------------------------------------------------*/
			/* Initialise 'new_b []'.  (N.B. the extra elements are left as	*/
			/* NULL, following our convention for collapsed dims).		*/
			/*--------------------------------------------------------------*/
			for (f_epp = distp->b,  t_epp = new_b,  dim = 0;    dim < src_rank;    ++dim)
				*t_epp++ = *f_epp++;
			
			/*--------------------------------------------------------------*/
			/* Delete the old arrays and attach the new ones to 'distp'	*/
			/*--------------------------------------------------------------*/
			if (src_rank) {
				free  ((char *) distp->fmt);
				free  ((char *) distp->b);
			}
			distp->fmt = new_fmt;
			distp->b   = new_b;
			/*}}}*/
		}
		else {
			/*{{{  delete superfluous dims in 'distp->fmt []' and 'b []' -- OK*/
			/*----------------------------------------------------------------
			(actual_src_rank < src_rank).
				Delete the superfluous entries in 'distp-> b []'.
			N.B.:
			-----
			--  We *delete* the extra 'distp->b []' entries because they're
			exprs.  If we didn't, it won't be done by 'delete_distr ()',
			so they would remain as garbage.
			
			-- If the reduced value of 'src_rank' (viz. 'actual_src_rank')
			causes *distributed* dims to be dropped from 'distp->fmt []',
			then we reduce 'proc_rank' accordingly and reset 'proc_sp' to
			point to the default processor array of the appropriate rank
			(viz. "$pN", where N = proc_rank).
			----------------------------------------------------------------*/
			int   new_proc_rank,  old_proc_rank;
			
			/*----------------------------------------------*/
			/* Delete unwanted entries in 'distp->b []'.	*/
			/*----------------------------------------------*/
			for (f_epp = distp->b + actual_src_rank,  dim = actual_src_rank;
			     dim < src_rank;    ++dim,  ++f_epp) {
				if (*f_epp)
					delete_expr (*f_epp, DECR_NLINKS);
				else
					clear_flags (distp->attrs, (DI_NEED_BLOCKSIZE << dim));
			}
			
			/*--------------------------------------------------------------*/
			/* If necessary, reduce 'proc_rank' and reset 'proc_sp' to	*/
			/* point to the default processor array of the appropriate rank	*/
			/* (viz. "$pN", where N = proc_rank).				*/
			/*--------------------------------------------------------------*/
			new_proc_rank = old_proc_rank =
			                 (distp->attrs & DI_PROC_RANK_MASK) >> DI_PROC_RANK_SHIFT;
			
			for (f_ip = distp->fmt + actual_src_rank,  dim = actual_src_rank;
			     dim < src_rank;    ++dim,  ++f_ip) {
			
				if (*f_ip & (DFMT_BLOCK | DFMT_CYCLIC)) {
					distp->attrs -= (1 << DI_PROC_RANK_SHIFT);
					new_proc_rank--;
				}
			}
			if (new_proc_rank != old_proc_rank)
				distp->proc_sp = default_proc_sp [new_proc_rank];
			
			/*--------------------------------------------------------------*/
			/* If the distributee is actually a scalar, delete arrays	*/
			/* 'distp->fmt []' and 'b []' themselves.			*/
			/*--------------------------------------------------------------*/
			if ( ! actual_src_rank) {
				free  ((char *) distp->fmt);
				free  ((char *) distp->b);
			}
			/*}}}*/
		}
		/*}}}*/
	}
}
/*}}}*/
/*{{{  check_distr_target (distp, sp) -- ok*/
check_distr_target (distp, sp)
/*--------------------------*/
	Distr   *distp;		/* ...which has 'distp->proc_sp' defined! */
	Symbol  *sp;		/* distributee -- may be NULL; for errmsgs only */
/*----------------------------------------------------------------
This is called by:
-- 'install_distr ()', when installing distribution info ('distp')
	where the target processor array's dimensions are already known;
-- 'process_decls ()', when it sees a 'distp' whose target processor
	arrangement wasn't 'checked out' as above (in which case the
	flag 'DI_CHECKED_PROC_DIMS' isn't set in 'distp->attrs').

It does the following:
-- checks that the target processor arrangement has the correct rank
	(whether derived from an explicit format list or deduced from
	the source rank), and if not gives an errmsg and changes
	the distribution format list or 'proc_sp' so that it does;
-- supplies any missing default blocksizes in 'distp->b []',
	provided that 'check_distributee ()' has already been called.

N.B.  If any modifications are made to 'distp' they apply to *all*
alignees that 'distp' is attached to, so we *don't* copy it if
(distp->nlinks > 1).  (This is different from 'check_distributee ()').
----------------------------------------------------------------*/
{
	int      actual_proc_rank,  dim,  proc_rank;
	short   *ip;
	Symbol  *proc_sp = distp->proc_sp;	/* not NULL! */

	proc_rank        = (distp->attrs & DI_PROC_RANK_MASK) >> DI_PROC_RANK_SHIFT;
	actual_proc_rank = proc_sp->arp ? (proc_sp->arp->attrs & RANK_MASK) : 0;

	if (actual_proc_rank != proc_rank) {
		/*{{{  give errmsg & 'fix' 'distp' to have correct #(target dims)*/
		/*----------------------------------------------------------------
		Either:
		 (i) the format list in the DISTRIBUTE directive / attr implies an
		incorrect rank for the distribute target  (if flag [EF] set);  or
		
		 (ii) there wasn't a format list, and the target rank was set equal
		to the known source rank, but in fact it's not equal  ([DP] set); or
		
		 (iii) there wasn't a format list, and the target rank was unknown,
		so it was (temporarily) assumed to be 0.
		
			'Fix' 'distp' so that (proc_rank == # (BLOCK / CYCLIC
		entries in 'fmt []') == actual rank of 'proc_sp'), and in cases
		(i) & (ii) give an errmsg.
		
		N.B.  In this case, although we're about to modify 'distp', we
		don't make a copy of it if (distp->nlinks > 1) as these changes
		apply to *all* distributees that 'distp' is attached to.  (This is
		different from 'check_distributee ()').
		----------------------------------------------------------------*/
		/*{{{  print errmsg*/
		/*{{{  << EXPLANATION >>*/
		/*----------------------------------------------------------------
		<< EXPLANATION >>
		-----------------
		Giving an appropriate errmsg is quite complicated!
			The possible situations are as follows (where S = source,
		P = target processors,  EF = EXPL_FMT_LIST,  DF = DEDUCED_FMT_LIST,
		DP = DEDUCED_PROC_RANK,  #S & #P = rank of S and P):
		
		case	DISTRIBUTE directive / attr	    flags;	(#dims)
		~~~~	~~~~~~~~~~~~~~~~~~~~~~~~~~~	    ~~~~~~~~~~~~~~~~~~~
		after 'install_distr()'
		----------------------
		1:	DISTRIBUTE [S][*] (fmt-list) ...    [EF]	(#S > 0);
		
		2a: }	DISTRIBUTE [*] ONTO [*] P :: ...S { [DF]  if #P known before DISTRIBUTE
		    }					  {          and not INHERIT_FMT:	
		    }					  {		(#S == #P > 0);
		2b: }					  { [-]   otherwise
		    }					  {		(#S == #P == 0) for now
		
		after 'check_distributee ()'
		----------------------------
		3:	DISTRIBUTE [*] ONTO [*] P :: ...S { if (not INHERIT_FMT && #S > 0)
							  {	converts [-]  to  [DP]
							  {	sets (#P == #S)
							  
		Errmsgs if (proc_rank != actual_proc_rank):
		~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		N.B.:
		-----
		--  Case (2a) needn't be considered, as it implies that #P is correct.
		
		-- In case (1), this could be called from 'install_distr ()' with
		distributee 'sp' undefined (i.e. NULL), e.g.:
			DISTRIBUTE [*] (fmt-list) onto P ...   ! with #P known.
		This must be borne in mind when generating errmsgs.
		
		if (actual_proc_rank > 0)
		-------------------------
		if (EF set)		-- case 1   (DISTRIBUTE [S][*] (fmt-list) ...)
			"DISTRIBUTE mismatch: %d dimensions [of SRC] are distributed onto
		%d-dimensional processor array PROC"
		
			-- N.B. This message isn't ideal if SRC is known and 'fmt-list'
			has a different rank from it, since then the msg will claim
			that N dims of SRC are distributed when SRC may actually have
			< N dims, and could even be scalar.
				Still, it shouldn't be too confusing.  E.g., if SRC
			is actually scalar then that's reported *after* this msg.
			Another consideration is that, if this is called from
			'process_decls ()', then we must mention *some* SRC in order
			to locate which DISTRIBUTE directive is being referred to!
		
		else if (DP set)	-- case 3   (DISTRIBUTE ONTO [*] P :: ...S  ! #S known)
			"DISTRIBUTE mismatch: %d-dimensional array SRC is distributed onto
		%d-dimensional processor array PROC"
		
		else		-- case 2b,   i.e. no flags set
				                  (DISTRIBUTE ONTO [*] P :: ...S  ! #S unknown)
			-- No errmsg, since (#S == #P) could be anything, but wasn't
				established before now.
			-- Set  (#S = #P = actual_proc_rank), and set [DF].
			-- 'check_distributee ()' will be called from 'process_decls()'
			immediately after this returns, and will give an errmsg
			since in fact (#S == 0) rather than (== #P > 0).
				[Why?  [-] means that #P was unknown in 'install_distr()',
			so this must be called from 'process_decls()'.  The fact that it's
			still [-] means that S hasn't been declared as an array --
			otherwise 'check_distributee()' would have changed it to [DP].
			Hence S is scalar, so 'check_distributee()' is called
			immediately after this and will complain about the rank mismatch.
				N.B. An advantage of this way of organising the errmsgs
			is that a separate errmsg will be given for *every* scalar
			distributee with this 'distp'.]
		
		else    -- i.e. (actual_proc_rank == 0)
		---------------------------------------
		N.B.:
		-- This must be called from 'process_decls ()', as it's only called
		from 'install_distr ()' if (#P > 0).  Hence distributee 'sp' is defined.
		
		-- (proc_rank > 0), so either [EF] or [DP] must be set.  In the
		former case (#S >= 0); in the latter, (#S > 0), where #S is the
		*actual* source rank.
			If (#S == 0) we choose not to give an errmsg here, as
		the following errmsg is given anyway by 'check_distributee()',
		which is called immediately after this returns:
		
			"scalar object SRC has a distribution format list in a DISTRIBUTE directive"
		
		There's no point giving 2 errmsgs for 1 error!
			Hence, we only give an errmsg if (#S > 0).
		
		
		if (#S > 0)	-- cases 1 & 3   (DISTRIBUTE [...] ONTO P)
		
			"in DISTRIBUTE directive, all dimensions of SRC must be collapsed
		(i.e. specified as '*') onto scalar processor arrangement PROC"
		----------------------------------------------------------------*/
		/*}}}*/
		
		if ((actual_proc_rank > 0)  &&  ! (distp->attrs &
		                         (DI_EXPL_FMT_LIST | DI_DEDUCED_PROC_RANK)) ) {
						/* case 2b, i.e. no flags set ([-]) */
		
			/*{{{  no errmsg;  update info in 'distp' & return*/
			/*----------------------------------------------------------------
			The DISTRIBUTE attr was:
			
				DISTRIBUTE ONTO [*] P ::  ... S
			
			where (#S == #P) was unknown in 'install_distr ()', so was set to 0.
				Now the rank of P, #P, is known, so do the same as would
			be done by 'install_distr ()' if #P were known earlier, viz:
			
			-- set (src_rank = #P = actual_proc_rank),
			-- set the distribution format to 'DFMT_BLOCK' in each dimension
			[this is an arbitrary choice!]
			-- set [DF].
			
			N.B.  'check_distributee ()' will be called from 'process_decls()'
			immediately after this returns, and will give an errmsg since in fact
			(#S == 0) rather than (== #P > 0).
				[Why?  Because [-] means that #P was unknown in
			'install_distr()', so this must be called from 'process_decls()'.
			The fact that it's still [-] means that S hasn't been declared as
			an array -- otherwise 'check_distributee()' would have changed
			it to [DP].  Hence S is scalar, so 'check_distributee()' is
			called immediately after this, and will complain about the rank mismatch].
			----------------------------------------------------------------*/
			if (distp->attrs & DI_INHERIT_FMT) {
				/*------------------------------------------------------*/
				/* Set (proc_rank = actual_proc_rank) in 'distp->attrs'.*/
				/*------------------------------------------------------*/
				distp->attrs |= ((actual_proc_rank << DI_PROC_RANK_SHIFT) |
				                 DI_CHECKED_PROC_DIMS);
			}
			else {
				/*------------------------------------------------------*/
				/* Allocate arrays 'distp->fmt []' & 'b []'.		*/
				/*							*/
				/* N.B. We zero the latter as its entries start out NULL*/
				/*------------------------------------------------------*/
				distp->fmt = (short *) emalloc (actual_proc_rank * sizeof (short));
				distp->b   = (Expr **) ecalloc (actual_proc_rank * sizeof (Expr *));
			
				for (ip = distp->fmt,  dim = 0;   dim < actual_proc_rank;   ++dim,  ++ip) {
					*ip = DFMT_BLOCK;		/* an arbitrary choice! */
					distp->attrs |= (DI_NEED_BLOCKSIZE << dim);
				}
			
				/*------------------------------------------------------*/
				/* Set (src_rank = proc_rank = actual_proc_rank), and	*/
				/* [DF] flag, in 'distp->attrs'.			*/
				/*------------------------------------------------------*/
				distp->attrs |= (actual_proc_rank | (actual_proc_rank << DI_PROC_RANK_SHIFT) |
				                 DI_DEDUCED_FMT_LIST | DI_CHECKED_PROC_DIMS);
			}
			return;		/* since all done now! */
			/*}}}*/
		}
		else if ( (distp->attrs & DI_CHECKED_PROC_TYPE)  &&
		        ! (distp->proc_sp->attrs & AT_PROCESSORS) )
		
			/*------------------------------------------------------*/
			/* An errmsg has already been given because 'proc_sp'	*/
			/* *isn't* a processor arrangement.  Don't bother to	*/
			/* give another one!					*/
			/*------------------------------------------------------*/
			;
		
		else if (actual_proc_rank > 0) {
			if (distp->attrs & DI_EXPL_FMT_LIST) {		/* case 1 */
		
				begin_errmsg ("DISTRIBUTE mismatch: %d dimension%c ",
				               proc_rank,  ((proc_rank == 1) ? '\0' : 's'));
				if (sp)
					cont_errmsg ("of %q ", sp->name);
		
				end_errmsg ("%s distributed onto %d-dimensional processor array %q",
				            ((proc_rank == 1) ? "is" : "are"),
				            actual_proc_rank,  proc_sp->name);
			}
			else if (distp->attrs & DI_DEDUCED_PROC_RANK)	/* case 3 */
		
				errmsg ("DISTRIBUTE mismatch: %d-dimensional array %q is distributed onto %d-dimensional processor array %q",
				         proc_rank,  sp->name,  actual_proc_rank,
				         proc_sp->name);
		}
		else if (sp->arp)	/* i.e. (actual_proc_rank == 0);  cases 1 & 3 */
		
			errmsg ("in DISTRIBUTE directive, all dimensions of %q must be collapsed (i.e. specified as \'*\') onto scalar processor arrangement %q",
			         sp->name,  proc_sp->name);
		/*}}}*/
		
		if (actual_proc_rank > proc_rank) {
			/*{{{  change 'proc_sp' to the default proc array with 'proc_rank' dims*/
			/*----------------------------------------------------------------
			(actual_proc_rank > proc_rank).
				In this case we just reset 'proc_sp' to point to the
			default processor arrangement with rank 'proc_rank'  (viz.
			"$pN", where N = proc_rank).
			
			N.B.  We choose not to attempt the possible alternative approach
			of trying to change enough 'DFMT_COLLAPSED' entries in
			'distp->fmt []' to 'DFMT_BLOCK' to make the format list have
			'actual_proc_rank' distributed dims.  We don't do it since
			**IT MAY BE IMPOSSIBLE**, as the distributee may have fewer dims
			than the proc target, in which case it doesn't have enough dims
			to distribute!
			----------------------------------------------------------------*/
			
			distp->proc_sp = default_proc_sp [proc_rank];
			/*}}}*/
		}
		else if (actual_proc_rank < proc_rank) {
			/*{{{  collapse enough distributed dims to get 'proc_rank' right -- OK*/
			/*----------------------------------------------------------------
			(actual_proc_rank < proc_rank).
				Convert enough distributed dims  (i.e. 'DFMT_BLOCK /_CYCLIC'
			entries in 'distp->fmt []')  to collapsed ones (i.e. 'DFMT_COLLAPSED')
			to make the format list have the correct number of distributed dims.
			----------------------------------------------------------------*/
			int     surplus_distr_dims;
			Expr  **epp;
			
			surplus_distr_dims = proc_rank - actual_proc_rank;
			
			for (ip = distp->fmt, epp = distp->b;   surplus_distr_dims;   ip++, epp++) {
			
				if (*ip & (DFMT_BLOCK | DFMT_CYCLIC)) {
					/*----------------------------------------------*/
					/* a distributed dim -- change it to collapsed	*/
					/*----------------------------------------------*/
					*ip = DFMT_COLLAPSED;
					if (*epp) {	/* delete the explicit blocksize */
						delete_expr (*epp, DECR_NLINKS);
						*epp = NULL_Expr;
					}
					else
						clear_flags (distp->attrs,
						            (DI_NEED_BLOCKSIZE << (ip - distp->fmt)) );
			
					distp->attrs -= (1 << DI_PROC_RANK_SHIFT);
					surplus_distr_dims--;
				}
			}
			/*}}}*/
		}
		/*}}}*/
	}
	distp->attrs |= DI_CHECKED_PROC_DIMS;
}
/*}}}*/

/*{{{  calc_default_blocksizes (sp) -- ok*/
calc_default_blocksizes (sp)
/*------------------------*/
	Symbol  *sp;		/* has 'sp->distp' defined! */
/*----------------------------------------------------------------
	'sp->distp->attrs' has some flags:

	(DI_NEED_BLOCKSIZE << dim),  0 <= dim < src_rank

set, indicating that dimension 'dim' is BLOCK distributed with
default blocksize, and that the latter hasn't been evaluated yet,
i.e. (b [dim] == NULL).  For every such dimension, this function
evaluates its blocksize and attaches it to 'distp->b [dim]'.
	The default blocksize is:

	b = ceil (n / np)
	  = (n + np - 1) / np
where:
	n  = extent of dimension 'dim' of 'sp' (= sp->arp->n [dim]);
	np = extent of the proc array dimension over which 'dim'
	              is distributed.
Hence we need both 'sp->arp' and 'proc_sp->arp'.

N.B.
----
--  This is called by 'process_decls ()' after both 'check_distributee ()'
and 'check_distr_target ()' have been called for 'sp', and only if
some '(DI_NEED_BLOCKSIZE << dim)' flags are set.  Hence we know
that 'sp->arp' and 'proc_sp->arp' are both defined, and that
*some* blocksizes 'distp->b [dim]' must be supplied.

--  These default blocksizes are specific to 'sp's array shape,
so a new copy of 'distp' is made if (sp->distp->nlinks > 1).
----------------------------------------------------------------*/
{
	/*{{{  decls*/
	Distr   *distp = sp->distp;
	int32    attrs = distp->attrs;
	int      dim,  src_rank = (attrs & DI_SRC_RANK_MASK);
	short   *fmt_p;
	Expr   **b_p,  **n_p,  **np_p;
	/*}}}*/

	/*{{{  make a new copy of 'distp' if its (nlinks > 1)*/
	/*------------------------------------------------------*/
	/* Make a new copy of 'sp->distp' if its (nlinks > 1)	*/
	/*------------------------------------------------------*/
	if (distp->nlinks > 1) {
		distp->nlinks--;		/* as it will be detached from 'sp' */
		distp = sp->distp = copy_distr (distp);
		distp->nlinks = 1;		/* as it's only attached to 'sp' */
	}
	/*}}}*/

	b_p   = distp->b;		/* blocksize */
	fmt_p = distp->fmt;		/* distribution format */
	n_p   = sp->arp->n;		/* array extent 'n' */
	np_p  = distp->proc_sp->arp->n;	/* proc array extent 'np' */

	for (dim = 0;    dim < src_rank;    ++dim,  ++fmt_p,  ++b_p,  ++n_p) {
		if (attrs & (DI_NEED_BLOCKSIZE << dim))
			attach_expr (*b_p,  ceiling_div (*n_p,  *np_p));

	        if (*fmt_p & (DFMT_BLOCK | DFMT_CYCLIC))
			++np_p;
	}
	clear_flags (distp->attrs,  DI_NEED_BLOCKSIZE_MASK);
}
/*}}}*/
/*}}}*/

/*{{{  <<< HPF 'Map' functions >>>*/
/*--------------------------------------------------------------*/
/* HPF 'Map' handling functions.				*/
/*								*/
/*** N.B. I don't think 'install_id_map ()' handles (INHERIT	*/
/* accompanied by DISTRIBUTE) properly! *****			*/
/*--------------------------------------------------------------*/

/*{{{  install_id_map (sp) -- ok  (but ignores DISTRIBUTE with INHERIT!)*/
install_id_map (sp)
/*---------------*/
	Symbol  *sp;		/* has (sp->alignp || sp->distp) defined */
/*{{{  EXPLANATION  *** MODIFY THIS!*/
/*------------------------------------------------------------------
	Create and initialise the 'mapping vector' ('sect_e->mapp')
of an expression node 'sect_e' corresponding to a named array,
array section or element.  (I.e 'sect_e->val.sp' is the symbol
table entry of an array name, and 'sect_e->right' may point
to the expression tree of a subscript list).

**** Q1: WHAT ABOUT ARRAY FUNCTION, ARRAY CONSTRUCTOR, ETC??

N.B.
---
-- It's assumed that 'sect_e->dvec' doesn't exist yet!
-- We don't delete the subscript expressions, even though they're
not directly used after 'Map' has been installed, as they're involved
in the expressions in 'Map' (i.e. the 'Map' expressions contain links
to them!).  All these expressions are deleted after the statement
has been processed and output.

Discussion
----------
	Recall that 'dvec' = (n, d, sectn, sectl, sects).
These parameters are NULL only if their values are arbitrary (i.e.
if they can be anything!).  They *aren't* NULL simply because they're
non-constant.  In particular, they aren't NULL if 'array_name' is
assumed shape or distribution;  instead, exprs involving 'i$at_name(..)'
are stored.  This is necessary, for otherwise no information can be
stored about the section parameters in 'sectn, l, s' (as they may be
functions of 'n' -- e.g. for subscript (:), sectn = n).  Also, the
parameters of different terms in an expression may be comparable,
or even equal, even though they're expressed in terms of variables
like 'i$at_' (e.g. 2 terms may both have n = i$at_A(10)).
	(Cf. 'sp' contains values of (n, lb and d) for the array,
where 'n' and 'd' are NULL for assumed shape and distribution.
This is desirable for 'sp', to distinguish assumed shape dummy
arrays (n = NULL) from local arrays whose shape is the same as
that of an assumed shape dummy (when an explicit expr is stored,
e.g. n = i$at_D(10))).

Parameters of a whole array (from which section is selected) are:

(i) Explicit-shape, or allocated array
--------------------------------------
sp->arp->n, sp->distp->d

Since contiguous storage, dvec->sectn, l, s are derived from these, viz:
sectl = 0, sectn = n, sects = 1.

(ii) (Associated) pointer array
-------------------------------
Parameters of (associated) pointer array are stored in the 'Expr' node
for the associated (sub)array, viz:

sp->info.init->dvec->n, d, sectn, sectl, sects

(The first 2 of these are also stored in, and hence identical to:
sp->arp->n,  sp->distp->d)

(iii) Assumed-shape
-------------------
Parameters of assumed-shape array are:

sp->arp->n,  sp->distp->d.
The sectional parameters are derived automatically (as indeed are these
also(!)) -- they're just "i$at_name(..)" elements!.

Unallocated array, unassociated pointer
---------------------------------------
Clearly, this has no parameters, & can't take section of it!

Q: What to do in this case??  Maybe should delete whole 'sect_e'
and just replace it by '1' -- so it plays no further part, and causes no
other problems.  Yep -- do it here.
But we also want to do some checks here (e.g. shape conforms??).

------------------------------------------------------------------*/
/*}}}*/
{
/******************************************
N.B.  *** This ignores 'DISTRIBUTE' accompanying an 'INHERIT' !!
******************************************/

	/*** char  *name = sp->name;	****/
	/*{{{  decls*/
	bool     aligned_with_inherit,  inherited;
	int      array_rank,  dim,  embed_rank,  real_embed_rank,  tgt_dim;
	Align   *alignp;
	Distr   *distp;
	Map     *mapp;
	Symbol  *tgt_sp;
	/*}}}*/

	alignp = sp->alignp;				/* may be NULL */
	tgt_sp = alignp ? alignp->tgt_sp : NULL_Sym;	/* may be NULL */
	inherited = (alignp  &&  ! tgt_sp);		/* TRUE for 'INHERIT'*/
	aligned_with_inherit = (alignp  &&  tgt_sp  &&  tgt_sp->alignp);
						/* TRUE for 'ALIGN -> INHERIT'*/
	distp = (alignp && tgt_sp) ? tgt_sp->distp : sp->distp;
							/* may be NULL */
	mapp = (Map *) ecalloc (sizeof (Map));
	attach_map (sp->mapp, mapp);

	/*{{{  find 'array_rank' & 'real_embed_rank'*/
	array_rank = sp->arp ? (sp->arp->attrs & RANK_MASK) : 0;
	
	if (alignp  &&  distp  &&
	 ! (inherited || aligned_with_inherit || (distp->attrs & DI_INHERIT_FMT)) ) {
		short  *ip;
	
		/*------------------------------------------------------*/
		/* Find 'real_embed_rank' -- the number of *processor*	*/
		/* array dims in which 'sp' is embedded.		*/
		/*							*/
		/* N.B.							*/
		/* ----							*/
		/* -- This may be < 'embed_rank' (the number of *align	*/
		/* target* dims in which 'sp' is embedded), as we don't	*/
		/* count embedding in *collapsed* align target dims.	*/
		/* -- We don't do this for INHERIT_FMT, since then we	*/
		/* can't tell which embedded dims are 'real' as we	*/
		/* don't known which target dims are collapsed.		*/
		/*------------------------------------------------------*/
		embed_rank = (alignp->attrs & AL_EMBED_RANK_MASK) >>
		                              AL_EMBED_RANK_SHIFT;
		real_embed_rank = embed_rank;
	
		for (dim = embed_rank,  ip = alignp->ed;    dim;    --dim,  ++ip) {
			/*----------------------------------------------*/
			/* '*ip' is the target dim.  If it's collapsed,	*/
			/* then 'dim' is embedded in a collapsed dim,	*/
			/* so forget it  (i.e. don't count it in	*/
			/* 'real_embed_rank').				*/
			/*----------------------------------------------*/
			if  (distp->fmt [*ip] & DFMT_COLLAPSED)
				--real_embed_rank;		/* omit this dim! */
		}
	}
	else
		real_embed_rank = 0;
	/*}}}*/

	/*{{{  set 'mapp->attrs'*/
	mapp->attrs = (sp->type & TY_TYPE) |
	              (array_rank << MA_ARRAY_RANK_SHIFT) |
	              (real_embed_rank << MA_EMBED_RANK_SHIFT) |
	              ( (distp || inherited || aligned_with_inherit) ?
	                                 MA_DISTRIB : MA_GLOBALLY_REPLIC );
	
	                    /* ...i.e. 'MA_DISTRIB' is set if 'sp' is  explicitly
	                       distributed, inherited or aligned with INHERIT!  */
	
	if (inherited)
		mapp->attrs |= MA_INHERITED;
	
	else if (aligned_with_inherit)
		mapp->attrs |= MA_ALIGNED_WITH_INHERIT;
	
	if (distp) {
		if (distp->attrs & DI_INHERIT_FMT)
			mapp->attrs |= MA_INHERIT_FMT;
	
		if (distp->attrs & DI_INHERIT_PROCS)
			mapp->attrs |= MA_INHERIT_PROCS;
	}
	
	if (array_rank  &&  ! (sp->arp->attrs & EXPL_SHAPE))
		/*--------------------------------------------------------
		The following flag means that 'sp' has assumed (or deferred!)
		shape or assumed size, so 'sp->arp->n' is set to an element
		of a 'mapvec'.  In this case we must ignore this 'mapp'
		for a dummy arg in a proc interface (i.e. *don't* remap the
		actual arg according to it), as it's meaningless in that context!
		--------------------------------------------------------*/
		mapp->attrs |= MA_ASSUMED_SHAPE;
	/*}}}*/
	if (array_rank) {
		/*{{{  install (al, an, as, b, tn, fmt, d) -- ok*/
		/*------------------------------------------------------*/
		/* Install  (al[], an[], as[], tn[], b[], fmt[], d[])	*/
		/*------------------------------------------------------*/
		Expr  **epp,  **alp,  **anp,  **asp,  **tnp,  **bp,  **np;
		short  *fmtp,  *dp;
		
		alp = mapp->al = epp = (Expr **) emalloc (5 * array_rank * sizeof (Expr *));
		anp = mapp->an = (epp += array_rank);
		asp = mapp->as = (epp += array_rank);
		tnp = mapp->tn = (epp += array_rank);
		 bp = mapp->b  = (epp += array_rank);
		
		fmtp = mapp->fmt = (short *) emalloc (2 * array_rank * sizeof (short));
		dp   = mapp->d   = (fmtp + array_rank);
		
		np = sp->arp->n;
		
		for (dim = 0;   dim < array_rank;
		      ++dim,  ++alp,  ++anp,  ++asp,  ++tnp,  ++bp,  ++fmtp,  ++dp,  ++np) {
		
			attach_expr (*anp, *np);	/* an = sp->arp->n [dim] in all cases */
		
			tgt_dim = alignp ? alignp->d [dim] : dim;	/* -1 for collapsed;
									   -2 for INHERIT */
			if (tgt_dim == -2) {
				/*{{{  INHERIT*/
				/*------------------------------------------------------*/
				/* A dimension of an INHERITed object.			*/
				/*							*/
				/* N.B.   We ignore any accompanying 'DISTRIBUTE'!	*/
				/*------------------------------------------------------*/
				attach_expr (*alp, alignp->al [dim]);
				attach_expr (*asp, alignp->as [dim]);
				attach_expr (*tnp, map_primary (sp->name, 0, map_subscr_tn (array_rank, dim)));
				attach_expr (*bp,  map_primary (sp->name, 0, map_subscr_b (array_rank, dim)));
				
				*fmtp = DFMT_INHERIT;
				*dp   = -2;		/* meaning 'unknown', as inherited! */
				/*}}}*/
			}
			else if (aligned_with_inherit  &&  (tgt_dim != -1)) {
				/*{{{  aligned (not collapsed) with INHERITed object*/
				/*--------------------------------------------------------------*/
				/*	A dimension that is aligned (*not* collapsed) with an	*/
				/* INHERITed object.						*/
				/*	In this case, the alignment parameters are combined	*/
				/* just as for any other multiple alignment (see		*/
				/* 'compose_align()'),  i.e. if 'sp' has alignment parameters	*/
				/* (al1, as1, tgt_dim) and dimension 'tgt_dim' of the INHERITed	*/
				/* object has alignment parameters (al2, as2, d2):		*/
				/*								*/
				/*	al =  al2 + (al1 * as2)					*/
				/*	as =  (as1 * as2)					*/
				/*	 d =  d2		{-2 here!}			*/
				/*								*/
				/* N.B. We ignore any 'DISTRIBUTE' for the INHERITed object!	*/
				/*--------------------------------------------------------------*/
				Expr  *al1 = alignp->al [dim],
				      *as1 = alignp->as [dim],
				      *al2 = tgt_sp->alignp->al [tgt_dim],
				      *as2 = tgt_sp->alignp->as [tgt_dim];
				
				attach_expr (*alp,  expr (al2, ND_PLUS_OP, expr (al1, ND_TIMES_OP, as2)));
				attach_expr (*asp,  expr (as1, ND_TIMES_OP, as2));
				attach_expr (*tnp,  map_primary (tgt_sp->name, 0, map_subscr_tn (array_rank, tgt_dim)));
				attach_expr (*bp,   map_primary (tgt_sp->name, 0, map_subscr_b  (array_rank, tgt_dim)));
				
				*fmtp = DFMT_INHERIT;
				*dp   = -2;		/* meaning 'unknown', as inherited! */
				
				/*}}}*/
			}
			else if  ( ! distp  ||  tgt_dim == -1  ||
			        distp->fmt [tgt_dim] & DFMT_COLLAPSED) {
		
				/*{{{  collapsed: (al,as,tn, b,fmt,d) = (0,1,n, n, COLLAPSED, -1)*/
				/*--------------------------------------------------------------*/
				/* A collapsed dimension.  Set its params to:			*/
				/*	(al, as, tn, b, fmt, d) = (0, 1, n, n, COLLAPSED, -1)	*/
				/*--------------------------------------------------------------*/
				attach_expr (*alp, int_lit_0);
				attach_expr (*asp, int_lit_1);
				attach_expr (*tnp, *np);
				attach_expr (*bp,  *np);
				
				*fmtp = DFMT_COLLAPSED;		/* collapsed */
				*dp   = -1;			/*     "     */
				/*}}}*/
			}
			else {
				if (alignp) {
					/*{{{  aligned:  (al,as,tn) = (alignp->al/as/->tgt->n)*/
					/*------------------------------------------------------*/
					/* Dimension aligned with 'tgt_sp'.  Set:		*/
					/*	(al, as) =  alignp->al / as [dim]		*/
					/*	     tn  =  tgt_sp->n [tgt_dim]			*/
					/*------------------------------------------------------*/
					attach_expr (*alp, alignp->al [dim]);
					attach_expr (*asp, alignp->as [dim]);
					attach_expr (*tnp, tgt_sp->arp->n [tgt_dim]);
					/*}}}*/
				}
				else {
					/*{{{  distrib directly:  (al,as,tn) = (0,1,n)*/
					/*----------------------------------------------*/
					/* Dimension distributed directly.  Set:	*/
					/*	(al, as) = (0, 1)			*/
					/*	     tn  =  n				*/
					/*----------------------------------------------*/
					attach_expr (*alp, int_lit_0);
					attach_expr (*asp, int_lit_1);
					attach_expr (*tnp, *np);
					
					/*}}}*/
				}
				/*{{{  (b, fmt, d) = (distr->b/fmt, procdim)*/
				/*--------------------------------------------------------------*/
				/*	b / fmt  =  distp->b / fmt [tgt_dim]			*/
				/*				-- i.e. format & blocksize of	*/
				/*				        ultimate align target.	*/
				/*	      d  =  proc_dim [tgt_dim]				*/
				/*								*/
				/* N.B. This also handles "DISTRIBUTE *"  (i.e. 'distp->attrs	*/
				/* & DI_INHERIT_FMT'), in which case:				*/
				/*								*/
				/*	distp->b []   =  "mapvec_target_name (...)"		*/
				/*	distp->fmt [] =  DFMT_INHERIT;				*/
				/*	procdim ()    =  -2  (meaning "unknown as inherited").	*/
				/*								*/
				/* There are no problems with 'INHERIT_PROCS'.			*/
				/*--------------------------------------------------------------*/
				attach_expr (*bp, distp->b [tgt_dim]);
				
				*fmtp = distp->fmt [tgt_dim];
				*dp   = proc_dim (distp->fmt, tgt_dim);
				/*}}}*/
			}
		}
		/*}}}*/
	}
	if (real_embed_rank) {
		/*{{{  install (ep, ed) -- ok*/
		/*--------------------------------------------------------------*/
		/* Install (ep [], ed []) for (real) embedded dimensions.	*/
		/*								*/
		/* N.B.  We *don't* have 'INHERIT_FMT', as in that case we set	*/
		/* (real_embed_rank = 0) as we can't statically determine which	*/
		/* dims are *really* embedded!					*/
		/*--------------------------------------------------------------*/
		Expr  **epp;
		short  *edp;
		
		epp = mapp->ep = (Expr **) emalloc (real_embed_rank * sizeof (Expr *));
		edp = mapp->ed = (short *) emalloc (real_embed_rank * sizeof (short));
		
		for (dim = 0;    dim < embed_rank;    ++dim) {
			Expr  *block_no;
		
			tgt_dim = alignp->ed [dim];
		
			if  (distp->fmt [tgt_dim] & DFMT_COLLAPSED)
				continue;	/* ...embedded in a collapsed dim; forget it! */
		
			*edp = proc_dim (distp->fmt, tgt_dim);
		
			/*--------------------------------------------------------
			Calculate 'ep' (= proc no):
			       { ei / b			-- BLOCK (b)
			  ep = { MOD (ei/b, n_procs)	-- CYCLIC (b), explicit procs
			       { "mapvec (...)"		-- CYCLIC (b), inherited procs
			--------------------------------------------------------*/
			block_no = expr (alignp->ei [dim], ND_DIV_OP, distp->b [tgt_dim]);
		
			if  (distp->fmt [tgt_dim] & DFMT_BLOCK)
				attach_expr (*epp, block_no);
		
			else if (distp->proc_sp) {
				/*----------------------------------------------*/
				/* explicit procs: "DISTRIBUTE ... [ONTO [*] P]"*/
				/*----------------------------------------------*/
				attach_expr (*epp,  mod_func (block_no,
				                           distp->proc_sp->arp->n [*edp]));
		
				delete_expr (block_no, LEAVE_NLINKS);	/* in case unused */
			}
			else {
				/*----------------------------------------------*/
				/* inherited procs:  "DISTRIBUTE ... ONTO *"	*/
				/*----------------------------------------------*/
				int  subscr = map_subscr_ep (array_rank, real_embed_rank, dim);
		
				attach_expr (*epp, map_primary (sp->name, 0, subscr));
		
				delete_expr (block_no, LEAVE_NLINKS);	/* since unused! */
			}
			++epp,  ++edp;
		}
		/*}}}*/
	}

	attach_expr (mapp->mapvec_ep, map_primary (sp->name, 0, -1));
	if (distp)
		mapp->proc_sp = distp->proc_sp;  /* NULL for 'inherit procs'*/
}

#if 0
***********************************
	/*{{{  DELETE THIS LATER !!*/
	/*----------------------------------------------------*/
	/* Install the sectional params, 'mapp->sectn, l, s' */
	/*----------------------------------------------------*/
	if (sect_e->token == ND_WHOLE_ARRAY) {
		/*{{{  whole array reference -- ok*/
		/*--------------------------------------------------------------*/ 
		/* Whole array reference.  Initialise 'mapp->sectn, l, s' to:	*/
		/*								*/
		/*	      |  contiguous   |  assumed shape |   pointer	*/
		/*	------|---------------|----------------|--------------	*/
		/*	sectl |       0       | i$at_n (sectl) |   copied	*/
		/*	sectn |       n       | i$at_n (sectn) |    from	*/
		/*	sects |       1       | i$at_n (sects) |   target	*/
		/*--------------------------------------------------------------*/ 
		Expr  **parent_sectl_p, **parent_sectn_p, **parent_sects_p;
		
		switch (array_type) {
		case  CONTIGUOUS :
		
			n_p = mapp->n;		/* re-initialise this */
			for (dim = 0;  dim < array_rank;  dim++) {
				attach_expr (*sectl_p++, int_lit_0);
				attach_expr (*sectn_p++, *n_p++);
				attach_expr (*sects_p++, int_lit_1);
			}
			break;
			
		case  ASSUMED_SHAPE :
			for (dim = 0;  dim < array_rank;  dim++) {
				attach_expr (*sectl_p++, map_primary (name, 0, map_sectl_index (dim)) );
				attach_expr (*sectn_p++, map_primary (name, 0, map_sectn_index (dim)) );
				attach_expr (*sects_p++, map_primary (name, 0, map_sects_index (dim)) );
			}
			break;
		
		case  ASSOC_PTR_SHAPE :
			parent_sectl_p = parent_mapp->sectl;
			parent_sectn_p = parent_mapp->sectn;
			parent_sects_p = parent_mapp->sects;
		
			for (dim = 0;  dim < array_rank;  dim++) {
				attach_expr (*sectl_p++, *parent_sectl_p++);
				attach_expr (*sectn_p++, *parent_sectn_p++);
				attach_expr (*sects_p++, *parent_sects_p++);
			}
			break;
		}
		/*}}}*/
	}
	else {
		/*{{{  array element or section -- ok*/
		/*--------------------------------------------------------------*/ 
		/* Initialise 'mapp->sectn, l & s' for an element or section.	*/
		/* (In the following, '*sp_lb_p' points to the declared lower	*/
		/* bound in each dimension.)					*/
		/*--------------------------------------------------------------*/
		Expr  *parent_sectl,  *parent_sectn,  *parent_sects,
		      **sp_lb_p,  *subscr,  *subscrlist;
		
		
		for (subscrlist = sect_e->right,  sp_lb_p = sp->arp->lb,  dim = 0;
		     subscrlist;
		     subscrlist = subscrlist->right,  sp_lb_p++,  dim++) {
		
			subscr = subscrlist->left;	/* expr node for a subscript */
		
			if (dim == array_rank) {
				errmsg ("reference to array %q has too many dimensions", name);
				return  FALSE;
			}
			
			/*{{{  set 'parent_sectl/n/s'*/
			switch (array_type) {
			case  CONTIGUOUS :
				parent_sectl = int_lit_0;
				parent_sectn = mapp->n [dim];
				break;
			case  ASSUMED_SHAPE :
				/*------------------------------------------------------*/
				/* N.B. These are garbage if unused, so 'protect' them	*/
				/* below and 'unprotect' them at the end (i.e. delete	*/
				/* them if their (nlinks == 0)).			*/
				/*------------------------------------------------------*/
				parent_sectl = map_primary (name, 0, map_sectl_index (dim));
				parent_sectn = map_primary (name, 0, map_sectn_index (dim));
				parent_sects = map_primary (name, 0, map_sects_index (dim));
			
				protect_expr (parent_sectl);
				protect_expr (parent_sectn);
				protect_expr (parent_sects);
				break;
			case  ASSOC_PTR_SHAPE :
				parent_sectl = parent_mapp->sectl [dim];
				parent_sectn = parent_mapp->sectn [dim];
				parent_sects = parent_mapp->sects [dim];
				break;
			}
			
			/*}}}*/
		
			if (subscr->token != ND_SUBSCR_TRIPLET) {
				/*{{{  a single subscript (l)*/
				/*-----------------------------------------------------------------*/
				/* A single subscript  (l)   (which, BTW, is always defined).      */
				/* Then:                                                           */
				/*	sectl = { (l - lb)                           -- contiguous */
				/*	        { parent_sectl + (l-lb)*parent_sects  -- sectional */
				/*	sectn = -1                                                 */
				/*	sects = undefined.                                         */
				/*-----------------------------------------------------------------*/
				Expr  *l_norm = expr (subscr, ND_MINUS_OP, *sp_lb_p);
				
				attach_expr (*sectl_p++, (array_type == CONTIGUOUS) ?
							   l_norm :
							   expr (parent_sectl, ND_PLUS_OP,
								 expr (l_norm, ND_TIMES_OP, parent_sects))
					    );
					    
				attach_expr (*sectn_p++,  int_lit_m1);
				/*}}}*/
			}
			else {
				/*{{{  a subscript triplet (l:u:s)*/
				/*--------------------------------------------------------------*/
				/* A subscript triplet  ([l]:[u][:s])  (where any or all of	*/
				/* 'l', 'u' and 's' may be NULL, i.e. any form of subscript	*/
				/* containing a colon ':').					*/
				/*								*/
				/* subscr->right = s, which may be NULL!			*/
				/*								*/
				/* subscr->left = 'ND_DOUBLET' or NULL!				*/
				/* In the former case, setting 'subscr' to the ND_DOUBLET, then:*/
				/*	subscr->left  = l;					*/
				/*	subscr->right = u;					*/
				/* either of which may be NULL!					*/
				/*--------------------------------------------------------------*/
				Expr  *l_norm,  *l_orig,  *u_orig,  *s_orig;
				
				s_orig = subscr->right ? subscr->right : int_lit_1;	/* nlinks > 0, so ok */
				attach_expr (*sects_p++, (array_type == CONTIGUOUS) ?
							   s_orig :
							   expr (s_orig, ND_TIMES_OP, parent_sects));
				
				if (subscr = subscr->left) {
					/*{{{  general section (l:u:s).  Set 'mapp->sectl' & 'n'*/
					/*-----------------------------------------------------------------*/
					/* General section ((l:u:s), with at least one of 'l' or 'u'       */
					/* given).  Then:                                                  */
					/*	sectl = { (l - lb)                           -- contiguous */
					/*	        { parent_sectl + (l-lb)*parent_sects  -- sectional */
					/*	sectn = MAX ((u-l+s)/s, 0)                                 */
					/*-----------------------------------------------------------------*/
					l_orig = subscr->left ? subscr->left : *sp_lb_p;
					u_orig = subscr->right ? subscr->right :
								 expr (*sp_lb_p, ND_PLUS_OP,
								       expr (parent_sectn, ND_MINUS_OP, int_lit_1));
					l_norm = subscr->left ? expr (l_orig, ND_MINUS_OP, *sp_lb_p) : int_lit_0;
					
					attach_expr (*sectl_p++, (array_type == CONTIGUOUS) ?
								   l_norm :
								   expr (parent_sectl, ND_PLUS_OP,
									 expr (l_norm, ND_TIMES_OP, parent_sects))
						    );
						   
					attach_expr (*sectn_p++,  sect_extent_func (l_orig, u_orig, s_orig));
					
					
					/*--------------------------------------------------------------*/
					/* ( N.B.:							*/
					/* 'l_orig' has (nlinks > 0), so it's ok.			*/
					/* 'u_orig' may have (nlinks == 0); but used once, so ok.	*/
					/* 'l_norm' may have (nlinks == 0); but used once, so ok.)	*/
					/*--------------------------------------------------------------*/
				/*}}}*/
				}
				else {
					/*{{{  section of the form (:[:s]).  Set 'mapp->sectl' & 'n'*/
					/*------------------------------*/
					/* section of the form (: [:s])	*/
					/*------------------------------*/
					attach_expr (*sectl_p++,  parent_sectl);
					
					if ((s_orig->type & EX_INT_LIT_CONST) && (int_val (s_orig)==1)) /* s=1 */
						attach_expr (*sectn_p++,  parent_sectn);
					else
						attach_expr (*sectn_p++,
							      sect_extent_func (int_lit_1,	/* l */
										parent_sectn,	/* u */
										s_orig));	/* s */
					/*}}}*/
				}
				/*}}}*/
			}
			/*{{{  delete 'parent_sectl/n/s' if they are garbage*/
			/*--------------------------------------------------------------*/
			/* If (array_type == ASSUMED_SHAPE), we created expression	*/
			/* nodes 'parent_sectl/n/s' that might not have been used.	*/
			/* To avoid them hanging around unused (i.e. as garbage), we	*/
			/* try to delete them.  This is risk-free, as the deletion	*/
			/* will only succeed if their (nlinks == 0) (i.e. if there are	*/
			/* no pointer references) to them.				*/
			/*--------------------------------------------------------------*/
			if (array_type == ASSUMED_SHAPE) {
				unprotect_expr (parent_sectl);
				unprotect_expr (parent_sectn);
				unprotect_expr (parent_sects);
			}
			/*}}}*/
		}
		
		if (dim < array_rank) {
			errmsg ("reference to array %q has too few dimensions", name);
			return  FALSE;
		}
		/*}}}*/
	}
	/*}}}*/
***********************************
#endif
/*}}}*/
/*}}}*/

/*}}}*/

#if 0
/*------------------ not needed by IDA -----------

/***************************************************************
'install_bounds()' AND install_array()' CURRENTLY DON'T CHANGE 'arp->n[]'
FROM 'ub -> n' -- I.E. STILL NEED TO CALL:
	np = *npp = array_extent_func (lbp, np);
SOMEWHERE!!
THIS WAS DONE IN THE OLD VERSION ('OLD_process_array_spec()'),
SO KEEP IT FOR TIME BEING!...

/*{{{  OLD_process_array_spec (arp, sp) *** DELETE EVENTUALLY -- MERGED WITH install_bounds()*/
OLD_process_array_spec (arp, sp)
/*----------------------------*/
	Array  *arp;
	Symbol  *sp;	/* may be NULL */
/*--------------------------------------------------------------*/
/* (1) Perform the following checks on the declared lower and	*/
/* upper bounds of an array (which, on entry, are stored in	*/
/* 'arp->lb' and 'arp->n' respectively):			*/
/*								*/
/* -- check that all bounds are legal specification exprs;	*/
/* -- classify the array shape as const, explicit, deferred,	*/
/*	assumed, assumed-size or mixed (the latter is an error!)*/
/* -- check that rank <= MAX_RANK.				*/
/*								*/
/* (2) Replace the contents of these slots as follows:		*/
/*								*/
/*	lb:	NULL -> 1 (unless 'deferred' -- i.e. ptr/alloc)	*/
/*		expr -- unchanged				*/
/*								*/
/*	n:	NULL -> "I$AT_name(..)" (unless 'deferred')	*/
/*		'*'  -- unchanged				*/
/*		expr -> n (i.e. MAX (ub-lb+1,0))		*/
/*								*/
/*	'sp' is passed in for the purpose of errmsgs.  If it	*/
/* isn't NULL, then the errmsg refers to 'sp->name'.  Otherwise,*/
/* it's assumed that the array spec appears in a type decl and	*/
/* no specific identifier is referred to!			*/
/*--------------------------------------------------------------*/

/*{{{  *** BIG DISCUSSION ****/
/*--------------------------------------------------------------
-- here, if all expl -- ok.
-- if ass shape (ub never given, but some lb are given), replace with
"I$AT_name (..)" -- **but** must know name to do this!!
-- if deferred (lb and ub) not given, leave NULL.
-- So, when we attach the 'arp' to an 'sp', we may have to modify entries
depending on the 'sp' type, eg:

If 'sp' given, can do some of these checks (viz, we know if dummy;
may not yet know about ptr/alloc/global):
Anyhow, check if DUMMY, POINTER, ALLOC, GLOBAL; if so, make nec mods.
-- If we have to invent 'var' dims, use 'map_primary()' ??!!

-- Errmsgs:

If 'sp == NULL' (i.e. in attrs of a typedecl):
	mixed: "cannot mix explicit and assumed/deferred dimensions"
	At end, leave as it is!! (Is this ok -- can we distinguish
typedecl from array specific spec??)
	When we install array spec to 'sp':
	-- dummy ->assumed,  && no more errmsgs!
	-- global ->var,  &&  errmsg
	-- alloc/ptr -> deferred,  &&  errmsg
	-- if not one of above, leave as it is (and fix at the end of decls!)
	
if (sp == dummy)
	mixed -- cannot mix explicit and assumed dimensions.
	At end ->assumed.
else if (sp == alloc/ptr)
	not def: -- dimensions of ALLOC/PTR array must be deferred (i.e. omitted)
	At end, -> def.
else if (sp == global)
	not const: -- dimensions of global array must be constant
	At end, -> var.
otherwise
	mixed -- cannot mix explicit and deferred dimensions.
	(but leave as it is).
	
At end of decls, check all array decls again (since now may have extra info!)




shape  |     dummy     |  alloc / ptr  |     global    |  local / auto
-------|---------------|---------------|---------------|-----------------
const  |       ok      |  ERR (->def)  |       ok      |  ok (not auto)
var    |       ok      |  ERR (->def)  |  ERR (->var)  |    ok (auto)
defer  |   ok (->ass)  |       ok      |  ERR (->var)  | ERR (->var,auto)
ass    |       ok      |  ERR (->def)  |  ERR (->var)  | ERR (->var,auto)
mixed  |  ERR (->ass)  |  ERR (->def)  |  ERR (->var)  | ERR (->var,auto)
-------------------------------------------------------------------------


sp = dummy:  expl ok;
--------------------------------------------------------------*/

/*}}}*/
{
	/*{{{  decls*/
	Expr  *lbp,  *np,  **lbpp,  **npp;
	int   arp_attrs = DEFERRED_SHAPE;
	int   dim;
	bool  ass_size_check = FALSE,
	      exprs_ok       = TRUE,
	      too_many_dims  = ((arp->attrs & RANK_MASK) > MAX_RANK);
	
	/*}}}*/

	if (too_many_dims) {
		clear_flags (arp->attrs, RANK_MASK);
		arp->attrs += MAX_RANK;
	}
	for (dim = 0,  lbpp = arp->lb,  npp = arp->n;
	     dim < (arp->attrs & RANK_MASK);
	     dim++,  lbpp++,  npp++) {
	     	
		/*{{{  check each dimension 'dim' & install correct 'lb' & 'n'*/
		
		lbp = *lbpp;
		np = *npp;
		
		/*{{{  check lower bound (in 'lbp')*/
		if (! lbp)
			/*------------------------------------------------------*/
			/* Lower bound not given -- replace it by '1'.		*/
			/*	N.B. If this were a pointer or allocatable	*/
			/* array, then we should really leave this NULL --	*/
			/* however, we're not considering such arrays!		*/
			/*------------------------------------------------------*/
			lbp = *lbpp = int_lit_1;
		else {
			/*----------------------*/
			/* Explicit lower bound */
			/*----------------------*/
			if (exprs_ok)
				exprs_ok = mustbe_spec_expr (lbp, sp);
			if (arp_attrs == DEFERRED_SHAPE)
				arp_attrs = ASSUMED_SHAPE;
		}
		/**** lbp->attrs |= EX_SAVE; ***/
		/*}}}*/
		
		/*{{{  check upper bound (stored in 'np') and replace by actual shape*/
		if (! np) {
			/*-----------------------*/
			/* upper bound not given */
			/*-----------------------*/
			if (arp_attrs & EXPL_SHAPE) {
				/*{{{  err -- cannot mix explicit and assumed shape dimensions*/
				if (sp)
					errmsg ("array %q: cannot mix explicit and assumed shape dimensions",
						sp->name);
				else
					errmsg ("cannot mix explicit and assumed shape dimensions");
				/*}}}*/
				arp_attrs = MIXED_SHAPE;
			}
		}
		
		else if (np->token == '*'  &&  !ass_size_check) {
			/*--------------------*/
			/* upper bound is '*' */
			/*--------------------*/
			if (dim > 0 && arp_attrs & (ASSUMED_SHAPE | DEFERRED_SHAPE)) {
				/*{{{  errmsg -- assumed shape array cannot have assumed size declarator (*)*/
				if (sp)
					errmsg ("assumed shape array %q cannot have assumed size declarator (*)",
						sp->name);
				else
					errmsg ("assumed shape array cannot have assumed size declarator (*)");
				
				arp_attrs = MIXED_SHAPE;
				/*}}}*/
			}
			if (dim < (arp->attrs & RANK_MASK) - 1) {
				/*{{{  err -- assumed size decl (*) can only appear in the last dim*/
				if (sp)
					errmsg ("array %q: assumed size declarator (*) can only appear in the last dimension",
						sp->name);
				else
					errmsg ("assumed size declarator (*) can only appear in the last dimension");
				
				arp_attrs = MIXED_SHAPE;
				/*}}}*/
			}
			ass_size_check = TRUE;
			if (arp_attrs != MIXED_SHAPE)
				arp_attrs = ASSUMED_SIZE;
		
			/**** COULD REMOVE THIS?? ALL CHECKING DONE OUTSIDE THIS FN?? *****/
			if (sp && not_set (sp->attrs, AT_DUMMY))
				/*{{{  err -- assumed size array 'A' is not a dummy array*/
				errmsg ("assumed size array %q is not a dummy array", sp->name);
				/*}}}*/
		}
		
		else {
			/*----------------------*/
			/* explicit upper bound */
			/*----------------------*/
			if (dim == 0)
				arp_attrs = CONST_SHAPE;
		
			else if (arp_attrs & (DEFERRED_SHAPE | ASSUMED_SHAPE)) {
				/*{{{  err -- cannot mix explicit and assumed shape dimensions*/
				if (sp)
					errmsg ("array %q: cannot mix explicit and assumed shape dimensions",
						sp->name);
				else
					errmsg ("cannot mix explicit and assumed shape dimensions");
				/*}}}*/
				arp_attrs = MIXED_SHAPE;
			}
			if (exprs_ok)
				exprs_ok = mustbe_spec_expr (np, sp);
		
			/*-------------------------------------------------------*/
			/* replace the upper bound (in 'np') by the extent */
			/*-------------------------------------------------------*/
			np = *npp = array_extent_func (lbp, np);
		
			if ((arp_attrs == CONST_SHAPE) && !(np->attrs & EX_INT_LIT_CONST))
				arp_attrs = VAR_SHAPE;
			
		/****	np->attrs |= EX_SAVE; ****/
		}
		/*}}}*/
		
		
		
		
		/*}}}*/
	}

	if (too_many_dims)
		errmsg ("maximum rank exceeded");

	arp->attrs = arp_attrs;
}



/*}}}*/
***************************************************************/
#endif
/*}}}*/
